#include "app.h"

/*!
 * \file   app.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:40:01 2006
 * 
 * \brief  Implementation of the application class  
 */

const int MODE_RADIOSITY = 0x0;
const int MODE_GLWIDGET  = 0x1;

const int mode = MODE_GLWIDGET;

bool Application::Init()
{
    FUNC("App::Init");

    AddSignalHandlers();

    kernel = new Kernel;
    if(mode == MODE_GLWIDGET)
	new SDLInput(kernel, false);
    else
	new SDLInput(kernel, true);
    new GLRender(kernel);

    kernel->AddChild(CATCHER(this),
		     MASKRender | MASKUpdate | MASKWindow |
		     MASKKeyboard | MASKMouse | MASKGFXResources |
		     MASKInitialSetup);

    kernel->MainLoop();
    ENDRETURN(true);
}

//! A vertex
struct Vert
{
  public:
    //! the position
    Vectorf pos;
    //! diffuse colour
    Vectorf diffuse;
    //! the normal
    Vectorf normal;
    //! the u co-ordinate
    float u,
	//! the v co-ordinate
	v;
    //! the bone this vertex is attached to
    int boneID;
};

bool Application::GotEvent(Event event)
{
    FUNC("App::GotEvent");
    switch(event.type)
    {
	case MSGRender:
	{
	    Event ev;
	    ev.type = MSGMouseMove;
	    Render();
	    assert(kernel);

	    ENDRETURN(kernel->SendEvent(CATCHER(this), ev));
	    break;
	}
	case MSGUpdate:
	{
	    if(!Update(event.update.time))
	    {
		Event ev;
		ev.type = MSGCloseWindow;
		assert(kernel);
		kernel->SendEvent(NULL, ev);
		ENDRETURN(false);
	    }
	    break;
	}
	case MSGReleaseResources:
	{
	    Quit();
	    break;
	}
	case MSGInitialSetup:
	{
	    srand((unsigned int)time(NULL));
	    x = y = z = 0;
	    z = 180.0f*20.0f;

	    objects.push_back(Obj(Vectorf(-3.0f, 15.0f, 0.0f), Vectorf(), 10.0f));
	    objects.push_back(Obj(Vectorf( 3.0f, 15.0f, 0.0f), Vectorf(), 10.0f));
	    objects.push_back(Obj(Vectorf(-3.0f, 10.0f, 0.0f), Vectorf(), 10.0f));
	    objects.push_back(Obj(Vectorf( 3.0f, 10.0f, 0.0f), Vectorf(), 10.0f));

	    break;
	}
	case MSGKey:
	{
	    keys = event.key;
	    if(event.key.keys[ESCAPE])
		ENDRETURN(false);

	    GLWidget::TextKeyPressWidgets(keys.txtkeys);

	    break;
	}
	case MSGMouse:
	{
	    cam.Look(event.mouse.dx,
		     event.mouse.dy,
		     event.mouse.dz);
	    
	    if(mode == MODE_GLWIDGET)
	    {
		GLWidget::MouseMoveWidgets(Vectorf(event.mouse.x,
						   event.mouse.y));
		GLWidget::MouseClickWidgets(event.mouse.left,
					    event.mouse.middle,
					    event.mouse.right);
	    }
	    break;
	}
	case MSGLoadGFXResources:
	{
	    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	    /*
	    tex.Load("neg_z.tga");

	    prog.AddVertShader("temp.vert");
	    prog.AddFragShader("temp.frag");
	    prog.Compile();

	    light.Init();
	    light.Enable();

	    light.Position(Vectorf(0.0f, 0.0f, -15.0f), true);
	    light.Attenuation(1.0f, 0.0f, 0.0f);
	    light.Ambient(Vectorf(0.1f, 0.1f, 0.1f));
	    light.Diffuse(Vectorf(0.9f, 0.9f, 0.9f));

	    cam.Init(0.02f, 0.5f);

	    */

	    if(mode == MODE_RADIOSITY)
	    {
		room = Model::New("room_nonsub.ms3d");
		Log << "subdividing" << endl;
		room->Setup(2);
		Log << "done subdividing" << endl;
		room->GenNormals(true);
		Log << "done normaling" << endl;
		room->SetMode(SEQUENTIAL, true);
		Log << "done updating" << endl;
		//cube->SetFrame(0.0f);
		//cube->SetMode(IMMEDIATE);
	    }
	    else if(mode == MODE_GLWIDGET)
		GLWidget::StartGUI("my.gui");
	    break;
	}
    }
    ENDRETURN(true);
}

bool Application::Update(int milliseconds)
{
    FUNC("App::Update");
    framecount++;
    timecount += milliseconds;
    cam.Move(keys.keys[int('w')], keys.keys[int('s')],
	     keys.keys[int(' ')], keys.mod & LCONTROL,
	     keys.keys[int('a')], keys.keys[int('d')]);

    cam.Update(milliseconds);
    
    if(mode == MODE_GLWIDGET)
    {
	if(GLWidget::GetByTag("red") &&
	   GLWidget::GetByTag("green") &&
	   GLWidget::GetByTag("blue"))
	{
	    glClearColor(GLWidget::GetByTag("red")->GetLevel(),
			 GLWidget::GetByTag("green")->GetLevel(),
			 GLWidget::GetByTag("blue")->GetLevel(),
			 0.0f);
	}

	char text[256];

	if(ISWIDGET(horz1) && ISWIDGET(horzl1))
	{
	    sprintf(text, "Slider 1: %01.2f", GLWidget::GetByTag("horz1")->GetLevel());
	    GLWidget::GetByTag("horzl1")->SetText(string(text));
	}
	if(ISWIDGET(horz1) && ISWIDGET(horzl2))
	{
	    sprintf(text, "Slider 2: %01.2f", GLWidget::GetByTag("horz2")->GetLevel());
	    GLWidget::GetByTag("horzl2")->SetText(string(text));
	}

	if((GLWidget::GetByTag("exitbut") && GLWidget::GetByTag("exitbut")->Clicked()) ||
	   (GLWidget::GetByTag("combo1") && GLWidget::GetByTag("combo1")->GetTextID() == 3))
	    ENDRETURN(false);

	int message = GLWidget::GetMessage();

	if(message == OK_BUTTON)
	    GETWIDGET(status)->SetText(string("Status: ") + "YOU CLICKED OK LOL");

	if(message == CANCEL_BUTTON)
	    GETWIDGET(status)->SetText(string("Status: ") + "YOU CLICKED CANCEL LOL");

	if(message == YES_BUTTON)
	    GETWIDGET(status)->SetText(string("Status: ") + "YOU CLICKED YES LOL");

	if(message == NO_BUTTON)
	    GETWIDGET(status)->SetText(string("Status: ") + "YOU CLICKED NO LOL");

	if(GETWIDGET(redbut)->BeenClicked())
	    GLWidget::MessageBox("John wanted this to do something, so it does something.", OK_BUTTON|CANCEL_BUTTON);

	if(GETWIDGET(greenbut)->BeenClicked())
	    GLWidget::MessageBox("John also wanted this to do something, so it does something.", YES_BUTTON|NO_BUTTON);

	if(GETWIDGET(bluebut)->BeenClicked())
	    GLWidget::MessageBox("John also wanted this to do something, so it does something.", YES_BUTTON|NO_BUTTON|CANCEL_BUTTON);

	if(GETWIDGET(mario)->BeenClicked())
	    GLWidget::MessageBox("This is a test of the emergency broadcast system.\nIf it had been a real emergency, you would be dead by now.");
    }
    
#ifdef DEBUG_FPS
    if(timecount > 5000)
    {
	if(mode == MODE_GLWIDGET)
	{
	    ostringstream oss;
	    oss << "FPS: " << int(1000.0f*framecount/timecount);

	    if(GLWidget::GetByTag("fps"))
		GLWidget::GetByTag("fps")->SetText(oss.str());
	    
	}

	Log << framecount << " frames in " << timecount/1000.0f << " seconds = " << 1000.0f*framecount/timecount << " FPS" << endl;
	timecount = framecount = 0;
    }
#endif

    for(size_t i=0; i < objects.size(); i++)
    {
	Vectorf force, acc;
	force += Vectorf(0.0f, -9.8f, 0.0f) * objects[i].mass;
	force += objects[i].vel * objects[i].vel * 0.1f;
	acc = force/objects[i].mass;
	float time = milliseconds/1000.0f;

	objects[i].pos += (acc/2.0f)*time*time + objects[i].vel*time;
	objects[i].vel += acc*time;
    }

    if(mode == MODE_RADIOSITY)
    {
	// 0
	static int curtri = -1;
	if(keys.txtkeys[int('r')])
	{
	    curtri++;
	    if(curtri > (signed)room->Triangles.size()-1)
		curtri -= room->Triangles.size();

	    Log << curtri << " " << room->Triangles.size() << ":-    " << endl;

	    int view[4];
	    int size=100;
	    glGetIntegerv(GL_VIEWPORT, view);
	    glViewport(0, 0, size, size);

	    tri = curtri;

	    FPSCamera a = cam;
	    SetupCam(curtri, 0, 0);
	    Render(false);
	    //cam = a;

	    float *data = new float[size*size];
	    glReadPixels(0, 0, size, size, GL_RED, GL_FLOAT, data);

	    float total = 0.0f;
	    for(int i=0; i < size*size; i++)
		total += data[i];

	    Log << "a: " << total << endl;

	    total /= (size*size);

	    total *= 10.0f;

	    total = max(0.0f, min(1.0f, total));

	    Log << "b: " << total << endl;

	    room->Triangles[curtri].diffuse = 
		Vectorf(total, total, total);

	    glViewport(0, 0, view[2], view[3]);
	    delete[] data;
	}

	if(keys.txtkeys[int('t')])
	{
	    Log << "updating" << endl;
	    room->SetMode(SEQUENTIAL, true, true);
	    Log << "done updating" << endl;
	}

	const int iterstodo = 1;
	static int numitersdone = 0;

	if(keys.txtkeys[int('y')] || numitersdone < iterstodo)
	{
	    numitersdone++;
	    Log << "pass " << numitersdone << endl;
	    int view[4];
	    glGetIntegerv(GL_VIEWPORT, view);
	    int size=100;
	    glViewport(0, 0, size, size);
	    float *data = new float[size*size];
	    FPSCamera a = cam;
	    for(size_t curtri=0; curtri < room->Triangles.size(); curtri++)
	    {
		room->Triangles[curtri].diffuse =
		    Vectorf(-1.0f, -1.0f, -1.0f);

		//room->Triangles[curtri].diffuse[0] =
		//Vectorf(-1.0f, -1.0f, -1.0f);
		//room->Triangles[curtri].diffuse[1] =
		//Vectorf(-1.0f, -1.0f, -1.0f);
		//room->Triangles[curtri].diffuse[2] =
		//Vectorf(-1.0f, -1.0f, -1.0f);
	    }

	    // 4743
	    for(size_t curtri=0; curtri < room->Triangles.size(); curtri++)
	    {
		Log << float(curtri)/float(room->Triangles.size())*100.0f << "%" << endl;

		/*
		  for(int curvert=0; curvert < 3; curvert++)
		  {
		*/
		/*
		  if(room->Vertices[room->Triangles[curtri].vertices[curvert]].diffuse.x > -0.5f)
		  {
		  Log << "!" << room->Vertices[room->Triangles[curtri].vertices[curvert]].diffuse.x << "! ";
		  continue;
		  }
		*/
		//if(room->Triangles[curtri].diffuse[curvert].x > -0.5f)
		//continue;

		float total = 0.0f;
		for(int view=0; view < 5; view++)
		{
		    SetupCam(curtri, 0, view);
		    Render(false);
		    glReadPixels(0, 0, size, size, GL_LUMINANCE, GL_FLOAT, data);
		    
		    float avg = 0.0f;
		    for(int i=0; i < size*size; i++)
			avg += data[i];
		    avg /= float(size*size);

		    if(view == 0)
			total += 1.8f*avg*0.5f;
		    else
			total += 1.8f*avg*0.25f;
		}

		//Log << total << " ";

		total = max(0.0f, min(1.0f, total));

		//Log << total << " ";

		room->Triangles[curtri].diffuse =
		    Vectorf(total, total, total);
		//room->Vertices[room->Triangles[curtri].vertices[curvert]].diffuse =
		//Vectorf(total, total, total);
		//room->Triangles[curtri].diffuse[curvert] =
		//Vectorf(total, total, total);
		//}
		//Log << endl;

	    }
	    delete[] data;
	    cam = a;
	    glViewport(0, 0, view[2], view[3]);

	    Log << "updating" << endl;
	    room->SetMode(SEQUENTIAL, true, true);
	    Log << "done updating" << endl;
	}
    }
    else if(mode == MODE_GLWIDGET)
    {
	GLWidget::UpdateWidgets(milliseconds);
    }

    ENDRETURN(true);
}

void Application::SetupCam(int t, int v, int i)
{
    //Vectorf pos = room->Vertices[room->Triangles[t].vertices[v]].pos;
    Vectorf pos = (room->Vertices[room->Triangles[t].vertices[0]].pos +
		   room->Vertices[room->Triangles[t].vertices[1]].pos +
		   room->Vertices[room->Triangles[t].vertices[2]].pos)/3.0f;
    norm = room->Triangles[t].normal;
    if(fabs(norm.y) > 0.9999f)
	right = Vectorf(1.0f, 0.0f, 0.0f);
    else
	right = norm | Vectorf(0.0f, 1.0f, 0.0f);

    up = right | norm;

    if(i == 0)
	cam.LookAt(pos, pos+norm, up);
    else if(i == 1)
	cam.LookAt(pos, pos+right, up);
    else if(i == 2)
	cam.LookAt(pos, pos-up, norm);
    else if(i == 3)
	cam.LookAt(pos, pos-right, up);
    else if(i == 4)
	cam.LookAt(pos, pos+up, -norm);
}

bool Application::Render(bool real)
{
    FUNC("App::Render");
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();

    glClearColor(0.0f, 0.0f, 1.0f, 1.0f);

    if(mode == MODE_RADIOSITY)
    {
	glEnable(GL_COLOR_MATERIAL);
	glDisable(GL_LIGHTING);

	static bool kd = false;

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	if(real && (keys.txtkeys[int('k')] || kd))
	{
	    if(keys.txtkeys[int('k')])
		kd = !kd;
	    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	    glDisable(GL_COLOR_MATERIAL);
	    glEnable(GL_LIGHTING);
	}

	static int view = 0;

	if(keys.txtkeys[int('l')])
	{
	    tri--;
	    tri %= room->Triangles.size();
	}
	if(keys.txtkeys[int('i')])
	{
	    tri++;
	    tri %= room->Triangles.size();
	}
	if(keys.txtkeys[int('j')])
	{
	    Log << tri << " " << room->Triangles[tri].diffuse << endl;
	}
	if(keys.txtkeys[int('o')])
	{
	    tri = rand()%room->Triangles.size();view = 0;
	    SetupCam(tri, 0, view);
	    Log << tri << " " << room->Triangles[tri].diffuse << endl;
	}
	if(keys.txtkeys[int('p')])
	{
	    view++;
	    if(view > 4)
		view = 0;
	    SetupCam(tri, 0, view);
	}

	glEnable(GL_CULL_FACE);

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);

	cam.Transform();

	glColor3f(1.0f, 1.0f, 1.0f);

	light.Position(Vectorf(0.0f, 0.0f, -15.0f), true);
	room->Render(true);

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	tex.Bind();

	glColor3f(1.0f, 0.0f, 0.0f);
	glBegin(GL_POINTS);
	glVertex3fv((room->Vertices[room->Triangles[tri].vertices[0]].pos+
		     room->Vertices[room->Triangles[tri].vertices[1]].pos+
		     room->Vertices[room->Triangles[tri].vertices[2]].pos)/3.0f);
	glEnd();

	FPSCamera a = cam;

	glLineWidth(4.0f);
    
	if(keys.keys[int('[')])
	{
	    glBegin(GL_LINES);
	    for(size_t i=0; i < room->Triangles.size(); i++)
	    {
		SetupCam(i, 0, 0);
		glColor3f(1.0f, 0.0f, 0.0f);
		glVertex3fv((room->Vertices[room->Triangles[i].vertices[0]].pos+
			     room->Vertices[room->Triangles[i].vertices[1]].pos+
			     room->Vertices[room->Triangles[i].vertices[2]].pos)/3.0f);
		glVertex3fv((room->Vertices[room->Triangles[i].vertices[0]].pos+
			     room->Vertices[room->Triangles[i].vertices[1]].pos+
			     room->Vertices[room->Triangles[i].vertices[2]].pos)/3.0f+norm);

		glColor3f(0.0f, 1.0f, 0.0f);
		glVertex3fv((room->Vertices[room->Triangles[i].vertices[0]].pos+
			     room->Vertices[room->Triangles[i].vertices[1]].pos+
			     room->Vertices[room->Triangles[i].vertices[2]].pos)/3.0f);
		glVertex3fv((room->Vertices[room->Triangles[i].vertices[0]].pos+
			     room->Vertices[room->Triangles[i].vertices[1]].pos+
			     room->Vertices[room->Triangles[i].vertices[2]].pos)/3.0f+up);

		glColor3f(0.0f, 0.0f, 1.0f);
		glVertex3fv((room->Vertices[room->Triangles[i].vertices[0]].pos+
			     room->Vertices[room->Triangles[i].vertices[1]].pos+
			     room->Vertices[room->Triangles[i].vertices[2]].pos)/3.0f);
		glVertex3fv((room->Vertices[room->Triangles[i].vertices[0]].pos+
			     room->Vertices[room->Triangles[i].vertices[1]].pos+
			     room->Vertices[room->Triangles[i].vertices[2]].pos)/3.0f+right);
	    }
	    glEnd();
	}
    
	cam = a;

	glTranslatef(-90.0f, 90.0f, 0.0f);
	glPointSize(5.0f);
	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_QUADS);
	glTexCoord2f(1.0f, 1.0f);glVertex3f(-80.0f,  80.0f, 120.0f);
	glTexCoord2f(0.0f, 1.0f);glVertex3f( 80.0f,  80.0f, 120.0f);
	glTexCoord2f(0.0f, 0.0f);glVertex3f( 80.0f, -80.0f, 120.0f);
	glTexCoord2f(1.0f, 0.0f);glVertex3f(-80.0f, -80.0f, 120.0f);
	glEnd();
    }
    else if(mode == MODE_GLWIDGET)
    {
	GLWidget::RenderWidgets();
    }
    ENDRETURN(true);
}

bool Application::Quit()
{
    FUNC("App::Quit");

    if(mode == MODE_RADIOSITY)
	delete room;

    GLWidget::Quit();

    ENDRETURN(true);
}

int main(int argc, char **argv)
{
    BeginDebug();

    FUNC("main");

    {
	Pointer<Application> app = new Application;
	app->Init();
    }

    MMObject::CollectGarbage();
    MMObject::CollectAll();

    ENDRETURN(0);
}
