#ifndef CAMERA_H
#define CAMERA_H

/*!
 * \file   camera.h
 * \author  <baldurk@baldurk.org>
 * \date   Tue Feb 14 16:41:26 2006
 * 
 * \brief  classes to control cameras
 */

#include "general.h"
#include <GL/glu.h>

//! a general abstract camera
class Camera
{
  public:
    //! a virtual destructor
    virtual ~Camera() {}

    /*!
     * Initialise the camera, set everything to defaults and
     * set up the settings for speed
     * 
     * \param tspeed the translation factor
     * \param rspeed the rotation factor
     */
    virtual bool Init(float tspeed, float rspeed) = 0;

    /*!
     * Reinitialise the camera,
     * setting only the settings for speed
     * 
     * \param tspeed the translation factor
     * \param rspeed the rotation factor
     */
    virtual bool Reinit(float tspeed, float rspeed) = 0;

    //! Transform the current view by the camera
    virtual bool Transform() = 0;

    /*!
     * Move the camera
     * 
     * \param forward  if the camera should be moved forward
     * \param backward if the camera should be moved back
     * \param up       if the camera should be moved up
     * \param down     if the camera should be moved down
     * \param left     if the camera should be moved left
     * \param right    if the camera should be moved right
     */
    virtual void Move(bool forward, bool backward,
		      bool up, bool down,
		      bool left, bool right) = 0;

    /*!
     * Look with the camera
     * 
     * \param dx the angle to rotate in the x direction
     * \param dy the angle to rotate in the y direction
     * \param dz the angle to rotate in the z direction
     */
    virtual void Look(float dx, float dy, float dz) = 0;

    /*!
     * Update the camera, given the time interval
     * 
     * \param time the time in milliseconds since the last update
     */
    virtual bool Update(float time) = 0;

    /*!
     * Translate the camera by a vector
     * 
     * \param trans the vector to translate by
     */
    virtual bool Translate(Vectorf trans) = 0;

    /*!
     * Set up the camera by a series of vectors
     * 
     * \param pos  the position the camera should be at
     * \param look the position the camera should look at
     * \param up   the direction that is "up" for the camera
     */
    virtual bool LookAt(Vectorf pos, Vectorf look, Vectorf up) = 0;

    //! Get the position of the camera
    virtual Vectorf GetPos() = 0;
};

class FPSCamera : public Camera
{
  public:
    /*!
     * Initialise the camera, set everything to defaults and
     * set up the settings for speed
     * 
     * \param ts the translation factor
     * \param rs the rotation factor
     */
    bool Init(float ts = 1.0f, float rs = 1.0f);

    /*!
     * Reinitialise the camera,
     * setting only the settings for speed
     * 
     * \param ts the translation factor
     * \param rs the rotation factor
     */
    bool Reinit(float ts = 1.0f, float rs = 1.0f);

    /*!
     * Move the camera
     * 
     * \param forward  if the camera should be moved forward
     * \param backward if the camera should be moved back
     * \param up       if the camera should be moved up
     * \param down     if the camera should be moved down
     * \param left     if the camera should be moved left
     * \param right    if the camera should be moved right
     */
    void Move(bool forward, bool backward,
	      bool up, bool down,
	      bool left, bool right);

    /*!
     * Look with the camera
     * 
     * \param dx the angle to rotate in the x direction
     * \param dy the angle to rotate in the y direction
     * \param dz the angle to rotate in the z direction
     */
    void Look(float dx, float dy, float dz);

    /*!
     * Update the camera, given the time interval
     * 
     * \param time the time in milliseconds since the last update
     */
    bool Update(float time);

    //! Transform the current view by the camera
    bool Transform();

    /*!
     * Translate the camera by a vector
     * 
     * \param Trans the vector to translate by
     */
    bool Translate(Vectorf Trans);

    /*!
     * Set up the camera by a series of vectors
     * 
     * \param pos  the position the camera should be at
     * \param look the position the camera should look at
     * \param up   the direction that is "up" for the camera
     */
    bool LookAt(Vectorf pos, Vectorf look, Vectorf up);

    //! Get the position of the camera
    Vectorf GetPos() { return trans; }
  private:
    //! the camera's position
    Vectorf trans,
	//! the camera's rotation
	rot;
    Vectorf up, right, forward;
    //! the last time given to update
    float Time,
	//! the speed at which to translate
	TSpeed,
	//! the speed at which to rotate
	RSpeed;
};

#endif
