
/*!
 * \file   compress.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:43:34 2006
 * 
 * \brief  Compression related functions
 */


#undef DEBUG
#undef PROFILE

#ifdef DEBUG_SLOW
#define DEBUG
#endif

#ifdef PROFILE_SLOW
#define PROFILE
#endif

#include "log.h"
#include "debug.h"

#define MAXBITS 15              // maximum bits in a code
#define MAXDCODES 30            // maximum number of distance codes
#define FIXLCODES 288           // number of fixed literal/length codes

static unsigned char *Source, *Dest;
static unsigned int srclen, *dstlen;
static int bitbuf = 0, bitcnt = 0, currentbyte = 0, currentbit = 1;

int bits(int n) 		// return n more bits
{
    FUNC("bits");
    long buf;           // first we grab as many bytes as we need into here

    buf = bitbuf;       // start with whatever we had left from the last byte
    while(bitcnt < n) // keep going until we have n bits
    {
	// bitcnt is how many bits left in the last byte.
	// we take the next byte and shift it left by that number
	// so that it neatly tacks on above whatever is in buf now
        buf |= (long)(Source[currentbyte++]) << bitcnt;

	// we have 8 more bits
        bitcnt += 8;
    } 

    // whatever is left is buf shifted right to lose the bits we're about to return
    // this means that bitbuf is whatever is left of the last byte we read in that
    // we didn't use (this could be 0 if we used all the byte)
    bitbuf = buf >> n;
    // subtract n bits, because we are about to return them
    bitcnt -= n;

    // if you shift 1 left by 4 bits, then you get 10000 = 1 then 4 zeros.
    // if you subtract 1 then you get 1s on the first 4 bits: 10000 - 1 = 1111
    // if you binary & this with the bytes we read in, you get the first 4 bits
    // of those bytes

    ENDRETURN(buf & ((1L << n) - 1));
}

// decompress a stored block

int stored()
{    
    FUNC("stored");
    // skip the rest of the byte
    bitbuf = bitcnt = 0;

    int len = bits(8) | bits(8) << 8; // first two bytes are the length (low end byte stored first, then high end)
    int nlen = bits(8) | bits(8) << 8; // second two bytes are the one's compliment of the length
    
    // binary NOT the nlength should be the length
    // Mask off the two low end bytes to ensure that
    // only two bytes are used in the calculations
    if((len & 0xffff) != (~nlen & 0xffff))
    {
	Log << "Stored block has invalid length: " << len << " " << nlen << endl;
	ENDRETURN(1);
    }
    
    if(Dest != NULL)	// if the Dest pointer isn't NULL (isn't just getting the length)
	for(int i=*dstlen; i < len+(signed)*dstlen; ++i)	// for every byte to be reat
	    Dest[i] = bits(8);

    *dstlen += len;		// increment the bytes outputted by the length
    ENDRETURN(0);		// everything went fine
}

struct huffman
{
    // the number of symbols of each length
    int count[MAXBITS+1];
    // the symbols in this tree
    int symbol[FIXLCODES];
};

// the trees for fixed codes - lengths and distances
huffman fixed_len, fixed_dist;

// construct a huffman tree from a series of lengths
int construct(huffman *h, int *lens, int n)
{
    FUNC("construct");
    // fill the counts, lengths, symbols with 0s
    for(int i=0; i < FIXLCODES; ++i)
	h->symbol[i] = 0;
    for(int i=0; i < MAXBITS; ++i)
	h->count[i] = 0;

    // get the number of each length into h->count (so h->count[4] contains the number of 4 long symbols)
    for(int i=0; i < n; ++i)
	++h->count[lens[i]];
    
    // this contains the starting number for each code length
    int offs[MAXBITS+1];

    // this time we're calculating where in h->symbol each length will start
    offs[0] = offs[1] = 0;
    for(int bits = 1; bits < MAXBITS; ++bits)
	offs[bits+1] = offs[bits] + h->count[bits];

    // go through all the symbols
    for(int i=0; i < n; ++i)
	// fill them into h->symbol in order of their lengths
	// so long as it isn't a 0 length (unused) code
	if(lens[i] != 0 && offs[lens[i]] < FIXLCODES)
	    h->symbol[offs[lens[i]]++] = i;

    ENDRETURN(0);
}

int getcode(huffman *h)
{
    FUNC("getcode");
    // faster code getting - don't get bits one by one, get them in bytes
    // this code is essentially a combination of bits() and getcode()

    int len;		// current number of bits in code
    int code;		// len bits being decoded
    int first;		// first code of length len
    int index;		// index of first code of length len in symbol table
    int btbuf;		// bits from stream
    int left;		// bits left in next or left to process

    // get initial settings
    btbuf = bitbuf;
    left = bitcnt;
    code = first = index = 0;
    len = 1;

    // loop forever (there are break conditions
    while(1)
    {
	// loop from left down to 0 - using the current bits in the stream
        while(left)
	{
	    // get the last bit in the stream and OR it onto the current code
            code |= btbuf & 1;
	    // shift off the last bit we just used
            btbuf >>= 1;
	    // get the count of the next code length
            int count = h->count[len];
	    
	    // if the code lies in this length's range
            if(code < first + count)
	    {
		// update the global stream with what's left of our stream
                bitbuf = btbuf;
                bitcnt = (bitcnt - len) & 7;
		// return the symbol
                ENDRETURN(h->symbol[index + (code - first)]);
            }

	    // update the index and the first value
	    index += count;
            first += count;
            first <<= 1;
	    // shift the code left, update the length and number of bits left
            code <<= 1;
            ++len;
	    --left;
        }
	
	// we're out of bits and no match found

	// the number of bits we need to finish processing all lengths
        left = (MAXBITS+1) - len;
	// if we've overflowed, or we don't need any more bits (finished processing)
	// then we haven't found a matching code. Error out
        if(left == 0 || (unsigned)(currentbyte+1) > srclen)
	    ENDRETURN(-1);
	// get the next byte from the stream
        btbuf = Source[currentbyte++];
	// we don't want more than 8 bits to be used next time round, because we only
	// got 8 more.
	if(left > 8)
	    left = 8;
    }
    ENDRETURN(-1337);
}

int decode(huffman *lencode, huffman *distcodes)
{
    FUNC("decode");
    // for i (0 to 28) then lens[i] is the lowest length for i where i is symbol-257
    static const short lens[29] = { 
        3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
        35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258};
    // for i (0 to 28) then lext[i] is the number of extra bits to read
    // to determine what the length is (lens[i] + bits_read)
    static const short lext[29] = {
        0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2,
        3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0};
    // same as above but for distances
    static const short dists[30] = {
        1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
        257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
        8193, 12289, 16385, 24577};
    // same as above but for distances
    static const short dext[30] = {
        0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,
        7, 7, 8, 8, 9, 9, 10, 10, 11, 11,
        12, 12, 13, 13};

    // the literal or length code
    int litlen;
    do
    {
	// get the literal/length code
	litlen = getcode(lencode);

	// if < 256 then it's just a literal byte, so store it
	if(litlen < 256)
	{
	    if(Dest != NULL)
		Dest[(*dstlen)++] = litlen;
	    else
		++(*dstlen);
	}
	// if it's over 256 then it's a literal/length pair
	if(litlen > 256)
	{
	    // get the actual length index into the lens[] array
	    litlen -= 257;
	    // add the item from the lens[] array and (optionally) more bits as
	    // necessary (short lengths are described in only the code, longer
	    // lengths are a range, and need a few extra bits to determine the
	    // exact length).
	    int length = lens[litlen] + bits(lext[litlen]);

	    // get the distance code (the distance analogue to litlen)
	    int distcode = getcode(distcodes);

	    if(distcode == -1)
		ENDRETURN(-1);

	    // then get the actual distance in the same way as with the length
	    int distance = dists[distcode] + bits(dext[distcode]);

	    // if we're writing data
	    if(Dest != NULL)
		// for the length
		for(int i=0; i < length; ++i)
		{
		    // copy to the current byte, from distance ago (as *dstlen
		    // increases this will move along
		    Dest[(*dstlen)] = Dest[(*dstlen)-distance];
		    // increment the current byte
		    ++(*dstlen);
		}
	    else
	    {
		// if we're just geting the length, just increment the length
		*dstlen += length;
	    }
	}
	// if the litlen is 256, that's the end of the block
    } while(litlen != 256);

    ENDRETURN(0);
}

int fixed()
{
    FUNC("fixed");
    // only build tables the first time
    static bool first = true;
    if(first)
    {
	// the lengths, the current symbol for persistent for loops
	int lengths[FIXLCODES], symbol;

	// these lengths are set by the standard

	for(symbol = 0; symbol < FIXLCODES; ++symbol)
            lengths[symbol] = 8;
        for(symbol = 144; symbol < 256; ++symbol)
            lengths[symbol] = 9;
        for(; symbol < 280; ++symbol)
            lengths[symbol] = 7;

	// construct the literal/length table
        construct(&fixed_len, lengths, FIXLCODES);

        for (symbol = 0; symbol < MAXDCODES; ++symbol)
            lengths[symbol] = 5;

	// construct the distance table
        construct(&fixed_dist, lengths, MAXDCODES);
	first = false;
    }

    ENDRETURN(decode(&fixed_len, &fixed_dist));
}

int dynamic()
{
    FUNC("dynamic");
    int numlitlen = bits(5) + 257;
    int numdist = bits(5) + 1;
    int numcodes = bits(4) + 4;
    int lengths[FIXLCODES];
    int dists[MAXDCODES];
    int order[19] =
        {16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15};

    for(int i=0; i < 19; ++i)
	lengths[i] = 0;
    for(int i=0; i < numcodes; ++i)
	lengths[order[i]] = bits(3);

    huffman lencode, distcode;

    construct(&lencode, lengths, 19);

    for(int i=0; i < numlitlen;)
    {
	int symbol = getcode(&lencode);
	if(symbol < 16)
	{
	    lengths[i++] = symbol;
	}
	else
	{
	    int copy = 0, rep = 0;
	    if(symbol == 16)
	    {
		rep = bits(2) + 3;
		copy = lengths[i-1];
	    }
	    else if(symbol == 17)
		rep = bits(3) + 3;
	    else if(symbol == 18)
		rep = bits(7) + 11;
	    
	    for(int j=0; j < rep; ++j)
		lengths[i++] = copy;
	}
    }

    for(int i=0; i < numdist;)
    {
	int symbol = getcode(&lencode);
	if(symbol < 16)
	    dists[i++] = symbol;
	else
	{
	    int copy = 0, rep = 0;
	    if(symbol == 16)
	    {
		rep = bits(2) + 3;
		copy = dists[i-1];
	    }
	    else if(symbol == 17)
		rep = bits(3) + 3;
	    else if(symbol == 18)
		rep = bits(7) + 11;
	    
	    for(int j=0; j < rep; ++j)
		dists[i++] = copy;
	}
    }

    construct(&lencode, lengths, numlitlen);
    construct(&distcode, dists, numdist);

    ENDRETURN(decode(&lencode, &distcode));
}

// main decompression hook function

int Inflate(unsigned char *dest,       // pointer to destination pointer
	    unsigned int *destlen,    // amount of output space
	    unsigned char *source,     // pointer to source data pointer
	    unsigned int sourcelen)  // amount of input available
{
    FUNC("Inflate");
    // get the parameters
    Source = source;
    Dest = dest;
    srclen = sourcelen;
    dstlen = destlen;
    *dstlen = 0;
    // reset the buffer settings
    currentbyte = 0;currentbit = 1;
    bitcnt = 0;bitbuf = 0;

    int error = 0;		// is there an error
    int last;			// is this the last block
    do
    {
	last = bits(1);		// one bit indicates if this is the last block or not
	int type = bits(2);	// the next two indicate the type of block
	int ret = 0;		// return value from called function
	switch(type)
	{
	    case 0:		// stored literal bytes
		ret = stored();
		break;
	    case 1:		// compressed using fixed huffman trees
		ret = fixed();
		break;
	    case 2:		// compressed using dynamic huffman trees
		ret = dynamic();
		break;
	    case 3:		// any other values are invalid
	    default:
		Log << "Invalid block type" << endl;
		last = 1;error = 1;
		break;
	}

	if(ret != 0)		// if the decompression function didn't return 0, there's an error
	{
	    last = 1;error = 1;
	    Log << "Error decompressing block" << endl;
	}
    } while(last != 1);

    ENDRETURN(error);
}
