#ifndef DEBUG_H
#define DEBUG_H

/*!
 * \file   debug.h
 * \author  <baldurk@baldurk.org>
 * \date   Wed Jan  4 00:33:42 2006
 *
 * \brief  Debug routines  
 */

#define NO_MATH
#include "general.h"
#undef NO_MATH

#include <signal.h>

//! A structure that contains the details of a function call

struct FuncCall
{
    string name;		/*!< The name of the function */
};

//! The list of function calls
extern list <FuncCall> FunctionCalls;

//! This function begins debugging/profiling
extern void BeginDebug();
//! This function adds a function onto the funtion stack
extern void FuncEntry(const string funcname);
//! This function pops a function off the stack
extern void FuncLeave(const string funcname);
//! This marks the starting of a profiling section called funcname
extern void ProfileBegin(const string funcname);
//! This ends the profiling of the last function and returns the milliseconds taken
extern void ProfileEnd();
//! This will not return if value is false, and will output debugging information
void Assert(string expr, bool value);
//! This outputs a stacktrace to the Logger
extern bool StackTrace();
//! This sets up C-style signal handlers
extern bool AddSignalHandlers();

#undef assert

#if defined(DEBUG) && defined(PROFILE)

#define FUNC(a) const string _FuncName=a;FuncEntry(a); ProfileBegin(a);
#define END() {FuncLeave(_FuncName); ProfileEnd();}
#define ENDRETURN(a) {FuncLeave(_FuncName); ProfileEnd(); return (a);}
#define STACK() {StackTrace();}

/*! 
 * assert is included by many headers, so we can't have an assert function.
 * We declare our function as Assert then define assert to Assert.
 */
#define assert(a) Assert(#a, (a));

#elif defined(DEBUG) && !defined(PROFILE)

#define FUNC(a) const string _FuncName=a;FuncEntry(a);
#define END() FuncLeave(_FuncName)
#define ENDRETURN(a) {FuncLeave(_FuncName); return (a);}
#define STACK() StackTrace()

/*! 
 * assert is included by many headers, so we can't have an assert function.
 * We declare our function as Assert then define assert to Assert.
 */
#define assert(a) Assert(#a, (a));

#elif !defined(DEBUG) && defined(PROFILE)

#define FUNC(a) ProfileBegin(a)
#define END() ProfileEnd()
#define ENDRETURN(a) {ProfileEnd(); return (a);}
#define STACK()

/*! 
 * assert is included by many headers, so we can't have an assert function.
 * We declare our function as Assert then define assert to Assert.
 */
#define assert(a)

#else

#define FUNC(a)
#define END() 
#define ENDRETURN(a) return(a)
#define STACK()

/*! 
 * assert is included by many headers, so we can't have an assert function.
 * We declare our function as Assert then define assert to Assert.
 */
#define assert(a)

#endif

//! \todo implement an assert function
//! \todo Define macros to use profiling/debugging functions
//! \todo use the debugging functions to try out a fake game loop

#endif
