#ifndef EVENT_H
#define EVENT_H

#include "general.h"
#include "memory.h"
#include "messages.h"

/*!
 * \file   event.h
 * \author  <baldurk@baldurk.org>
 * \date   Fri Dec 30 21:46:59 2005
 *
 * \brief  Details of the Event class
 */

//! A macro to cast to a Pointer to an EventCatcher
#define CATCHER(a) Pointer<EventCatcher>(a)

//! EventMask is just an int
typedef int EventMask;

//! the data for a mouse event
struct MouseEvent
{
    MouseEvent() : x(0), y(0), z(0), left(false), middle(false), right(false) {}
    int x,	/*!< the x co-ordinate of the mouse */
	y,	/*!< the y co-ordinate of the mouse */
	z;	/*!< the z co-ordinate of the mouse */
    int dx,	/*!< change in the x co-ordinate of the mouse */
	dy,	/*!< change in the y co-ordinate of the mouse */
	dz;	/*!< change in the z co-ordinate of the mouse */
    bool left,  /*!< the status of the left button */
	middle, /*!< the status of the middle button */
	right;  /*!< the status of the right button */
};
//! the data for a keyboard event
struct KeyboardEvent
{
    KeyboardEvent() : mod(0) {FUNC("KeyboardEvent::KeyboardEvent");for(int i=0; i < 512; i++) keys[i]=txtkeys[i]=0;END();}
    bool keys[512];	/*!<
			 * An array specifying which of the keys are up
			 * or down at the time of this event
			 */
    bool txtkeys[512];	/*!<
			 * An array specifying which of the keys should
			 * be considered up or down to get the effects of
			 * normal text key pressing - an initial keypress,
			 * then a slow repetition after a pause.
			 */
    int mod;		/*!< the keyboard modifiers (shift, alt, capslock, etc) */
};
//! the data for a update event
struct UpdateEvent
{
    UpdateEvent() : time(0) {}
    int time;		/*!< the time since the last update event */
};

//! the data for a window opening
struct WinopenEvent
{
    WinopenEvent() : title(""), width(800), height(600), bpp(24), fullscreen(false) {}
    string title;	/*!< the title of the window */
    int width;		/*!< the width of the window */
    int height;		/*!< the height of the window */
    int bpp;		/*!< the bitdepth of the window */
    bool fullscreen;	/*!< whether the window is fullscreen or not */
};

/*! 
 * \brief A structure detailing all the information about an event
 *
 * This structure contains the type of and information about
 * a single event 
 */
struct Event
{
    Event() : type(0) {}
    int type;		/*!<
			 * Represents the type of the event,
			 * so that the correct data member can be used
			 */
    MouseEvent mouse;
    KeyboardEvent key;
    UpdateEvent update;
    WinopenEvent winopen;
};

/*! 
 * \brief A pure base class that receives events.
 *
 * If a class derives from this, it can receive events from a kernel.
 */
class EventCatcher : public MMObject
{
  public:
    //! a default constructor to optionally set the parent of the EventCatcher
    EventCatcher(Pointer<EventCatcher> rent = NULL) : parent(rent) {}
    //! a pure virtual function that inheriters must implement
    virtual bool GotEvent(Event) = 0;
  protected:
    Pointer<EventCatcher> parent; /*!< the eventcatcher's parent */
};

#endif
