#include "font.h"

/*!
 * \file   font.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:47:50 2006
 * 
 * \brief  Functions for rendering text
 */

bool FontRender::Init(string fontname)
{
    FUNC("FontRender::Init");

    FT_Face font;
    FT_Library lib;
    if(FT_Init_FreeType(&lib))
    {
	Log << "Error loading freetype library" << endl;
	ENDRETURN(false);
    }

    int error = 0;

    error = FT_New_Face(lib, fontname.c_str(), 0, &font);

    if(error == FT_Err_Unknown_File_Format)
    {
	Log << "The font file format is not recgonised" << endl;
	ENDRETURN(false);
    }
    else if(error)
    {
	Log << "Error loading the font" << endl;
	ENDRETURN(false);
    }

    const int dim = 32;

    if(FT_Set_Char_Size(font, dim*64, dim*64, 96, 96))
    {
	Log << "Error setting the font size" << endl;
	ENDRETURN(false);
    }

    chars = glGenLists(128);

    for(int i=0; i < 128; i++)
    {
	if(FT_Load_Glyph(font, FT_Get_Char_Index(font, i), FT_LOAD_DEFAULT))
	{
	    Log << "errors!" << endl;
	    ENDRETURN(false);
	}

	FT_Render_Glyph(font->glyph, ft_render_mode_normal);

	FT_Bitmap &bitmap = font->glyph->bitmap;

	int rval=1;while(rval < bitmap.width) rval<<=1;
	int width = rval;

	rval=1;while(rval < bitmap.rows) rval<<=1;
	int height = rval;

	unsigned char *buf = new unsigned char[width*height*4];
	memset(buf, 0, width*height*4);
	for(int y=0; y < bitmap.rows; y++)
	    for(int x=0; x < bitmap.width; x++)
	    {
		buf[4*(x+y*width)] = buf[4*(x+y*width)+1] = buf[4*(x+y*width)+2] =
		    bitmap.buffer[x+y*bitmap.width];
		buf[4*(x+y*width)+3] = bitmap.buffer[x+y*bitmap.width];
	    }

	TextureData texdata;
	texdata.Load(width, height, true, buf);

	Texture t;
	t.Load(texdata, TextureDetails("", XYCLAMP | BLACKWHITE));

	glNewList(chars+i, GL_COMPILE);
	t.Bind();

	glPushMatrix();
	glTranslatef(font->glyph->bitmap_left/float(dim),
		     (font->glyph->bitmap_top-bitmap.rows)/float(dim), 0.0f);

	float x = (float)bitmap.width / (float)width, y = (float)bitmap.rows / (float)height;

	glBegin(GL_QUADS);
	glTexCoord2f(0.0f, 0.0f); glVertex2f(0.0f, (bitmap.rows)/float(dim));
	glTexCoord2f(0.0f, y   ); glVertex2f(0.0f, 0.0f);
	glTexCoord2f(x,    y   ); glVertex2f((bitmap.width)/float(dim), 0.0f);
	glTexCoord2f(x,    0.0f); glVertex2f((bitmap.width)/float(dim), (bitmap.rows)/float(dim));
	glEnd();

	glPopMatrix();
	glTranslatef((font->glyph->advance.x >> 6)/float(dim), 0, 0);

	lengths[i] = (font->glyph->advance.x >> 6)/float(dim);

	glEndList();
    }

    ENDRETURN(true);
}

float FontRender::Length(const char *str, ...)
{
    FUNC("FontRender::Length");

    char text[512];
    memset(text, 0, 512);

    va_list ap;
    va_start(ap, str);
    vsprintf(text, str, ap);
    va_end(ap);

    vector<string> lines;
    const char *c;
    char buf[512];
    c = text;
    memset(buf, 0, 512);

    while(strchr(c, '\n'))
    {
	int len = strchr(c, '\n')-c;
	strncpy(buf, c, len);
	lines.push_back(string(buf));
	c = strchr(c, '\n')+1;
    }
    strcpy(buf, c);
    lines.push_back(string(buf));

    float maxwid = 0.0f;
    for(size_t i=0; i < lines.size(); i++)
    {
	float wid = 0.0f;
	for(size_t j=0; j < lines[i].size(); j++)
	    wid += lengths[(int)lines[i][j]]*xsize;

	if(wid > maxwid)
	    maxwid = wid;
    }

    ENDRETURN(maxwid);
}

bool FontRender::Print(const char *str, ...)
{
    FUNC("FontRender::Print");
    glPushAttrib(GL_ENABLE_BIT | GL_DEPTH_BUFFER_BIT);

    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);

    char text[256];

    va_list ap;
    va_start(ap, str);
    vsprintf(text, str, ap);
    va_end(ap);

    vector<string> lines;
    const char *c;
    char buf[512];
    c = text;
    memset(buf, 0, 512);

    while(strchr(c, '\n'))
    {
	int len = strchr(c, '\n')-c;
	strncpy(buf, c, len);
	lines.push_back(string(buf));
	c = strchr(c, '\n')+1;
	memset(buf, 0, sizeof(buf));
    }
    strcpy(buf, c);
    lines.push_back(string(buf));

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);      

    glListBase(chars);

    Matrix<float> modelview;modelview.GetModelView();

    assert(lines.size());

    for(int i=0; i < (signed)lines.size(); i++)
    {
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(0.0f, -1.3f*float(i)*ysize, 0.0f);
	modelview.ApplyModelView();

	glScalef(xsize, ysize, 0.0f);

	glCallLists(lines[i].length(), GL_UNSIGNED_BYTE, lines[i].c_str());

	glPopMatrix();
    }

    glPopAttrib();

    ENDRETURN(true);
}

