#include "glrender.h"

/*!
 * \file   glrender.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:48:27 2006
 * 
 * \brief  SDL and GL rendering system
 */

bool GLRender::GotEvent(Event e)
{
    FUNC("GLRender::GotEvent");
    if(e.type == MSGUpdate)
	ENDRETURN(Update());

    if(e.type == MSGRender)
	ENDRETURN(Render());

    if(e.type == MSGOpenWindow)
	ENDRETURN(OpenWindow(e.winopen.title, e.winopen.width, e.winopen.height,
			     e.winopen.bpp, e.winopen.fullscreen));

    if(e.type == MSGCloseWindow)
	ENDRETURN(Close());
    ENDRETURN(true);
}

bool GLRender::OpenWindow(string title, int width, int height, int bpp,
			  bool fullscreen)
{
    FUNC("GLRender::OpenWindow");

    if(SDL_WasInit(SDL_INIT_VIDEO) == 0)
    {
	if(SDL_InitSubSystem(SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) < 0)
	{
	    Log << "SDL_Init() failed. SDL Error: " << SDL_GetError() << endl;
	    SDL_QuitSubSystem(SDL_INIT_VIDEO);
	    ENDRETURN(false);
	}
    }

    if(fullscreen)
        SDL_ShowCursor(0);
    else
        SDL_ShowCursor(1);

    const SDL_VideoInfo *VideoInfo = SDL_GetVideoInfo();      // grab the video info
    
    if(!VideoInfo)                       // if we didn't get it
    {
        Log << "SDL_GetVideoInfo() failed. SDL Error: " << SDL_GetError() << endl;
        SDL_QuitSubSystem(SDL_INIT_VIDEO);
	ENDRETURN(false);
    }
    
    // add the default flags
    int Flags = SDL_OPENGL | SDL_GL_DOUBLEBUFFER | SDL_HWPALETTE;
    
    if(VideoInfo->hw_available)          // if hardware surfaces are enabled
        Flags |= SDL_HWSURFACE;
    else
        Flags |= SDL_SWSURFACE;

    if(fullscreen)                       // if we want fullscreen
        Flags |= SDL_FULLSCREEN;

    if(VideoInfo->blit_hw)               // if blitting is supported in hardware
        Flags |= SDL_HWACCEL;

    //! \todo add options for gl attributes
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1); // enable OpenGL Doublebuffering

    SDL_Surface *Surface = SDL_SetVideoMode(width, height, bpp, Flags); // create our surface
    
    if(!Surface)                         // if the surface wasn't created
    {
        Log << "SDL_SetVideoMode() Failed. SDL_Error: " << SDL_GetError() << endl;
        SDL_QuitSubSystem(SDL_INIT_VIDEO);
	ENDRETURN(false);
    }

    SDL_WM_SetCaption(title.c_str(), NULL);

    //! \todo make the opengl parameters more general

    // setup the OpenGL view
    glViewport(0, 0, width, height);
    
    glMatrixMode(GL_PROJECTION);
    
    gluPerspective(45.0f, float(width)/float(height), 1.0f, 10000.0f);
    
    glMatrixMode(GL_MODELVIEW);

    glClearColor(0.0f, 0.0f, 1.0f, 0.0f);

    if(!GLeeInit())
    {
	Log << "error initialising GLee" << endl;
	ENDRETURN(false);
    }

    ENDRETURN(true);
}

bool GLRender::Update()
{
    FUNC("GLRender::Update");

    SDL_Event e;
    vector<SDL_Event> saved;

    SDL_PumpEvents();

    while(SDL_PollEvent(&e))
    {
	if(e.type == SDL_QUIT)
	{
	    ENDRETURN(false);
	}
	else
	    saved.push_back(e);
    }

    SDL_Event *events = new SDL_Event[saved.size()];

    if(saved.size() && e.type != SDL_VIDEOEXPOSE)
    {
	for(int i=0; i < (signed)saved.size(); i++)
	    events[i] = saved[i];
	SDL_PeepEvents(events, saved.size(), SDL_ADDEVENT, 0);
    }

    delete[] events;
    ENDRETURN(true);
}

bool GLRender::Render()
{
    FUNC("GLRender::Render");
    //! \todo render here
    SDL_GL_SwapBuffers();
    ENDRETURN(true);
}

bool GLRender::Close()
{
    FUNC("GLRender::Close");
    SDL_QuitSubSystem(SDL_INIT_VIDEO);
    ENDRETURN(true);
}

GLRender::GLRender(Pointer<Kernel> k)
{
    FUNC("GLRender::GLRender(k)");
    assert(k);
    k->AddChild(CATCHER(this), MASKWindow | MASKUpdate | MASKRender, 700);
    END();
}
