#ifndef GLRENDER_H
#define GLRENDER_H

#include "GLee.h"
#include <GL/glu.h>

/*!
 * \file   glrender.h
 * \author  <baldurk@baldurk.org>
 * \date   Tue Jan  3 15:32:31 2006
 *
 * \brief  Contains a SDL/GL Rendering class  
 */


#include "general.h"
#include "kernel.h"
#include "memory.h"
#include "event.h"

/*! 
 * \brief A class to render to the screen
 *
 * This is a class that will render the current scene to the
 * screen using OpenGL. There shouldn't be any application
 * specific code in here. The application should set up
 * the initial conditions and pass the data for GLRender to
 * manage.
 */

class GLRender : public EventCatcher
{
  public:

    //! A default constructor. This shouldn't be used and will error when used.
    GLRender()
    { FUNC("GLRender::GLRender");assert(1==0 && "Don't use the default constructor");END(); }
    /*!
     * \brief A constructor that sets the kernel and adds itself as a child.
     * 
     * This constructor sets the kernel that the class should send
     * updates about the keyboards and mice. It then adds itself as
     * a child of that kernel.
     *
     * \param k the kernel this class should send updates to
     */

    GLRender(Pointer<Kernel> k);

    /*!
     * \brief This allows the class to receive events
     *
     * When the class receives an event, it comes into this
     * function, which is parsed and possibly passed into internal
     * functions.
     * 
     * \param e the event being received.
     */
    bool GotEvent(Event e);
  private:
    /*!
     * \brief Opens a window with the specified parameters
     * 
     * This function uses SDL to open up a window with the specified
     * parameters.
     *
     * \param title      The window's title [Default: 'BKEng mk.II']
     * \param width      The width of the window [Default: 800]
     * \param height     The height of the window [Default: 600]
     * \param bpp        The bitdepth of the window [Default: 24]
     * \param fullscreen Is the window fullscreen [Default: false]
     */
    bool OpenWindow(string title = "BKEng mk.II", int width = 800,
		    int height = 600, int bpp = 24, bool fullscreen = false);

    /*!
     * \brief Updates the window
     * 
     * This function pumps the SDL Event loop and takes any that it wants
     * and acts on them.
     */
    bool Update();

    /*!
     * \brief Renders the window
     * 
     * This function renders the current scene and swaps the buffers in
     * SDL.
     */
    bool Render();

    /*!
     * \brief Closes the window
     * 
     * This function closes the window.
     */
    bool Close();
};

#endif
