#ifndef GLWIDGET_H
#define GLWIDGET_H

#include "general.h"
#include "font.h"
#include "input.h"

/*!
 * \file   glwidget.h
 * \author  <baldurk@baldurk.org>
 * \date   Wed Mar  1 18:38:26 2006
 * 
 * \brief  class definitions for widgets
 *
 * \todo release widgets instantiated
 */

//! a theme class (glorified struct). These should only be
//! instantiated once per theme, and then pointed to in a
//! static map in GLWidget. They are listed in the way that
//! they should be read, so that if necessary binary read/writes
//! can be done. The last string is a char arrays for precisely
//! this reason. It must also be loaded into the Texture class
//! provided.
class Theme
{
  public:
    Vectorf textcol, bordercol, maincol,
	secondcol;

    float mainalpha, secondalpha;

    float border, insideborder;

    float corner, dimspeed;

    Vectorf textsize;

    float dimfact, brightfact;

    Vectorf clickoffset;

    float itemsize;

    char texture[512];

    bool highlight;

    Texture tex;

    void Read(string file);
    void Default();
};

#define OK_BUTTON     0x1
#define CANCEL_BUTTON 0x2
#define YES_BUTTON    0x4
#define NO_BUTTON     0x8

#define ISWIDGET(a) GLWidget::GetByTag(#a)
#define GETWIDGET(a) GLWidget::GetByTag(#a) && GLWidget::GetByTag(#a)

//! a generic widget class. It also contains static members
//! that can be accessed by all deriving members.
class GLWidget
{
  public:
    //! a virtual destructor
    virtual ~GLWidget() {}
    /*!
     * a constructor to set values common to all widgets
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     */
    GLWidget(string Tag, GLWidget *Parent, Vectorf Pos,
	     Theme *theme);

    //! returns true/false depending on whether the widget has focus
    bool HasFocus() { FUNC("GLWidget::HasFocus"); ENDRETURN(focus); }

    //! returns true/false depending on whether the widget is clicked.
    //! this means it is true all the time that the widget is clicjed
    bool Clicked() { FUNC("GLWidget::Clicked"); ENDRETURN(clicked); }

    //! returns true/false depending on whether the widget has been clicked
    bool BeenClicked() { FUNC("GLWidget::BeenClicked"); bool bc = beenclicked; beenclicked = false; ENDRETURN(bc); }

    //! returns the level - used for sliders and such.
    float GetLevel() { FUNC("GLWidget::GetLevel"); ENDRETURN(level); }

    //! sets the level - used for sliders and such.
    //! \param l the new level
    bool SetLevel(float l) { level = l; return true; }

    //! Get the text
    string GetText() { FUNC("GLWidget::GetText"); ENDRETURN(Text); }

    //! Brings up a popup with the specified message
    //! and optional buttons
    //! \param message the message
    //! \param l the new level
    static bool MessageBox(string message, int buttons=OK_BUTTON);

    //! Gets the message returned by the last messagebox, or 0 if
    //! none have returned. This also resets the last return to 0.
    static int GetMessage();

    //! Set the text
    //! \param t the text to set
    bool SetText(string t) { Text = t; return true; }
    
    //! Get the ID of the currently selected text - 0 to n
    int GetTextID() { FUNC("GLWidget::GetTextID"); ENDRETURN((int)currenttext); }

    //! Get the current text
    string GetCurrentText()
    {
	FUNC("GLWidget::GetCurrentText");
	ENDRETURN(currenttext < Texts.size() ? Texts[currenttext] : "");
    }

    //! Set which of the current texts is selected
    bool SetCurrentText(size_t i)
    {
	FUNC("GLWidget::GetCurrentText");
	(i < Texts.size() ? currenttext=i:i);
	ENDRETURN(true);
    }

    //! Add a text string to a list of texts
    //! \param t the text to add
    virtual bool AddText(string t)
    {
	FUNC("GLWidget::AddText");
	Texts.push_back(t);
	ENDRETURN(true);
    }

    //! centre a child widget in 1 or 2 dimensions
    //! \param child the child to centre
    //! \param x     whether to centre the child in the x axis, default true
    //! \param y     whether to centre the child in the y axis, default true
    virtual bool Centre(GLWidget *child, bool x=true, bool y=true);

    //! a static function to return the widget, referenced
    //! by a given string
    //! \param Tag the tag to look up
    static GLWidget *GetByTag(string Tag)
    { FUNC("GLWidget::GetByTag"); ENDRETURN(tags[Tag]); }

    //! A pure virtual function to process a mouse click
    //! 
    //! \note call this function AFTER MouseMove, otherwise results
    //!       are undefined
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    virtual bool MouseClick(bool left, bool middle, bool right) = 0;

    //! A mouse move function - this converts real co-ordinates to
    //! widget space
    //! 
    //! \note call this function BEFORE MouseClick, otherwise results
    //!       are undefined
    //! \param mp the mouse position
    //! \todo make the space constant size - say 600px = 1.0, so that
    //!       the windows scale properly with size of window
    virtual bool MouseMove(Vectorf mp);

    //! A virtual key press function
    //! \param keys the keys
    virtual bool TextKeyPress(bool keys[256]);

    //! A virtual mouse move function
    //! \param x the mouse x pos
    //! \param y the mouse y pos
    virtual bool MouseMove(float x, float y);

    //! A virtual update function
    virtual bool Update(int time);

    //! A pure virtual update function
    virtual bool Render() = 0;

    //! an overloaded += operator to move the object
    //! \param mod the vector to modify the position by
    virtual Vectorf operator +=(Vectorf mod)
    { FUNC("GLWidget::+=");pos += mod;ENDRETURN(pos); }

    //! Give the widget focus
    bool Focus() { FUNC("GLWidget::Focus");focus = true;focusdt = 1.0f;ENDRETURN(true); }

    //! Enable/click the widget
    bool Enable() { FUNC("GLWidget::Enable");clicked = true;clickdt = 0.0f;ENDRETURN(true); }

    //! returns if the widget should be unique in a given context
    virtual bool Unique() { FUNC("GLWidget::Unique"); ENDRETURN(false); }

    //! returns the widget's position
    Vectorf GetPos() { FUNC("GLWidget::GetPos"); ENDRETURN(pos); }

    //! returns the widget's tag
    string GetTag() { FUNC("GLWidget::GetTag"); ENDRETURN(tag); }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("GLWidget"); }

    //! returns the widget's depth
    float GetDepth() { FUNC("GLWidget::GetDepth"); ENDRETURN(z); }

    //! Add a child to this widget
    //! \param a the child to add
    virtual bool AddChild(GLWidget *a);

    //! return the bounding box. (x,y) is the position,
    //! (z,w) is the position plus the dimensions
    virtual Vectorf Getbbox()
    {
	FUNC("GLWidget::Getbbox");
	Vectorf vec;vec.x = pos.x;vec.y = pos.y;
	vec.z = pos.x+dim.x;vec.w = pos.y+dim.y;
	ENDRETURN(vec);
    }

    //! return the valid area - the area of the widget that children
    //! can appear in.
    virtual Vectorf GetValidArea()
    {
	FUNC("GLWidget::GetValidArea");
	ENDRETURN(Vectorf());
    }

    //! this should be called after all children have been added
    //! it will autoresize the widget to fit the children in
    virtual bool Solidify();

    //! A virtual key press function
    //! \param keys the keys
    static bool TextKeyPressWidgets(bool keys[256]);

    //! a static function that mousemoves all widgets
    static bool MouseMoveWidgets(Vectorf v);

    //! a static function that mousemoves all widgets
    static bool MouseClickWidgets(bool a, bool b, bool c);

    //! a static function that updates all widgets
    static bool UpdateWidgets(int i);

    //! a static function that renders all widgets
    static bool RenderWidgets();

    //! a static function to initialise the GUI from a file.
    //! \param fname the config file. Contains details of the gui
    static bool StartGUI(string fname);

    //! initialise the GUI given a config file. Should be called first
    static bool Init(string conffile);

    //! quit the GLWidget setup
    static bool Quit();

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    virtual bool SetFocussed(GLWidget *widget) = 0;

    //! Add a tab with an index and a name
    //! \param tab     the index of the tab to add
    //! \param tabname the name of the tab to be added
    virtual bool AddTab(int tab, string tabname)
    { FUNC("GLWidget::AddTab");ENDRETURN(true); }

    //! Add a widget to a tab given an index and a child
    //! \param tab the index of the tab to add to
    //! \param wid the widget to add to the tab
    virtual bool AddToTab(int tab, GLWidget *wid)
    { FUNC("GLWidget::AddToTab");ENDRETURN(true); }
  protected:
    //! returns true if the mouse is within the widget
    virtual bool Within();

    GLWidget *parent;		/*!< the parent of this widget */

    vector<string> Texts;	/*!< a list of the texts in this widget */
    size_t currenttext;		/*!< the index of the current text */

    static FontRender font;	/*!< an object to render text to the screen */

    int Time;			/*!< the current time difference */
    float z;			/*!< the depth of the widget */

    float focusdt,		/*!< a variable that goes from 0-1 to control focus highlighting */
	clickdt;		/*!< a variable that goes from 0-1 to control click lighting */

    float level;		/*!< the 0 - 1 level, used in sliders etc */
    bool lastleft,		/*!< the previous status of the left button */
	lastmiddle,		/*!< the previous status of the middle button */
	lastright;		/*!< the previous status of the right button */
    Vectorf pos,		/*!< the widget's position */
	dim,			/*!< the widget's dimensions */
	mpos;			/*!< the last mouse position */
    bool focus,			/*!< if the widget has focus */
	clicked,		/*!< if the widget is clicked */
	beenclicked;		/*!< if the widget has just been clicked */

    string Text;		/*!< the widget's "text" */

    static vector<GLWidget *> roots; /*!< the root widgets */

    static map<string,Theme*> themes; /*!< a static map of themes */

    Theme *theme;		/*!< the theme to use */

    static map<string,GLWidget*> tags; /*!< a static map of tags to widget pointers */
    string tag;			/*!< the widget's tag */

    static int viewport[4];	/*!< the window's viewport */
    static int msgreturn;	/*!< the message box return */
    static bool rendering;	/*!< whether or not we have rendering set up */
    static Vectorf optionbox,   /*!< the texture co-ordinates for option boxes */
	optiondot,              /*!< the texture co-ordinates for option box dots */
	checkbox,   		/*!< the texture co-ordinates for check boxes */
	checkdot,   		/*!< the texture co-ordinates for check box dots */
	slider; 		/*!< the texture co-ordinates for slider handles */

    static float timescale;	/*!< the scale that things change with time */

    vector<GLWidget *> children; /*!< this widget's children */
};

//! a window with a title and children
class GLWindow : public GLWidget
{
  public:
    /*!
     * A default constructor
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     * \param Title     the window's title
     * \param Draggable if the window can be dragged
     */
    GLWindow(string Tag, GLWidget *Parent, Vectorf Pos, Theme *theme,
	     string Title = "", bool Draggable = true) :
	GLWidget(Tag, Parent, Pos, theme)
    {
	FUNC("GLWindow::GLWindow()");
	dragging = false;
	Text = Title;
	draggable = Draggable;
	Tex = NULL; 
	END();
    }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("GLWindow"); }

    //! A virtual mouse move function
    //! \param x the mouse x pos
    //! \param y the mouse y pos
    bool MouseMove(float x, float y);

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right);

    //! A pure virtual update function
    bool Render();

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget);

    //! Add a child to this widget
    //! \param a the child to add
    bool AddChild(GLWidget *a);

    //! this should be called after all children have been added
    //! it will autoresize the widget to fit the children in
    bool Solidify();

    //! return the bounding box. (x,y) is the position,
    //! (z,w) is the position plus the dimensions
    Vectorf Getbbox()
    {
	FUNC("GLWidget::Getbbox");
	Vectorf vec;vec.x = pos.x;vec.y = pos.y;
	vec.z = pos.x+dim.x;vec.w = pos.y+dim.y;
	ENDRETURN(vec);
    }

    //! return the valid area - the area of the widget that children
    //! can appear in.
    Vectorf GetValidArea()
    {
	FUNC("GLWidget::GetValidArea");
	Vectorf vec;vec.x = pos.x+dim.x*0.01f;vec.y = pos.y+dim.x*0.01f;
	vec.z = pos.x+dim.x*0.99f;
	vec.w = pos.y+dim.y-dim.x*0.01f-theme->border-theme->textsize.y*1.5f;
	ENDRETURN(vec);
    }
  protected:
    //! returns true if the mouse is within the widget
    bool Within();

    //! the window's texture.
    Texture *Tex;

    //! the previous position of the mouse
    Vectorf prevmpos;

    //! if the window is being dragged
    bool dragging,
	//! if the window can be dragged
	draggable;
};

//! a checkbox
class Checkbox : public GLWidget
{
  public:
    /*!
     * A default constructor
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     * \param title     the check box's text
     */
    Checkbox(string Tag, GLWidget *Parent, Vectorf Pos,
	     Theme *theme, string title) :
	GLWidget(Tag, Parent, Pos, theme)
    {
	FUNC("Checkbox::Checkbox()");
	Text = title;
	if(Text != "")
	    dim = Vectorf(theme->itemsize+font.Length((Text+" ").c_str())+theme->textsize.x*1.0f,
			  max(theme->itemsize,theme->textsize.y*2.0f));
	else
	    dim = Vectorf(theme->itemsize, theme->itemsize);
	clicked = false;
	END();
    }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("Checkbox"); }

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right);

    //! A pure virtual update function
    bool Render();

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget)
    { FUNC("Checkbox::SetFocussed");ENDRETURN(true);}
  protected:
};

//! an optionbox
class Optionbox : public Checkbox
{
  public:
    /*!
     * A default constructor
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     * \param title     the option box's text
     */
    Optionbox(string Tag, GLWidget *Parent, Vectorf Pos,
	      Theme *theme, string title) :
	Checkbox(Tag, Parent, Pos, theme, title) { }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("Optionbox"); }

    //! A pure virtual update function
    bool Render();

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget);

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right);

    virtual bool Unique() { FUNC("GLWidget::Unique"); ENDRETURN(true); }
  private:
};

//! a button
class Button : public GLWidget
{
  public:
    /*!
     * A default constructor
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     * \param text      the button's text
     */
    Button(string Tag, GLWidget *Parent, Vectorf Pos,
	   Theme *theme, string text) :
	GLWidget(Tag, Parent, Pos, theme)
    {
	FUNC("Button::Button(string)");
	dim = Vectorf(font.Length(text.c_str())+theme->textsize.x*1.0f,
		      theme->textsize.y*1.5f, 0.0f);
	Text = text;
	END();
    }

    /*!
     * A constructor for buttons with textures
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     * \param Dim       the button's dimensions
     */
    Button(string Tag, GLWidget *Parent, Vectorf Pos, 
	   Theme *theme, Vectorf Dim) :
	GLWidget(Tag, Parent, Pos, theme)
    {
	FUNC("Button::Button(Vectorf)");
	dim = Dim;
	Text = "";
	Tex = theme->tex;
	END();
    }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("Button"); }

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right);

    //! A pure virtual update function
    bool Render();

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget)
    { FUNC("Checkbox::SetFocussed");ENDRETURN(true);}

    //! A virtual mouse move function
    //! \param x the mouse x pos
    //! \param y the mouse y pos
    bool MouseMove(float x, float y);
  private:
    //! the button's texture
    Texture Tex;
};

//! a label
class Label : public GLWidget
{
  public:
    /*!
     * A default constructor
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param text      the label's text
     */
    Label(string Tag, GLWidget *Parent, Vectorf Pos, Theme *theme, string text) :
	GLWidget(Tag, Parent, Pos, theme)
    {
	FUNC("Label::Label()");
	dim = Vectorf(font.Length(text.c_str())+theme->textsize.x*1.0f,
		      theme->textsize.y*1.5f*max(1,substr_count(text, "\n")), 0.0f);
	Text = text;
	END();
    }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("Label"); }

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right)
    {FUNC("Label::MouseClick");ENDRETURN(true);}

    //! A pure virtual update function
    bool Render();

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget)
    { FUNC("Label::SetFocussed");ENDRETURN(true);}
  private:
};

//! a slider - slides between 0 and 1 - horizontal or vertical
class Slider : public GLWidget
{
  public:
    /*!
     * A default constructor
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     * \param End       the slider's end point (\note this should only be vertical or horizontal)
     */
    Slider(string Tag, GLWidget *Parent, Vectorf Pos,
	   Theme *theme, Vectorf End) :
	GLWidget(Tag, Parent, Pos, theme)
    {
	FUNC("Slider::Slider()");
	end = dim = End;level = 0.5f;
	dragging = false;prevmpos = Vectorf(-5.0f, -5.0f);
	END();
    }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("Slider"); }

    //! A virtual mouse move function
    //! \param x the mouse x pos
    //! \param y the mouse y pos
    bool MouseMove(float x, float y);

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right);

    //! A pure virtual update function
    bool Render();

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget)
    { FUNC("Checkbox::SetFocussed");ENDRETURN(true);}

    //! return the bounding box. (x,y) is the position,
    //! (z,w) is the position plus the dimensions
    Vectorf Getbbox()
    {
	FUNC("GLWidget::Getbbox");
	Vectorf vec;
	if(end.x == 0.0f)
	{
	    vec.x = pos.x-theme->itemsize/2.0f;vec.y = pos.y;
	    vec.z = pos.x+dim.x+theme->itemsize/2.0f;vec.w = pos.y+dim.y+theme->itemsize;
	}
	else
	{
	    vec.x = pos.x;vec.y = pos.y-theme->itemsize/2.0f;
	    vec.z = pos.x+dim.x+theme->itemsize;vec.w = pos.y+dim.y+theme->itemsize/2.0f;
	}
	ENDRETURN(vec);
    }

  private:
    //! returns true if the mouse is within the widget
    bool Within();

    //! if the slider is being dragged
    bool dragging;

    //! the previous position of the mouse
    Vectorf prevmpos;

    //! the endpoint of the slider
    Vectorf end;
};

//! a tabbed window
class TabWindow : public GLWindow
{
  public:
    /*!
     * A default constructor
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     */
    TabWindow(string Tag, GLWidget *Parent, Vectorf Pos, Theme *theme) :
	GLWindow(Tag, Parent, Pos, theme, "", false)
    {
	FUNC("TabWindow()");
	Tex = NULL;
	currenttab = 0;
	END();
    }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("TabWindow"); }

    //! A virtual update function
    bool Update(int time);

    //! A pure virtual update function
    bool Render();

    //! Add a tab with an index and a name
    //! \param tab     the index of the tab to add
    //! \param tabname the name of the tab to be added
    bool AddTab(int tab, string tabname);

    //! Add a widget to a tab given an index and a child
    //! \param tab the index of the tab to add to
    //! \param wid the widget to add to the tab
    bool AddToTab(int tab, GLWidget *wid);

    //! A virtual mouse move function
    //! \param x the mouse x pos
    //! \param y the mouse y pos
    bool MouseMove(float x, float y);

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right);

    //! this should be called after all children have been added
    //! it will autoresize the widget to fit the children in
    bool Solidify();

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget);

    //! an overloaded += operator to move the object
    //! \param mod the vector to modify the position by
    Vectorf operator +=(Vectorf mod)
    {
	FUNC("GLWidget::+=");
	pos += mod;
	for(size_t i=0; i < tabs.size(); i++)
	    (*tabs[i].but) += mod;
	ENDRETURN(pos);
    }
  private:
    //! a structure to contain details of a tab
    struct Tab
    {
	Tab() : name(""), but(NULL) {}
	//! the name of the tab
	string name;
	//! the children in this tab
	vector<GLWidget *> children;
	//! the button to switch to this tab
	GLWidget *but;
    };
    //! a list of tabs
    vector<Tab> tabs;
    //! an index to the current tab
    int currenttab;
};

//! a frame - like a window but with no title bar and optionally no border.
//! Used for positioning and arrangement
class Frame : public GLWidget
{
  public:
    /*!
     * A default constructor
     * 
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     * \param Bordered  if the frame should have a border
     */
    Frame(string Tag, GLWidget *Parent, Vectorf Pos, Theme *theme,
	  bool Bordered = false) :
	GLWidget(Tag, Parent, Pos, theme)
    {
	FUNC("Frame::Frame()");
	focus = false;focusdt = 0.0f;bordered = Bordered;
	END();
    }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("Frame"); }

    //! A pure virtual update function
    bool Render();

    //! A virtual mouse move function
    //! \param x the mouse x pos
    //! \param y the mouse y pos
    bool MouseMove(float x, float y);

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right);

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget);

    //! return the bounding box. (x,y) is the position,
    //! (z,w) is the position plus the dimensions
    Vectorf Getbbox()
    {
	FUNC("GLWidget::Getbbox");
	Vectorf vec;vec.x = pos.x;vec.y = pos.y;
	vec.z = pos.x+dim.x;vec.w = pos.y+dim.y;
	ENDRETURN(vec);
    }

    //! return the valid area - the area of the widget that children
    //! can appear in.
    Vectorf GetValidArea()
    {
	FUNC("GLWidget::GetValidArea");
	Vectorf vec;vec.x = pos.x+dim.x*0.01f;vec.y = pos.y+dim.x*0.01f;
	vec.z = pos.x+dim.x*0.99f;
	vec.w = pos.y+dim.y-dim.x*0.01f-theme->border-theme->textsize.y*1.5f;
	ENDRETURN(vec);
    }
  protected:
    //! if the frame has a border
    bool bordered;
};

//! a combo box
class Combobox : public GLWidget
{
  public:
    /*!
     * A default constructor
     *
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     */
    Combobox(string Tag, GLWidget *Parent, Vectorf Pos, Theme *theme) :
	GLWidget(Tag, Parent, Pos, theme)
    {
	FUNC("Combobox::Combobox()");

	if(themes["combo_button.theme"] == NULL)
	{
	    themes["combo_button.theme"] = new Theme;
	    themes["combo_button.theme"]->Read("combo_button.theme");
	}

	children.push_back(new Button(string("__") + Tag + "dropbox", this,
				      Vectorf(), themes["combo_button.theme"], " "));
	dim = Vectorf(children[0]->Getbbox().z + theme->insideborder*6.0f,
		      children[0]->Getbbox().w + theme->insideborder*6.0f);
	menu = false;
	highlightidx = -1;
	END();
    }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("Combobox"); }

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right);

    //! A pure virtual update function
    bool Render();

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget)
    { FUNC("Checkbox::SetFocussed");ENDRETURN(true);}

    //! A virtual mouse move function
    //! \param x the mouse x pos
    //! \param y the mouse y pos
    bool MouseMove(float x, float y);

    //! Add a text string to a list of texts
    //! \param t the text to add
    bool AddText(string t);
  private:
    //! if the menu/dropdown is true
    bool menu,
	//! if the last update saw a click
	prevclick;
    //! the current index to highlight
    int highlightidx;
};

//! Text box - one line
class Textbox : public GLWidget
{
  public:
    /*!
     * A default constructor
     *
     * \param Tag       a unique string to identify the widget
     * \param Parent    the parent of this widget
     * \param Pos       the position of the widget
     * \param theme     the theme to use on this widget
     * \param width     the width of this widget
     * \param height     the height of this widget
     */
    Textbox(string Tag, GLWidget *Parent, Vectorf Pos, Theme *theme,
	    int width, int height) :
	GLWidget(Tag, Parent, Pos, theme)
    {
	FUNC("Textbox::Textbox()");
	Log << "textbox " << width << " " << height << endl;
	dim = Vectorf(font.Length(string(width, 'w').c_str())+theme->textsize.x*1.0f,
		      theme->textsize.y*1.3f*height, 0.0f);
	Log << "dim: " << dim << endl;
	Height = height;
	END();
    }

    //! returns the widget's type
    virtual string GetType() { FUNC("GLWidget::GetType"); ENDRETURN("Textbox"); }

    //! A virtual mouse move function
    //! \param x the mouse x pos
    //! \param y the mouse y pos
    bool MouseMove(float x, float y);

    //! A pure virtual function to process a mouse click
    //! \param left   if the left mouse button has been clicked
    //! \param middle if the middle mouse button has been clicked
    //! \param right  if the right mouse button has been clicked
    bool MouseClick(bool left, bool middle, bool right);

    //! a pure virtual function that indicates the focus has been
    //! given to another widget on the same level - only called
    //! if the widget returns true to Unique
    bool SetFocussed(GLWidget *widget);
    bool Unique() { FUNC("Textbox::Unique"); ENDRETURN(true); }

    virtual bool TextKeyPress(bool keys[]);

    //! A pure virtual update function
    bool Render();
  private:
    string start, end;
    int Height;
};

#endif
