#include "kernel.h"

/*!
 * \file   kernel.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:52:10 2006
 * 
 * \brief  implementation of the kernel class
 */

bool Kernel::AddChild(Pointer<EventCatcher> child, EventMask mask, int priority)
{
    FUNC("Kernel::AddChild");
    Child chld;
    chld.catcher = child;
    chld.mask = mask;
    chld.priority = priority;
    children.push_back(chld);

    OrganiseChildren();
    ENDRETURN(true);
}

bool Kernel::ChangeProps(Pointer<EventCatcher> child, EventMask mask, int priority)
{
    FUNC("Kernel::ChangeProps");
    for(vector<Child>::iterator it = children.begin(); it != children.end(); ++it)
	if((*it).catcher == child)
	{
	    (*it).mask = mask;
	    (*it).priority = priority;
	}

    OrganiseChildren();
    ENDRETURN(true);
}

bool Kernel::RemoveChild(Pointer<EventCatcher> child)
{
    FUNC("Kernel::RemoveChild");
    for(vector<Child>::iterator it = children.begin(); it != children.end();)
    {
	vector<Child>::iterator ptr = it;
	++it;
	if((*ptr).catcher == child)
	    children.erase(ptr);
    }

    OrganiseChildren();
    ENDRETURN(true);
}

bool Kernel::MainLoop()
{
    FUNC("Kernel::MainLoop");
    done = false;
    if(!timer)
	timer = new Timer;

    Event e;
    e.type = MSGLoadResources;
    done = !SendEvent(NULL, e);

    if(done)
    {
	Log << "Error sending load resources event" << endl;
	ENDRETURN(false);
    }

    e.type = MSGOpenWindow;
    //! \todo read files from a config. Shit, write a config reader.
    e.winopen.title = "BKEng Application";
    e.winopen.width = 800;
    e.winopen.height = 600;
    e.winopen.bpp = 24;
    e.winopen.fullscreen = false;
    done = !SendEvent(NULL, e);

    if(done)
    {
	Log << "Error sending openwindow event" << endl;
	ENDRETURN(false);
    }

    e.type = MSGLoadGFXResources;
    done = !SendEvent(NULL, e);

    if(done)
    {
	Log << "Error sending load graphics resources event" << endl;
	ENDRETURN(false);
    }

    e.type = MSGInitialSetup;
    done = !SendEvent(NULL, e);

    if(done)
    {
	Log << "Error sending initial setup event" << endl;
	ENDRETURN(false);
    }

    internalmask = MASKWindow;

    int prevtime = timer->GetTime();
    while(!done)
    {
	int currtime = timer->GetTime();
	Event ev;ev.type = MSGUpdate;ev.update.time = currtime-prevtime;
	done = !SendEvent(NULL, ev);
	if(done)
	    break;
	ev.type = MSGRender;
	done = done || !SendEvent(NULL, ev);

	prevtime = currtime;
    }

    e.type = MSGReleaseResources;
    done = !SendEvent(NULL, e);

    if(done)
    {
	Log << "Error sending release resources event" << endl;
	ENDRETURN(false);
    }

    ENDRETURN(true);
}

bool Kernel::SendEvent(Pointer<EventCatcher> child, Event e)
{
    FUNC("Kernel::SendEvent");
    ENDRETURN(PropogateEvent(child, e, true));
}

bool Kernel::SendEventTo(Pointer<EventCatcher> child, Event e)
{
    FUNC("Kernel::SendEventTo");
    ENDRETURN(PropogateEvent(child, e, false));
}

bool Kernel::PropogateEvent(Pointer<EventCatcher> child, Event e,
			bool global)
{
    FUNC("Kernel::PropogateEvent");
    vector<int> mask;

    // determine the event event, add all the masks it could match to the list
    switch(e.type)
    {
	case MSGLoadResources:
	    mask.push_back(MASKResourcesID);break;
	case MSGLoadGFXResources:
	    mask.push_back(MASKGFXResourcesID);break;
	case MSGReleaseResources:
	    mask.push_back(MASKGFXResourcesID);
	    mask.push_back(MASKResourcesID);break;
	case MSGInitialSetup:
	    mask.push_back(MASKInitialSetupID);break;
	case MSGUpdate:
	    mask.push_back(MASKUpdateID);break;
	case MSGRender:
	    mask.push_back(MASKRenderID);break;
	case MSGMouseMove:
	    mask.push_back(MASKMouseMoveID);break;
	case MSGMouse:
	    mask.push_back(MASKMouseID);break;
	case MSGKey:
	    mask.push_back(MASKKeyboardID);break;
	case MSGCloseWindow:
	case MSGOpenWindow:
	    mask.push_back(MASKWindowID);break;
	default:
	    // we've encountered an event not implemented, or an invalid type
	    Log << e.type << endl;
	    assert("Undefined event" && 0);
    }

    if(mask.size() == 0)
	assert("No matching event found, but no error returned" && 0);

    // for each matching mask
    for(int i=0; i < (signed)mask.size(); i++)
    {
	// spy on messages
	if(internalmask & (1 << mask[i]))
	    if(!InternalEvent(e))
	    {
		Log << "Error processing an internal event" << endl;
		ENDRETURN(false);
	    }
	// for each child wanting this mask (precalculated in OrganiseChildren)
	for(int j=0; j < (signed)messages[mask[i]].size(); j++)
	    // send the event on.
	    if((children[messages[mask[i]][j]].catcher == child && !global) ||
	       (children[messages[mask[i]][j]].catcher != child && global))
		if(!children[messages[mask[i]][j]].catcher->GotEvent(e))
		    ENDRETURN(false);
    }

    ENDRETURN(true);
}

bool Kernel::InternalEvent(Event e)
{
    FUNC("Kernel::InternalEvent");

    if(e.type == MSGCloseWindow)
	done = true;

    ENDRETURN(true);
}

bool Kernel::GotEvent(Event e)
{
    FUNC("Kernel::GotEvent");
    ENDRETURN(SendEvent(this, e));
}

bool Kernel::OrganiseChildren()
{
    FUNC("Kernel::OrganiseChildren");
    for(int i=0; i < NumMasks; i++)
    {
	messages[i].clear();
	for(int j=0; j < (signed)children.size(); j++)
	    if(children[j].mask & (1 << i))
		messages[i].push_back(j);

	std::sort(messages[i].begin(), messages[i].end());
    }
    ENDRETURN(true);
}
