#ifndef KERNEL_H
#define KERNEL_H

#include "general.h"
#include "memory.h"
#include "event.h"
#include "timer.h"

/*!
 * \file   kernel.h
 * \author  <baldurk@baldurk.org>
 * \date   Fri Dec 30 21:46:59 2005
 *
 * \brief  Details of the Kernel class
 */


/*! 
 * \brief A structure containing details on a child.
 *
 * This class is only used in Kernel so that the Kernel
 * can keep track of all the children that it is managing
 * and translating messages between.
 */

struct Child
{
    Pointer<EventCatcher> catcher; /*!<
				    * This is a Pointer to the actual
				    * object in question. This allows
				    * the Kernel to call its GotEvent
				    * function.
				    */
    EventMask mask;		/*!<
				 * This is a bitfield that specifies
				 * the messages that should be passed
				 * to this child.
				 */
    int priority;		/*!<
				 * This is an integer in the range 0 - 1000
				 * which indicates where it should be sent messages
				 * in relation to other children. 0 is first, 1000
				 * is last.
				 */

    /*!
     * \brief sets this to another child
     * 
     * \param c the child to set to
     */

    Child &operator =(const Child &c)
    {
	FUNC("Child::=");
	catcher = c.catcher;
	mask = c.mask;
	priority = c.priority;
	ENDRETURN(*this);
    }

    /*!
     * \brief compares to another child by priority
     * 
     * \param c the child to compare to
     */

    bool operator <(Child c)
    {
	FUNC("Child::<");
	ENDRETURN(priority < c.priority);
    }
};

/*! 
 * \brief A class that ties everything together.
 *
 * This class will keep a track of a list of all the objects
 * being managed by the core of the engine. Note this does not
 * necessarily mean that all objects sending and receiving messages
 * will be managed by the Kernel, but it generally does.
 *
 * The Kernel is designed so that it can be stacked in a tree-like fashion.
 * A server could run a kernel without a renderer, or input children. Then
 * a remote Kernel can be added as a child, asking for update events.
 * It can also send events to that kernel to be distributed to the other
 * children.
 */

class Kernel : public EventCatcher
{
  public:
    /*!
     * \brief A default constructor
     * 
     * This constructor initialises the done variable to false
     * just in case. MainLoop also does this.
     */

    Kernel() : done(false), internalmask(0) {FUNC("Kernel::Kernel");END();}
    /*! 
     * \brief Gets an event, then passes it down to the apropriate children.
     *
     * This function takes an event, then iterates through the list
     * of children in order to determine which of them should have this
     * message passed onto them.
     * 
     * \param e the event received
     */
    bool GotEvent(Event e);
    
    /*! 
     * \brief Changes the properties of a Child
     *
     * This function takes a Pointer to an eventcatcher and new properties.
     * It then updates its list so that the child specified by that eventcatcher
     * has the new properties. This allows children to change their properties.
     *
     * \param child    the child requesting messages
     * \param mask     the new mask to be set for the child
     * \param priority the new priority to be set for the child
     */
    bool ChangeProps(Pointer<EventCatcher> child,
		     EventMask mask=0, int priority=500);

    /*! 
     * \brief Sends an event to all children but the specified one
     *
     * This function will send the specified event to all the Kernel's
     * children except the specified child. If the parameter is NULL,
     * all children receive the event.
     *
     * \param child the child sending the event, so this child
     *              will not receive the event (optional: if NULL,
     *                all children receive the event. This is more
     *                efficient if the child sending ignores the event).
     * \param e     the event being sent
     *
     * \note This function respects children's eventmasks.
     */
    bool SendEvent(Pointer<EventCatcher> child, Event e);

    /*! 
     * \brief Sends an event to the specified child
     *
     * This function will send the specified event only to the child
     * specified by the parameter.
     *
     * \param child the destination for the event
     * \param e     the event being sent
     *
     * \note This function respects child's eventmasks.
     */
    bool SendEventTo(Pointer<EventCatcher> child, Event e);

    /*! 
     * \brief starts the Kernel main loop pumping
     *
     * \note This function should only be called from main() or some kind of mainloop.
     *
     * This function starts the Kernel main loop. Children should have been initialised
     * and added to the kernel before this. The main loop will then call all the children
     * at various points during the startup
     */
    bool MainLoop();

    /*! 
     * \brief Adds a child
     *
     * This function adds a child to the list of children, and sets the initial
     * properties to the ones specified.
     *
     * \param child    this is the child to be added
     * \param mask     this is the initial mask for the child
     * \param priority this is the initial mask for the child
     */
    bool AddChild(Pointer<EventCatcher> child,
		  EventMask mask=0, int priority=500);

    /*!
     * \brief Removes a child
     * 
     * This function removes the specified children from the list of children.
     *
     * \param child the child to remove
     */
    bool RemoveChild(Pointer<EventCatcher> child);

    /*! 
     * \brief Organises the children for efficiency
     *
     * This function sets up lists for each message that includes details of
     * which children want which message.
     */
    bool OrganiseChildren();
  private:

    /*!
     * \brief Sends events out based on the parameters
     * 
     * This propogates an event to various children depending
     * on what the parameters are.
     *
     * \note this function will respect masks
     *
     * \param child  this is either the child to ignore, or send to depending on
     *               global.
     * \param e      this is the event to be propogated.
     * \param global if true, child is the child to be ignored (it can be NULL).
     *               if false, child is the child to send the event to.
     */
    bool PropogateEvent(Pointer<EventCatcher> child, Event e,
			bool global);

    /*!
     * \brief Processes an event for the kernel itself
     * 
     * This function is for events scooped off the propogation list.
     * Events here will not be passed on anywhere as they are already
     * being passed on elsewhere. This is purely for internal processing
     * of event data.
     */
    bool InternalEvent(Event e);

    /*!
     * One list for each mask, containing indices into children vector,
     * so that only the children accepting each mask need be iterated through
     */
    vector<int> messages[NumMasks];
    vector<Child> children;	/*!< The list of children */
    bool done;			/*!< This indicates whether the main loop is done */
    int internalmask;		/*!< This is the mask of the events the kernel wants for itself */
    Pointer<Timer> timer;	/*!< The timer used in the main loop */
};

#endif
