#ifndef LIGHT_H
#define LIGHT_H

#include "general.h"
#include "GLee.h"

/*!
 * \file   light.h
 * \author  <baldurk@baldurk.org>
 * \date   Fri Jan  6 18:31:54 2006
 *
 * \brief  A wrapper for lights  
 */


/*!
 * \brief a wrapper for a Light
 * 
 */

class Light
{
  public:
    /*!
     * \brief A default constructor
     * 
     * If the constructor has never been called before,
     * it initialises the static list of light allocations.
     * It also gets a new light allocation
     */
    Light()
    {
	FUNC("Light::Light()");
	if(!init)
	{
	    for(int i=0; i < GL_MAX_LIGHTS; i++)
		light[i] = false;
	    init = true;
	}

	Name = 0;

	Init();
	END();
    }

    /*!
     * \brief A default destructor
     * 
     * The destructor releases the light allocation
     */
    ~Light()
    {
	FUNC("Light::~Light()");
	Release();
	END();
    }

    /*!
     * \brief Gets a light allocation
     * 
     * This function iterates through the list of lights
     * looking for one which is currently unused, then nabs it.
     */
    bool Init()
    {
	FUNC("Light::Init()");
	for(int i=0; i < GL_MAX_LIGHTS; i++)
	    if(!light[i])
	    {
		Name = GL_LIGHT0+i;
		light[i] = true;
		ENDRETURN(true);
	    }

	ENDRETURN(false);
    }

    /*!
     * \brief Releases a light allocation
     * 
     * If we have a light allocated, we release it
     * so that another light can use it.
     */
    bool Release()
    {
	FUNC("Light::Release");
	if(Name)
	    light[Name-GL_LIGHT0] = false;
	Name = 0;
	ENDRETURN(true);
    }

    //! Enable the light
    bool Enable() { FUNC("Light::Enable");glEnable(Name);ENDRETURN(true); }

    //! Disable the light
    bool Disable() { FUNC("Light::Disable");glDisable(Name);ENDRETURN(true); }

    /*!
     * \brief Sets the light position
     * 
     * \param pos        the light position
     * \param positional true if the light is a positional light
     */
    bool Position(Vectorf pos, bool positional);

    /*!
     * \brief Sets the light direction
     * 
     * \param dir the light direction
     */
    void Direction(Vectorf dir);

    /*!
     * \brief Sets the ambient light colour
     * 
     * \param amb the ambient light colour
     */
    void Ambient(Vectorf amb);

    /*!
     * \brief Sets the diffuse light colour
     * 
     * \param dif the diffuse light colour
     */
    void Diffuse(Vectorf dif);

    /*!
     * \brief Sets the specular light colour
     * 
     * \param spec the specular light colour
     */
    void Specular(Vectorf spec);

    /*!
     * \brief Sets the light attenuation parameters
     * 
     * \param constant  the constant attenuation parameter
     * \param linear    the linear attenuation parameter
     * \param quadratic the quadratic attenuation parameter
     */
    void Attenuation(float constant, float linear, float quadratic);
  private:
    //! whether the light array has been initialised
    static bool init;
    //! an array containing details of which lights are used
    static bool light[GL_MAX_LIGHTS];
    //! the ID of the current light. 0 if none is allocated
    GLenum Name;
};

#endif
