#ifndef Matrix_H
#define Matrix_H

#include "general.h"
#include "GLee.h"

/*!
 * \file   matrix.h
 * \author  <baldurk@baldurk.org>
 * \date   Fri Dec 30 21:46:59 2005
 *
 * \brief  Details of the Matrix class
 */

/*! 
 * \brief a templated Matrix class
 *
 * This class is a matrix templated to work for chars, floats.
 * The matrices are always 4x4. They are constructed initially to 
 * the identity matrix.
 */

template <class T> class Matrix
{
  public:
    //! Default constructor
    Matrix() { FUNC("Matrix::Matrix"); identity(); END(); }
    //! Initialise from an array
    Matrix(T mat[16])
    {
	FUNC("Matrix::Matrix(T mat[16])");
	m[0] = mat[0]; m[4] = mat[1]; m[8] =  mat[2];  m[12] = mat[3];
	m[1] = mat[4]; m[5] = mat[5]; m[9] =  mat[6];  m[13] = mat[7];
	m[2] = mat[8]; m[6] = mat[9]; m[10] = mat[10]; m[14] = mat[11];
	m[3] = mat[12]; m[7] = mat[13]; m[11] = mat[14]; m[15] = mat[15];
	END();
    }

    //! set the matrix to the identity matrix
    void identity()
    {
	FUNC("Matrix::identity");
	m[0] = 1; m[4] = 0; m[8] =  0; m[12] = 0;
	m[1] = 0; m[5] = 1; m[9] =  0; m[13] = 0;
	m[2] = 0; m[6] = 0; m[10] = 1; m[14] = 0;
	m[3] = 0; m[7] = 0; m[11] = 0; m[15] = 1;
	END();
    }
    //! cast to an array of floats.
    float* mat() { FUNC("Matrix::float*");ENDRETURN(m); }

    //! add a translation specified by the vector onto the matrix
    void translate(Vector <T> v)
    { FUNC("Matrix::translate(v)"); translate(v.x, v.y, v.z); END(); }
    //! add a rotation specified by the vector onto the matrix
    void rotate(Vector <T> v)
    { FUNC("Matrix::rotate(v)"); rotate(v.x, v.y, v.z); END(); }

    //! add a translation specified by the parameters onto the matrix
    void translate(float x, float y, float z)
    {
	FUNC("Matrix::translate(x,y,z)");
	Matrix mat;
	mat.m[12] = x;mat.m[13] += y;mat.m[14] += z;
	*this = *this * mat;
	END();
    }
    //! add a rotation specified by the parameters onto the matrix
    void rotate(float x, float y, float z)
    {
	FUNC("Matrix::rotate(x, y, z)");
	float cx, cy, cz, sx, sy, sz;

	cx = (float)cos(x);sx = (float)sin(x);
	cy = (float)cos(y);sy = (float)sin(y);
	cz = (float)cos(z);sz = (float)sin(z);

	Matrix mat;

	mat.m[0] = cy * cz;
	mat.m[1] = cy * sz;
	mat.m[2] = -sy;
    
	mat.m[4] = (sx * sy * cz) - (cx * sz);
	mat.m[5] = (sx * sy * sz) + (cx * cz);
	mat.m[6] = sx * cy;
    
	mat.m[8] = (cx * sy * cz) + (sx * sz);
	mat.m[9] = (cx * sy * sz) - (sx * cz);
	mat.m[10] = cx * cy;
	
	*this = *this * mat;
	END();
    }

    //! specify a rotation around a specific vector
    // thanks to the flipcode algo for explaining this
    Vector <T> rot(Vector <T> v)
    {
	FUNC("Matrix::rot(v)");
	Vector<T> angles;
	angles.y = asin(m[2]);

	float cosy = cos(angles.y);

	if(fabs(cosy) > 0.005f)     // helps with gimball lock
	{
	    angles.x  = atan2(-m[6]/cosy, m[10]/cosy);
	    angles.z  = atan2(-m[1]/cosy, m[0]/cosy);
	}
	else
	{
	    angles.x  = 0;
	    angles.z  = atan2(m[4], m[5]);
	}

	while(angles.x < 0.0f)
	    angles.x += PI*2.0f;
	while(angles.y < 0.0f)
	    angles.y += PI*2.0f;
	while(angles.z < 0.0f)
	    angles.z += PI*2.0f;

	while(angles.x > PI*2.0f)
	    angles.x -= PI*2.0f;
	while(angles.y > PI*2.0f)
	    angles.y -= PI*2.0f;
	while(angles.z > PI*2.0f)
	    angles.z -= PI*2.0f;
	
	Vector<T> temp(v), final;
    
	// z rotation
	
	final.x = (cos(angles.z) * temp.x) - (sin(angles.z) * temp.y);
	final.y = (sin(angles.z) * temp.x) + (cos(angles.z) * temp.y);
	final.z = temp.z;

	temp = final;

	// y rotation

	final.x = (cos(angles.y) * temp.x) + (sin(angles.y) * temp.z);
	final.y = temp.y;
	final.z = -(sin(angles.y) * temp.x) + (cos(angles.y) * temp.z);

	temp = final;
	
	// x rotation
	
	final.x = temp.x;
	final.y = (cos(angles.x) * temp.y) - (sin(angles.x) * temp.z);
	final.z = (sin(angles.x) * temp.y) + (cos(angles.x) * temp.z);
	
	temp = final;
	
	ENDRETURN(final);
    }

    //! Get the rotation angles out of the matrix
    Vector <T> GetAngles()
    {
	FUNC("Matrix::GetAngles");
	static Vector<T> angles;
	angles.y = asin(m[2]);

	float cosy = cos(angles.y);

	if(fabs(cosy) > 0.005f)     // helps with gimball lock
	{
	    angles.x  = atan2(-m[6]/cosy, m[10]/cosy);
	    angles.z  = atan2(-m[1]/cosy, m[0]/cosy);
	}
	else
	{
	    angles.x  = 0;
	    angles.z  = atan2(m[4], m[5]);
	}

	while(angles.x < 0.0f)
	    angles.x += PI*2.0f;
	while(angles.y < 0.0f)
	    angles.y += PI*2.0f;
	while(angles.z < 0.0f)
	    angles.z += PI*2.0f;

	while(angles.x > PI*2.0f)
	    angles.x -= PI*2.0f;
	while(angles.y > PI*2.0f)
	    angles.y -= PI*2.0f;
	while(angles.z > PI*2.0f)
	    angles.z -= PI*2.0f;

	ENDRETURN(angles);
    }

    /*!
      get the current GL modelview matrix into this matrix

      \todo fix this - GL specific code should not be in this class
    */
    Matrix GetModelView()
    {
	FUNC("Matrix::GetModelView");
	glGetFloatv(GL_MODELVIEW_MATRIX, m);
	ENDRETURN(*this);
    }

    /*! 
      multiply this matrix onto the current modelview matrix.

      \todo fix this - GL specific code should not be in this class
    */
    Matrix ApplyModelView()
    {
	FUNC("Matrix::ApplyModelView");
	glMultMatrixf(m);
	ENDRETURN(*this);
    }

    void operator = (float val[16])
    {
	FUNC("Matrix::=val[16]");
	for(int i=0; i < 16; i++)
	    m[i] = val[i];
	END();
    }

    void operator *= (float f)
    {
	FUNC("Matrix::*=f");
	for(int i=0; i < 16; i++)
	    m[i] *= f;
	END();
    }

    Matrix operator *= (Matrix matr)
    {
	FUNC("Matrix::*=m");
	m[0] = m[0]*matr.m[0] + m[4]*matr.m[1] + m[8]*matr.m[2];
	m[1] = m[1]*matr.m[0] + m[5]*matr.m[1] + m[9]*matr.m[2];
	m[2] = m[2]*matr.m[0] + m[6]*matr.m[1] + m[10]*matr.m[2];
	m[3] = m[3]*matr.m[0] + m[7]*matr.m[1] + m[11]*matr.m[2];

	m[4] = m[0]*matr.m[4] + m[4]*matr.m[5] + m[8]*matr.m[6];
	m[5] = m[1]*matr.m[4] + m[5]*matr.m[5] + m[9]*matr.m[6];
	m[6] = m[2]*matr.m[4] + m[6]*matr.m[5] + m[10]*matr.m[6];
	m[7] = m[3]*matr.m[4] + m[7]*matr.m[5] + m[11]*matr.m[6];

	m[8] = m[0]*matr.m[8] + m[4]*matr.m[9] + m[8]*matr.m[10];
	m[9] = m[1]*matr.m[8] + m[5]*matr.m[9] + m[9]*matr.m[10];
	m[10] = m[2]*matr.m[8] + m[6]*matr.m[9] + m[10]*matr.m[10];
	m[11] = m[3]*matr.m[8] + m[7]*matr.m[9] + m[11]*matr.m[10];

	m[12] = m[0]*matr.m[12] + m[4]*matr.m[13] + m[8]*matr.m[14];
	m[13] = m[1]*matr.m[12] + m[5]*matr.m[13] + m[9]*matr.m[14];
	m[14] = m[2]*matr.m[12] + m[6]*matr.m[13] + m[10]*matr.m[14];
	m[15] = m[3]*matr.m[12] + m[7]*matr.m[13] + m[11]*matr.m[14];

	ENDRETURN(*this);
    }

    float &operator [](int i)
    {
	FUNC("Matrix::[]");
	ENDRETURN(m[i]);
    }

    Matrix operator * (Matrix matr)
    {
	FUNC("Matrix::*m");
	Matrix ret;

	ret.m[0] = m[0]*matr.m[0] + m[4]*matr.m[1] + m[8]*matr.m[2];
	ret.m[1] = m[1]*matr.m[0] + m[5]*matr.m[1] + m[9]*matr.m[2];
	ret.m[2] = m[2]*matr.m[0] + m[6]*matr.m[1] + m[10]*matr.m[2];
	ret.m[3] = m[3]*matr.m[0] + m[7]*matr.m[1] + m[11]*matr.m[2];

	ret.m[4] = m[0]*matr.m[4] + m[4]*matr.m[5] + m[8]*matr.m[6];
	ret.m[5] = m[1]*matr.m[4] + m[5]*matr.m[5] + m[9]*matr.m[6];
	ret.m[6] = m[2]*matr.m[4] + m[6]*matr.m[5] + m[10]*matr.m[6];
	ret.m[7] = m[3]*matr.m[4] + m[7]*matr.m[5] + m[11]*matr.m[6];

	ret.m[8] = m[0]*matr.m[8] + m[4]*matr.m[9] + m[8]*matr.m[10];
	ret.m[9] = m[1]*matr.m[8] + m[5]*matr.m[9] + m[9]*matr.m[10];
	ret.m[10] = m[2]*matr.m[8] + m[6]*matr.m[9] + m[10]*matr.m[10];
	ret.m[11] = m[3]*matr.m[8] + m[7]*matr.m[9] + m[11]*matr.m[10];

	ret.m[12] = m[0]*matr.m[12] + m[4]*matr.m[13] + m[8]*matr.m[14] + m[12];
	ret.m[13] = m[1]*matr.m[12] + m[5]*matr.m[13] + m[9]*matr.m[14] + m[13];
	ret.m[14] = m[2]*matr.m[12] + m[6]*matr.m[13] + m[10]*matr.m[14] + m[14];
	ret.m[15] = m[3]*matr.m[12] + m[7]*matr.m[13] + m[11]*matr.m[14] + m[15];

	ENDRETURN(ret);
    }

    Vector <T> operator * (Vector <T> vec)
    {
	FUNC("Matrix::*v");
	float a, b, c;

	a = vec.x * m[0] + vec.y * m[4] + vec.z * m[8] + m[12];
	b = vec.x * m[1] + vec.y * m[5] + vec.z * m[9] + m[13];
	c = vec.x * m[2] + vec.y * m[6] + vec.z * m[10] + m[14];

	ENDRETURN(Vector<T>(a, b, c));
    }

    void Invert()
    {
	FUNC("Matrix::invert");
	float tmp;
#define SWAP(a, b) tmp = m[a]; m[a] = m[b]; m[b] = tmp;

	SWAP(4, 1);SWAP(9, 6);SWAP(14, 11);
	SWAP(2, 8);SWAP(7, 13);SWAP(3, 12);
#undef SWAP
	END();
    }

    Matrix operator += (Matrix matr)
    {
	FUNC("Matrix::+=m");
	m[0] += matr.m[0]; m[4] += matr.m[4]; m[8] +=  matr.m[8];  m[12] += matr.m[12];
	m[1] += matr.m[1]; m[5] += matr.m[5]; m[9] +=  matr.m[9];  m[13] += matr.m[13];
	m[2] += matr.m[2]; m[6] += matr.m[6]; m[10] += matr.m[10]; m[14] += matr.m[14];
	m[3] += matr.m[3]; m[7] += matr.m[7]; m[11] += matr.m[11]; m[15] += matr.m[15];

	ENDRETURN(*this);
    }

    Matrix operator -= (Matrix matr)
    {
	FUNC("Matrix::-=m");
	m[0] -= matr.m[0]; m[4] -= matr.m[4]; m[8] -=  matr.m[8];  m[12] -= matr.m[12];
	m[1] -= matr.m[1]; m[5] -= matr.m[5]; m[9] -=  matr.m[9];  m[13] -= matr.m[13];
	m[2] -= matr.m[2]; m[6] -= matr.m[6]; m[10] -= matr.m[10]; m[14] -= matr.m[14];
	m[3] -= matr.m[3]; m[7] -= matr.m[7]; m[11] -= matr.m[11]; m[15] -= matr.m[15];

	ENDRETURN(*this);
    }

    bool operator == (Matrix matr)
    {
	FUNC("Matrix::==m");
	if(m[0] == matr.m[0] && m[4] == matr.m[4] && m[8] ==  matr.m[8] &&  m[12] == matr.m[12] &&
	   m[1] == matr.m[1] && m[5] == matr.m[5] && m[9] ==  matr.m[9] &&  m[13] == matr.m[13] &&
	   m[2] == matr.m[2] && m[6] == matr.m[6] && m[10] == matr.m[10] && m[14] == matr.m[14] &&
	   m[3] == matr.m[3] && m[7] == matr.m[7] && m[11] == matr.m[11] && m[15] == matr.m[15])
	    ENDRETURN(true);

	ENDRETURN(false);
    }

    bool operator != (Matrix matr)
    {
	FUNC("Matrix::!=m");
	if(!(*this == matr))
	    ENDRETURN(true);
	
	ENDRETURN(false);
    }


    float m[16];
};

template <class T> ostream &operator <<(ostream &o, Matrix <T> m)
{
    FUNC("Matrix::<<");
    o << "(" << m[0] << " " << m[4] << " " << m[8] << " " << m[12] << ")" << endl;
    o << "(" << m[1] << " " << m[5] << " " << m[9] << " " << m[13] << ")" << endl;
    o << "(" << m[2] << " " << m[6] << " " << m[10] << " " << m[14] << ")" << endl;
    o << "(" << m[3] << " " << m[7] << " " << m[11] << " " << m[15] << ")" << endl;
    ENDRETURN(o);
}

typedef Matrix<float> Matrixf;

#endif
