#ifndef MEMORY_H
#define MEMORY_H

#include "general.h"

/*!
 * \file   memory.h
 * \author  <baldurk@baldurk.org>
 * \date   Sat Dec 31 12:04:51 2005
 * 
 * \brief  Definitions of Smart Pointers and Memory Managed Objects  
 */


/*! 
 * \brief A base class which all objects to be automanaged should derive from
 * 
 * This class is the base class for all objects that are going to be
 * dynamically allocated via smart pointers so that they are automatically
 * garbage collected and reference counted. Garbage is collected via the
 * static member of this class.
 */
class MMObject
{
  public:
    /*! 
     * A default constructor that initialises the reference count to 0
     * and adds this object to the list of objects.
     */
    MMObject()
    {
	FUNC("MMObject");

	assert(!shutdown);
	refcount = 0;
	objects.push_back(this);
	END();
    }
    /*! 
     * If this object still has references, and it is being deleted,
     * something is very wrong. Unless we are deleting all objects.
     * This should only happen if you manually delete an object.
     */
    virtual ~MMObject()
    {
	FUNC("~MMObject");
	assert(refcount==0 || shutdown);
	END();
    }
    //! A new Pointer is pointing to this object
    void AddRef()
    {
	FUNC("MMObject::AddRef");
	if(shutdown){END();return;}
	++refcount;
	END();
    }
    //! A Pointer has stopped pointing to this object
    void Release()
    {
	FUNC("MMObject::Release");
	if(shutdown){END();return;}
	--refcount;
	END();
    }

    /*! 
     * \brief This static method deletes and removes all unused objects
     *
     * This static method will iterate through the list of objects and
     * remove all those that have a reference count of 0 (also if the count
     * is negative, although if this has happened, something has gone wrong).
     */

    static void CollectGarbage()
    {
	FUNC("MMObject::CollectGarbage");
	for(list<MMObject *>::iterator it = objects.begin();
	    it != objects.end();)
	{
	    MMObject *ptr = (*it);
	    ++it;
	    if(ptr->refcount <= 0)
	    {
		objects.remove(ptr);
		delete ptr;
	    }
	}
	END();
    }

    /*! 
     * \brief This static method deletes and removes all objects, unused or no
     *
     * This static method will iterate through the list of objects and
     * remove them all. It should therefore only be used at the end of a program
     * when al objects have been deleted, except those with interdependencies,
     * which would never be deleted as they reference each other - this means
     * their reference count stays above 0. Only call this once all Pointers have
     * been destroyed, or their destructor would break.
     */

    static void CollectAll()
    {
	FUNC("MMObject::CollectAll");
	shutdown = true;
	for(list<MMObject *>::iterator it = objects.begin();
	    it != objects.end();)
	{
	    MMObject *ptr = (*it);
	    ++it;
	    objects.remove(ptr);
	    delete ptr;
	}
	END();
    }

  private:
    int refcount;		/*!< The number of Pointers using this object */
    static list<MMObject *> objects; /*!< A list of all objects being managed */
    static bool shutdown;	/*!<
				 * a failsafe to prevent people calling
				 * CollectAll before the end of the program */
};


/*! 
 * \brief This class takes care of pointing to a memory managed object
 *
 * This class isn't much different from a reimplementation of pointers, but
 * with the important difference that it implements reference counting on
 * memory managed objects, so that they can be properly tracked and garbage
 * collected. For this reason the class T must derive from MMObject, or implement
 * the same interface and functionality.
 *
 * \note This class also protects against NULL pointer accesses.
 */

template <class T> class Pointer
{
  public:
    //! A default constructor to initialise the object to NULL
    Pointer() { FUNC("Pointer");obj = NULL;END(); }
    //! A constructor to initialise the object to a plain pointer
    Pointer(T *o) { FUNC("Pointer::Pointer(T*)");obj = NULL;*this = o;END(); }
    //! The copy constructor
    Pointer(const Pointer<T> &o)
    { FUNC("Pointer::Pointer(P)");obj = NULL; *this = o;END(); }
    //! The destructor releases the object - decrementing its reference count
    ~Pointer() { FUNC("~Pointer"); if(obj) obj->Release(); END();}

    //! An operator to cast to a plain pointer
    operator T*() const { FUNC("Pointer::T*");ENDRETURN(obj); }

    //! The not operator
    bool operator !() const { FUNC("Pointer::!");ENDRETURN(!obj); }
    //! Equality with a Pointer
    bool operator ==(Pointer<T> o) const
    { FUNC("Pointer::==P"); ENDRETURN(obj == o.obj); }
    //! Equality with a plain pointer
    bool operator ==(T *o) const { FUNC("Pointer::==T*");ENDRETURN(obj == o); }

    T &operator *() const
    { FUNC("Pointer::*");assert(obj!=NULL);ENDRETURN(*obj); }
    T *operator ->() const
    { FUNC("Pointer::->");assert(obj!=NULL);ENDRETURN(obj); }

    //! Assignment to a plain pointer
    void operator =(T *o)
    {
	FUNC("Pointer::=T*");
	if(obj) obj->Release();
	obj = o;
	if(obj) obj->AddRef();
	END();
    }

    //! Assignment to a Pointer
    void operator =(Pointer<T> &o)
    {
	FUNC("Pointer::=P");

	if(obj) obj->Release();
	obj = o.obj;
	if(obj) obj->AddRef();

	END();
    }
  private:
    T *obj;			/*!< the object being managed */
};

#endif
