#ifndef MODEL_H
#define MODEL_H

#include "general.h"
#include "texture.h"

/*!
 * \file   model.h
 * \author  <baldurk@baldurk.org>
 * \date   Fri Jan  6 19:29:36 2006
 * 
 * \brief  The details of a model  
 */

//! A vertex
struct Vertex
{
  public:
    //! the position
    Vectorf pos;
    //! the normal
    Vectorf normal;
    //Vectorf diffuse;
    //! the u co-ordinate
    float u,
	//! the v co-ordinate
	v;
    //! the bone this vertex is attached to
    int boneID;
};

//! A triangle
struct Triangle
{
  public:
    //! vertex indices
    int vertices[3];
    //! normal indices
    int normals[3];
    //! the smooth group this triangle is in
    int smoothgroup;
    Vectorf diffuse;
    //! the face normal
    Vectorf normal;
    //! the material ID
    int MaterialID;
};

enum Rendermode { NONE=0, IMMEDIATE, SEQUENTIAL };

//! A material
struct Material
{
  public:
    //! ambient colour
    Vectorf ambient,
	//! diffuse colour
	diffuse,
	//! specular colour
	specular,
	//! emissive colour
	emissive;
    //! the level of specular
    float shininess,
	//! the alpha value
	alpha;
    //! the texture filename
    string texture;
    //! the texture details
    Texture tex;
};

//! a keyframe, consisting of a rotation/position vector at a time index
struct Keyframe
{
    Keyframe() { Keyframe(0.0f, Vectorf()); }
    Keyframe(float t, Vectorf v) { time = t; vec = v; }
    float time;
    Vectorf vec;
};

//! a bone
struct Bone
{
    //Bone() : first(false) {};bool first;
    //! the name of this bone
    string name;
    //! a pointer to this bone's parent
    Bone *parent;
    //! this bone's children
    vector<int> children;
    //! this bone's vertices
    vector<int> vertices;
    //! the current Matrix
    Matrixf currentmatrix;
    //! bone flags
    int flags;
    //! the bone's initial position
    Vectorf firstpos;
    //! the bone's initial translation
    Vectorf initpos,
	//! the bone's initial rotation
	initrot;
    //! the bone's current position
    Vectorf curpos,
	//! the bone's current rotation
	currot;
    //! position keyframes
    vector<Keyframe> pos,
	//! rotation keyframes
	rot;
    int ID;
};

//! A model made up of several meshes and their materials
class Model
{
  public:
    Model() : vbomode(NONE) {}
    virtual ~Model() {}
    
    static Model *New(string Filename);

    /*!
     * A pure virtual function that loads a model.
     * This is inherited by format-specific implementations.
     */
    virtual int Load(string Filename)
    {
	assert("undefined load" && 0);
	return 0;
    }

    //! Write the model out to disk
    //! \param f filename
    virtual int Write(string f)
    {
	assert("undefined write" && 0);
	return 0;
    }

    void Setup(int subdivide = 0);

    //! sets what the current frame is
    void SetFrame(float f);

    //! sets what the current frame is
    void SetFrame(int idx, float f);

    //! Recurse through bones, initialising or updating vertices
    void BoneRecurse(Bone *b, Matrixf m=Matrixf(), bool initbone=false);

    //! Set the rendering mode
    //! \param mode  the mode to use
    //! \param regen whether or not to regenerate the data
    int SetMode(Rendermode mode, bool flat = false, bool regen = false);

    //! Draw the model
    int Render(bool flat = false);

    //! Release the model's resources
    virtual int Release();

    /*!
     * translate the model by a set amount
     * 
     * \param x the amount to translate by in x
     * \param y the amount to translate by in y
     * \param z the amount to translate by in z
     */
    void translate(float x, float y, float z)
    { FUNC("Model::translate");trans += Vectorf(x, y, z);END(); }

    /*!
     * rotate the model by a set amount
     * 
     * \param x the amount to rotate by in x
     * \param y the amount to rotate by in y
     * \param z the amount to rotate by in z
     */
    void rotate(float x, float y, float z)
    { FUNC("Model::rotate");rot += Vectorf(x, y, z);END(); }

    int GenNormals(bool flat = false);

    //! reset the rotation and translation
    void identity()
    { FUNC("Model::identity");trans = Vectorf(0.0f, 0.0f, 0.0f); rot = Vectorf(0.0f, 0.0f, 0.0f);END(); }

    //! the bones in the model
    vector<Bone*> Bones;

    //! the bones in the model
    vector<Bone*> RootBones;

    //! the vertices in the model
    vector<Vertex> Vertices, OrigVertices;

    //! the triangles in the model
    vector<Triangle> Triangles;

    //! the materials in the model
    vector<Material> Materials;
  protected:
    //! Write the mesh out to a buffer. If the buffer is NULL, 
    //! only the size of what would be written is returned.
    //! \param c pointer to a buffer.
    int WriteBuf(unsigned char *c)
    {
	assert("undefined write" && 0);
	return 0;
    }

    //! Autoload a cube
    bool Cube();

    //! Autoload a teapot
    bool Teapot();

    /*!
     * get a line from the file, ignoring comment lines and blank lines
     * 
     * \param fp the file to read from
     */
    string GetLine(FILE *fp);

    //! the model's translation
    Vectorf trans,
	//! the model's rotation
	rot;

    int Flat;

    int animmax;
    GLuint vtxvbo, normvbo, texvbo, diffvbo;
    Rendermode vbomode;
};

#include "modeltypes.h"

#endif
