#include "model.h"

/*!
 * \file   ms3dmodel.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:55:07 2006
 * 
 * \brief  Specific implementation of a model for MS3D files
 */

int MS3DModel::Load(string Filename)
{
    FUNC("MS3DModel::Load");
    FILE *fp = fopen(Filename.c_str(), "r");

    if(!fp)
    {
	Log << "'" << Filename << "': Invalid model filename" << endl;;
	ENDRETURN(0);
    }

    string line;

    int NumMeshes = 0;

    line = GetLine(fp);
    sscanf(line.c_str(), "Frames: %d", &animmax);
    line = GetLine(fp); // Frame
    line = GetLine(fp);
    sscanf(line.c_str(), "Meshes: %d", &NumMeshes);

    int voffset = 0, noffset = 0;
    for(int i=0; i < NumMeshes; i++)
    {
	int NumVertices, NumNormals, NumTriangles, MaterialID;

	string line;
	line = GetLine(fp); // Mesh
	sscanf(line.c_str(), "%*s %*d %d", &MaterialID);
	line = GetLine(fp); // Num Vertices
	sscanf(line.c_str(), "%d", &NumVertices);

	for(int i=0; i < NumVertices; i++)
	{
	    Vertex v;

	    line = GetLine(fp); // Vertex i
	    sscanf(line.c_str(), "%*d %f %f %f %f %f %d",
		   &v.pos.x, &v.pos.y, &v.pos.z,
		   &v.u, &v.v, &v.boneID);

	    Vertices.push_back(v);
	}

	line = GetLine(fp); // Num Normals

	sscanf(line.c_str(), "%d", &NumNormals);

	vector<Vectorf> Normals;

	for(int i=0; i < NumNormals; i++)
	{
	    Vectorf norm;

	    line = GetLine(fp); // Normal i
	    sscanf(line.c_str(), "%f %f %f",
		   &norm.x, &norm.y, &norm.z);

	    Normals.push_back(norm);
	}

	line = GetLine(fp); // Num Faces
	sscanf(line.c_str(), "%d", &NumTriangles);

	for(int i=0; i < NumTriangles; i++)
	{
	    Triangle tri;

	    int a = 0, b = 0, c = 0;
	    line = GetLine(fp); // Normal i
	    sscanf(line.c_str(), "%*d %d %d %d %d %d %d %d",
		   &tri.vertices[0], &tri.vertices[1], &tri.vertices[2],
		   &a, &b, &c, &tri.smoothgroup);

	    tri.vertices[0] += voffset;
	    tri.vertices[1] += voffset;
	    tri.vertices[2] += voffset;

	    tri.normals[0] = a+noffset;
	    tri.normals[1] = b+noffset;
	    tri.normals[2] = c+noffset;

	    Vertices[tri.vertices[0]].normal = Normals[tri.normals[0]];
	    Vertices[tri.vertices[1]].normal = Normals[tri.normals[1]];
	    Vertices[tri.vertices[2]].normal = Normals[tri.normals[2]];

	    tri.MaterialID = MaterialID;

	    Triangles.push_back(tri);
	}

	voffset = Vertices.size();
	noffset = Normals.size();
    }

    int NumMaterials;

    line = GetLine(fp);
    sscanf(line.c_str(), "Materials: %d", &NumMaterials);

    Materials.resize(NumMaterials);

    char tmpbuf[512];

    for(int i=0; i < NumMaterials; i++)
    {
	line = GetLine(fp); // Name
	line = GetLine(fp);
	sscanf(line.c_str(), "%f %f %f %*f", &Materials[i].ambient.x, &Materials[i].ambient.y, &Materials[i].ambient.z);
	line = GetLine(fp);
	sscanf(line.c_str(), "%f %f %f %*f", &Materials[i].diffuse.x, &Materials[i].diffuse.y, &Materials[i].diffuse.z);
	line = GetLine(fp);
	sscanf(line.c_str(), "%f %f %f %*f", &Materials[i].specular.x, &Materials[i].specular.y, &Materials[i].specular.z);
	line = GetLine(fp);
	sscanf(line.c_str(), "%f %f %f %*f", &Materials[i].emissive.x, &Materials[i].emissive.y, &Materials[i].emissive.z);
	line = GetLine(fp);
	sscanf(line.c_str(), "%f", &Materials[i].shininess);
	line = GetLine(fp);
	sscanf(line.c_str(), "%f", &Materials[i].alpha);
	line = GetLine(fp);
	sscanf(line.c_str(), "\"%s", tmpbuf);
	char *c = strchr(tmpbuf, '"');
	(c != NULL ? *c = 0 : *c);
	line = GetLine(fp);
	Materials[i].texture = tmpbuf;
	if(Materials[i].texture != "")
	    Materials[i].tex.Load(Materials[i].texture.c_str());
    }

    int NumBones = 0;
    line = GetLine(fp);
    sscanf(line.c_str(), "Bones: %d", &NumBones);

    Bones.resize(NumBones);

    for(int i=0; i < NumBones; i++)
    {
	Bone *b = new Bone;
	line = GetLine(fp); // name
	b->name = line;
	line = GetLine(fp); // parent
	if(line == "\"\"")
	    b->parent = NULL;
	else
	    for(int j=0; j < i; j++)
		if(line == Bones[j]->name)
		    b->parent = Bones[j];

	line = GetLine(fp); // initial setup
	sscanf(line.c_str(), "%d %f %f %f %f %f %f",
	       &b->flags,
	       &b->initpos.x, &b->initpos.y, &b->initpos.z,
	       &b->initrot.x, &b->initrot.y, &b->initrot.z);
	int poss = 0, rots = 0;

	line = GetLine(fp); // num pos keyframes
	sscanf(line.c_str(), "%d", &poss);
	b->pos.resize(poss);
	for(int j=0; j < poss; j++)
	{
	    line = GetLine(fp); // keyframe details
	    sscanf(line.c_str(), "%f %f %f %f", &b->pos[j].time, 
		   &b->pos[j].vec.x, &b->pos[j].vec.y, &b->pos[j].vec.z);
	}

	line = GetLine(fp); // num rot keyframes
	sscanf(line.c_str(), "%d", &rots);
	b->rot.resize(rots);
	for(int j=0; j < rots; j++)
	{
	    line = GetLine(fp); // keyframe details
	    sscanf(line.c_str(), "%f %f %f %f", &b->rot[j].time, 
		   &b->rot[j].vec.x, &b->rot[j].vec.y, &b->rot[j].vec.z);
	}

	b->ID = i;

	Bones[i] = b;
	b->curpos = b->initpos;b->currot = b->initrot;
	if(b->parent)
	    b->parent->children.push_back(i);
	else
	    RootBones.push_back(b);

	for(size_t j=0; j < Vertices.size(); j++)
	    if(Vertices[j].boneID == i)
		b->vertices.push_back(j);
    }

    fclose(fp);
    ENDRETURN(1);
}

int MS3DModel::Write(string f)
{
    assert("undefined write" && 0);
    return 0;
}
