#include "model.h"

/*!
 * \file   bkmodel.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Sat Feb  4 16:48:12 2006
 * 
 * \brief  Specific implementation of a model for BKM files
 */

int BKMModel::Load(string f)
{
    FUNC("BKMModel::Load");

    FILE *fp = fopen(f.c_str(), "rb");

    char id[5];
    int ver;
    memset(id, 0, 5);
    fread(id, sizeof(char), 4, fp);
    fread(&ver, sizeof(int), 1, fp);

    if(strcmp(id, "bkmb") || ver != 2)
    {
	Log << "not a BKM Binary file, or not the correct version (" << ver << ")" << endl;
	ENDRETURN(1);
    }

    unsigned long srcsize = 0;
    fread(&srcsize, sizeof(int), 1, fp);

    unsigned long dstsize = 0;
    fread(&dstsize, sizeof(int), 1, fp);

    unsigned char *srcbuf = new unsigned char[srcsize];
    unsigned char *dstbuf = new unsigned char[dstsize];

    fread(srcbuf, sizeof(unsigned char), srcsize, fp);

    fclose(fp);

    Inflate(dstbuf, &dstsize, srcbuf, srcsize);

    unsigned char *c = dstbuf;

    int NumVertices = 0, NumNormals = 0, NumTriangles = 0;

    memcpy(&NumVertices, c, sizeof(int));
    c += sizeof(int);

    Vertices.resize(NumVertices);

    for(int i=0; i < NumVertices; i++)
    {
	memcpy(&Vertices[i].pos, c, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(&Vertices[i].normal, c, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(&Vertices[i].u, c, sizeof(float));
	c += sizeof(float);

	memcpy(&Vertices[i].v, c, sizeof(float));
	c += sizeof(float);

	memcpy(&Vertices[i].boneID, c, sizeof(int));
	c += sizeof(int);
    }

    memcpy(&NumNormals, c, sizeof(int));
    c += sizeof(int);

    Normals.resize(NumNormals);

    memcpy(&Normals, c, sizeof(Vectorf)*NumNormals);
    c += sizeof(Vectorf)*NumNormals;

    memcpy(&NumTriangles, c, sizeof(int));
    c += sizeof(int);

    Triangles.resize(NumTriangles);

    for(int i=0; i < NumTriangles; i++)
    {
	memcpy(&Triangles[i].vertices, c, sizeof(int)*3);
	c += sizeof(int)*3;

	memcpy(&Triangles[i].normals, c, sizeof(int)*3);
	c += sizeof(int)*3;

	memcpy(&Triangles[i].smoothgroup, c, sizeof(int));
	c += sizeof(int);

	memcpy(&Triangles[i].normal, c, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(&Triangles[i].MaterialID, c, sizeof(int));
	c += sizeof(int);
    }

    memcpy(&NumMaterials, c, sizeof(int));
    c += sizeof(int);

    Materials.resize(NumMaterials);

    for(int i=0; i < NumMaterials; i++)
    {
	memcpy(&Materials[i].ambient, c, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(&Materials[i].diffuse, c, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(&Materials[i].specular, c, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(&Materials[i].emissive, c, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(&Materials[i].shininess, c, sizeof(float));
	c += sizeof(float);

	int dummy;
	memcpy(&dummy, c, sizeof(int));
	c += sizeof(int);
	Materials[i].texture.resize(dummy);
	memcpy(&Materials[i].texture, c, sizeof(char)*dummy);
	c += sizeof(char)*dummy;
    }

    ENDRETURN(0);
}

int BKMModel::Write(string f)
{
    FUNC("Model::Write(string)");

    int size = WriteBuf(NULL);

    unsigned char *srcbuf = new unsigned char[size];

    WriteBuf(srcbuf);

    unsigned char *destbuf = new unsigned char[DeflateSize(size)];

    unsigned long destsize = DeflateSize(size);

    Deflate(destbuf, &destsize, srcbuf, size);

    const char id[] = "bkmb";
    const int ver = 2;

    FILE *fp = fopen(f.c_str(), "wb");

    fwrite(id, sizeof(char), 4, fp);
    fwrite(&ver, sizeof(int), 1, fp);

    fwrite(&destsize, sizeof(int), 1, fp);
    fwrite(&size, sizeof(int), 1, fp);

    fwrite(destbuf, sizeof(unsigned char), destsize, fp);

    fclose(fp);

    ENDRETURN(0);
}

int BKMModel::WriteBuf(unsigned char *c)
{
    FUNC("Model::Write(char*)");
    assert(0);
    unsigned char *orig = c;


    int NumVertices = 0, NumNormals = 0, NumTriangles = 0;
    NumVertices = Vertices.size();
    NumNormals = Normals.size();
    NumTriangles = Triangles.size();

    memcpy(c, &NumVertices, sizeof(int));
    c += sizeof(int);

    for(int i=0; i < NumVertices; i++)
    {
	memcpy(c, &Vertices[i].pos, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(c, &Vertices[i].normal, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(c, &Vertices[i].u, sizeof(float));
	c += sizeof(float);

	memcpy(c, &Vertices[i].v, sizeof(float));
	c += sizeof(float);

	memcpy(c, &Vertices[i].boneID, sizeof(int));
	c += sizeof(int);
    }

    memcpy(c, &NumNormals, sizeof(int));
    c += sizeof(int);

    memcpy(c, &Normals, sizeof(Vectorf)*NumNormals);
    c += sizeof(Vectorf)*NumNormals;

    memcpy(c, &NumTriangles, sizeof(int));
    c += sizeof(int);

    for(int i=0; i < NumTriangles; i++)
    {
	memcpy(c, &Triangles[i].vertices, sizeof(int)*3);
	c += sizeof(int)*3;

	memcpy(c, &Triangles[i].normals, sizeof(int)*3);
	c += sizeof(int)*3;

	memcpy(c, &Triangles[i].smoothgroup, sizeof(int));
	c += sizeof(int);

	memcpy(c, &Triangles[i].normal, sizeof(Vectorf));
	c += sizeof(Vectorf);

	memcpy(c, &Triangles[i].MaterialID, sizeof(int));
	c += sizeof(int);
    }

    if(orig != NULL)
	memcpy(c, &NumMaterials, sizeof(int));
    c += sizeof(int);

    for(int i=0; i < NumMaterials; i++)
    {
	if(orig != NULL)
	    memcpy(c, &Materials[i].ambient, sizeof(Vectorf));
	c += sizeof(Vectorf);

	if(orig != NULL)
	    memcpy(c, &Materials[i].diffuse, sizeof(Vectorf));
	c += sizeof(Vectorf);

	if(orig != NULL)
	    memcpy(c, &Materials[i].specular, sizeof(Vectorf));
	c += sizeof(Vectorf);

	if(orig != NULL)
	    memcpy(c, &Materials[i].emissive, sizeof(Vectorf));
	c += sizeof(Vectorf);

	if(orig != NULL)
	    memcpy(c, &Materials[i].shininess, sizeof(float));
	c += sizeof(float);

	int dummy = Materials[i].texture.size();
	if(orig != NULL)
	    memcpy(c, &dummy, sizeof(int));
	c += sizeof(int);

	if(orig != NULL)
	    memcpy(c, &Materials[i].texture,
		   sizeof(char)*Materials[i].texture.size());
	c += sizeof(char)*Materials[i].texture.size();
    }

    ENDRETURN((int)(long)(c));
}
