#include "script.h"

/*!
 * \file   script.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:56:00 2006
 * 
 * \brief  Functions for dealing with scripted functions - in bytecode form
 */

ostream &operator <<(ostream &o, ScriptVariable var)
{
    FUNC("<< ScriptVariable");
    switch(var.type)
    {
	case VAR_INT:
	    ENDRETURN(o << var.i);
	case VAR_STRING:
	    ENDRETURN(o << var.s);
	case VAR_FLOAT:
	    ENDRETURN(o << var.f);
	case VAR_ARRAY:
	    o << "Array {";
	    for(int i=0; i < (signed)var.v.size(); i++)
		o << " [" << i << "]=" << var.v[i];
	    o << " }";
	    ENDRETURN(o);
	case VAR_NULL:
	    ENDRETURN(o << "NULL");
	default:
	    assert(0 && "unhandled type");
    }
    ENDRETURN(o);
}

#define TrimFirst(string) string = string.substr(1)
#define TrimLast(string) string = string.substr(0,string.length()-1)

ScriptVariable *Script::GetVar(string varname)
{
    FUNC("Script::GetVar");
    // If the string doesn't start with a variable, bail out
    if(StartVariable(varname))
    { 
	// grab the string into a working string
	string work = varname;
	// this will point at the current variable, initially NULL
	ScriptVariable *v = NULL;
	bool found = false;
	// while we don't have a blank string
	while(!IsSpace(work))
	{
	    // if we have a name of a variable - a-z, :, - and _
	    if(StartVariable(work))
	    {
		v = &vars[GetVariableName(work)];
		// we've got a variable
		found = true;
	    }
	    // if we get an array lookup
	    else if(Start(work, "["))
	    {
		// if we don't already have a variable name, get out and return NULL
		if(!found)
		    break;
		// trim off the [
		TrimFirst(work);
		// trim whitespace
		GetSpace(work);
		// if we don't have a number referencing the array, get out
		if(!StartNumber(work))
		    break;
		else
		{
		    // get the number
		    string num = GetNumber(work);
		    istringstream iss;int i; iss.str(num); iss >> i;
		    // reference into the current variable, and get the address
		    // of the array element
		    v = &((*v)[i]);
		}
		// trim off any remaining whitespace and the closing ]
		while(StartSpace(work) || Start(work, "]"))
		    TrimFirst(work);
	    }
	    else
	    {
		// we haven't found anything
		v = NULL;
		ENDRETURN(v);
	    }
	}
	if(found)
	    ENDRETURN(v);
    }
    ScriptVariable *v = NULL;
    ENDRETURN(v);
}

ScriptVariable Script::EvalExp(vector<token> tokens)
{
    FUNC("Script::EvalExp");

    // pop off NOPs from the end of the tokens
    while(tokens.size() && tokens[tokens.size()-1].type == NOP)
	tokens.pop_back();

    // parse block open/close tokens
    while(tokens.size() && 
	  (tokens[0].type == BLOCKOPEN || tokens[0].type == BLOCKCLOSE))
    {
	if(tokens.size())
	{
	    if(tokens[0].type == BLOCKOPEN)
		blocklevel++;
	    if(tokens[0].type == BLOCKCLOSE)
		blocklevel--;

	    for(size_t q=0; q < tokens.size()-1; q++)
		tokens[q] = tokens[q+1];
	    tokens.pop_back();
	}
    }

    // if the only tokens were NOPs or block open/close, size will be 0 or
    // -1, which means a pop_back() emptied the vector. That means the value
    // of this expression is NULL
    if(tokens.size() == 0 || tokens.size() == (size_t)-1)
	ENDRETURN(ScriptVariable());

    // if we only have one token to deal with
    if(tokens.size() == 1)
    {
	// if it is a value or string, we can just grab the token and return it
	if(tokens[0].type == VALUE || tokens[0].type == STRING)
	    ENDRETURN(ScriptVariable(tokens[0].var));
	// if it is a variable name, get a pointer to the variable and return what
	// it points to
	if(tokens[0].type == VARNAME)
	    ENDRETURN(*GetVar(tokens[0].name));
	// make a function call
	if(tokens[0].type == FUNCCALL)
	{
	    vector<token> ts;
	    vector<ScriptVariable> vs;
	    // for all the parameters passed, evaluate each and put it into
	    // a vector of values
	    for(size_t i=0; i < tokens[0].parameters.size(); i++)
	    {
		if(tokens[0].parameters[i].type == NOP)
		{
		    if(ts.size())
			vs.push_back(EvalExp(ts));
		    ts.clear();
		}
		else
		    ts.push_back(tokens[0].parameters[i]);
	    }

	    // copy the variable vector to a POD array
	    ScriptVariable *params = new ScriptVariable[vs.size()];

	    for(size_t i=0; i < vs.size(); i++)
		params[i] = vs[i];

	    // if there exists a compiled function with this name, give it precedence
	    if(compilefuncs[tokens[0].name])
	    {
		ENDRETURN(compilefuncs[tokens[0].name](tokens[0].name,
						       vs.size(), params));
	    }
	    // if there isn't a compiled function, try scripted functions
	    else if(scriptfuncs[tokens[0].name].tokens.size() > 0)
	    {
		// for each parameter the function expects.
		// this will stop at the earliest point. When no more parameters
		// have been passed, or when no more are expected.
		for(size_t q=0; scriptfuncs[tokens[0].name].parameters.size() &&
			q < scriptfuncs[tokens[0].name].parameters.size()-1 && q < vs.size(); q+=2)
		{
		    // if the parameter isn't a variable name, error out
		    if(scriptfuncs[tokens[0].name].parameters[q].type != VARNAME)
		    {
			Log << "token " << q << " is not a variable name" << endl;
			break;
		    }
		    
		    // store the current value of this variable in a global reference
		    // eg.
		    // func foo(bar);
		    // the value of bar is stored in ::bar before the call to 
		    // foo(4), then the value of the parameter, eg. 4, is stored
		    // in bar for the duration of the call
		    vars[string("::") +  scriptfuncs[tokens[0].name].parameters[q].name] =
			vars[scriptfuncs[tokens[0].name].parameters[q].name];
		    vars[scriptfuncs[tokens[0].name].parameters[q].name] = vs[q];
		}

		// evaluate the function tokens
		ScriptVariable ret = Eval(scriptfuncs[tokens[0].name].tokens);

		// for each parameter
		for(size_t q=0; scriptfuncs[tokens[0].name].parameters.size() &&
			q < scriptfuncs[tokens[0].name].parameters.size()-1 && q < vs.size(); q+=2)
		{
		    if(scriptfuncs[tokens[0].name].parameters[q].type != VARNAME)
		    {
			Log << "token " << q << " is not a variable name" << endl;
			break;
		    }

		    // restore the global value
		    vars[scriptfuncs[tokens[0].name].parameters[q].name] =
			vars[string("::") + scriptfuncs[tokens[0].name].parameters[q].name];
		}
		ENDRETURN(ret);
	    }
	    // no matching function found
	    else
	    {
		Log << "cannot find reference to " << tokens[0].name << endl;
		ENDRETURN(ScriptVariable());
	    }
	}
	// if it's a function definition, return NULL
	if(tokens[0].type == FUNCDEF)
	{
	    ENDRETURN(ScriptVariable());
	}
    }

    // iterate through the tokens
    for(vector<token>::iterator i=tokens.begin(); i != tokens.end(); ++i)
    {
	// if we've got a left bracket
	if(i->type == LBRACKET)
	{
	    // extract the contents of the bracket
	    int bracketlevel = 0;
	    int cnt = 0;
	    vector<token>::iterator j;
	    for(j=i; j != tokens.end(); ++j,++cnt)
	    {
		if((*j).type == LBRACKET)
		    bracketlevel++;
		if((*j).type == RBRACKET)
		    bracketlevel--;

		if(bracketlevel == 0)
		    break;
	    }

	    vector<token> ts(cnt);
	    // copy it to a separate vector
	    copy(i+1, j, ts.begin());
	    // recurse
	    ScriptVariable ret = EvalExp(ts);

	    // get rid of the bracketed expression
	    tokens.erase(i, j+1);

	    // insert the value of the returned value
	    if(ret.type == VAR_STRING)
		tokens.insert(i, token(STRING, ret));
	    else
		tokens.insert(i, token(VALUE, ret));
	}
    }

    // for all operations, in order of precedence
    for(int o=FIRSTOP; o < LASTOPNUM; o++)
    {
	// look through the tokens, rightmost first
	for(int i=tokens.size()-1; i >= 0; --i)
	{
	    if(tokens[i].type == o)
	    {
		// get the operations. Note, we might not have them so
		// put in a NOP if not.
		vector<token> first, snd;
		if(i)
		    first.push_back(tokens[i-1]);
		else
		    first.push_back(token(NOP));

		if(i+1 < (signed)tokens.size())
		    snd.push_back(tokens[i+1]);
		else
		    snd.push_back(token(NOP));

		// evaluate the tokens
		ScriptVariable a = EvalExp(first);
		ScriptVariable b = EvalExp(snd);

		// these variables should not be written to as they
		// are static.
		a.ro = b.ro = true;

		ScriptVariable ret;

		// if either token is a variable name we pass in a pointer
		// to it, otherwise we pass a pointer to the local (readonly)
		// variable.

		if(first[0].type == VARNAME && snd[0].type == VARNAME)
		    ret = PerformOp(GetVar(first[0].name),
				    toktype(o),
				    GetVar(snd[0].name));
		else if(first[0].type == VARNAME)
		    ret = PerformOp(GetVar(first[0].name),
				    toktype(o),
				    &b);
		else if(snd[0].type == VARNAME)
		    ret = PerformOp(&a,
				    toktype(o),
				    GetVar(snd[0].name));
		else
		    ret = PerformOp(&a, toktype(o), &b);

		int ref = i-1;
		if(ref == -1)
		    ref++;

		// replace the operation with the result

		for(size_t idx=ref; idx < tokens.size()-2; idx++)
		    tokens[idx] = tokens[idx+2];

		if(ret.type == VAR_STRING)
		    tokens[ref] = token(STRING, ret);
		else
		    tokens[ref] = token(VALUE, ret);

		for(size_t idx=0; idx < 2; idx++)
		    tokens.pop_back();
	    }
	}
    }

    // after any simplification, recurse. This will hit a base case when
    // only one token remains
    ENDRETURN(EvalExp(tokens));
}

// perform an operation. This only modifies the variables pointed to
// in the case they need to be - eg. increment, decrement or assignment.
// it's just a giant switch statement to perform the operations
ScriptVariable Script::PerformOp(ScriptVariable *a, toktype o, ScriptVariable *b)
{
    FUNC("Script::PerformOp");
    ScriptVariable ret;
    switch(o)
    {
	case INC:
	    if(*a == ScriptVariable())
	    {
		*b = *b + ScriptVariable(1);
		ret = *b;
	    }
	    if(*b == ScriptVariable())
	    {
		*a = *a + ScriptVariable(1);
		ret = *a;
	    }
	    break;
	case DEC:
	    if(*a == ScriptVariable())
	    {
		*b = *b - ScriptVariable(1);
		ret = *b;
	    }
	    if(*b == ScriptVariable())
	    {
		*a = *a - ScriptVariable(1);
		ret = *a;
	    }
	    break;
	case POW:
	    ret = *a ^ *b;
	    break;
	case MUL:
	    ret = *a * *b;
	    break;
	case DIV:
	    ret = *a / *b;
	    break;
	case MOD:
	    ret = *a % *b;
	    break;
	case ADD:
	    ret = *a + *b;
	    break;
	case SUB:
	    ret = *a - *b;
	    break;
	case NOT:
	    ret = !*b;
	    break;
	case AND:
	    ret = *a && *b;
	    break;
	case OR:
	    ret = *a || *b;
	    break;
	case LE:
	    ret = *a <= *b;
	    break;
	case LT:
	    ret = *a < *b;
	    break;
	case GE:
	    ret = *a >= *b;
	    break;
	case GT:
	    ret = *a > *b;
	    break;
	case EQ:
	    ret = *a == *b;
	    break;
	case NE:
	    ret = *a != *b;
	    break;
	case ASSIGN:
	    ret = *a = *b;
	    break;
	default:
	    break;
    }
    ENDRETURN(ret);
}

// Evaluate a series of tokens
ScriptVariable Script::Eval(vector<token> tokens)
{
    FUNC("Script::Eval");

    vector<token> ts(2);
    vector<token>::iterator prev=tokens.begin();
    int cnt=0;
    vector<token> init, test, inc;
    ScriptVariable ret;
    // for all the tokens
    for(vector<token>::iterator i=tokens.begin(); i != tokens.end();
	++i,++cnt)
    {
	// if we hit a return token, bail out
	if(i->type == RET)
	{
	    ret = (i+1)->var;
	    ENDRETURN(ret);
	}

	// if we hit an end of line token, we evaluate the tokens up until now
	if(i->type == EOL)
	{
	    ts.clear();
	    ts.resize(cnt);
	    // copy the currently read over tokens into a temporary array
	    copy(prev, i, ts.begin());

	    // if we have more than one token, or the first token isn't a
	    // execution control type statement, simply evaluate the tokens
	    // and store the return value
	    if(ts.size() != 1 ||
	       (ts[0].type != FOR &&
		ts[0].type != IF &&
		ts[0].type != WHILE))
	    {
		ScriptVariable r = EvalExp(ts);
		if(r != ScriptVariable())
		    ret = r;
	    }
	    // if we have more than one token, but the first IS an execution
	    // control statement, something has gone wrong
	    else if(ts.size() != 1)
	    {
		Log << "error parsing after while/for loop/if" << endl;
		ENDRETURN(ScriptVariable());
	    }
	    // check an if statement
	    else if(ts[0].type == IF)
	    {
		// evaluate the check, and if it is false
		if(EvalExp(ts[0].parameters) != ScriptVariable(1))
		{
		    // skip a block if there is one
		    if((i+1)->type == BLOCKOPEN)
		    {
			i += 2;
			int localblock = 1;
			while(localblock > 0)
			{
			    if(i->type == BLOCKOPEN)
				localblock++;
			    if(i->type == BLOCKCLOSE)
				localblock--;

			    i++;
			}			
		    }
		    // otherwise skip a single statement
		    else
			i++;
		}
	    }
	    // if we have a while statement
	    else if(ts[0].type == WHILE)
	    {
		vector<token>::iterator j = i+1;

		int num = 1;
		
		// grab the block of code (possibly a single statement)
		// into a vector
		if(j->type == BLOCKOPEN)
		{
		    j++;num++;
		    int localblock = 1;
		    while(localblock > 0)
		    {
			if(j->type == BLOCKOPEN)
			    localblock++;
			if(j->type == BLOCKCLOSE)
			    localblock--;
			
			j++;num++;
		    }
		}
		else
		{
		    j++;num++;
		}

		vector<token> block(num);
		copy(i, j, block.begin());

		// while the test is true, execute the code
		while(EvalExp(ts[0].parameters) == ScriptVariable(1))
		{
		    ScriptVariable r;
		    r = Eval(block);
		    if(r != ScriptVariable())
			ret = r;
		}
		i = j;
	    }
	    // do a for loop
	    else if(ts[0].type == FOR)
	    {
		vector<token>::iterator j = i+1;

		int num = 1;

		// grab the block of code
		if(j->type == BLOCKOPEN)
		{
		    j++;num++;
		    int localblock = 1;
		    while(localblock > 0)
		    {
			if(j->type == BLOCKOPEN)
			    localblock++;
			if(j->type == BLOCKCLOSE)
			    localblock--;

			j++;num++;
		    }
		}
		else
		{
		    j++;num++;
		}

		vector<token> block(num);
		copy(i, j, block.begin());

		// grab the init, test and inc statements by splitting
		// the parameters by EOLs.
		int exp = 0;
		init.clear();test.clear();inc.clear();
		for(size_t q=0; q < ts[0].parameters.size() && exp < 3; q++)
		{
		    if(ts[0].parameters[q].type == EOL)
		    {
			exp++;
			continue;
		    }
		    if(exp == 0)
			init.push_back(ts[0].parameters[q]);
		    if(exp == 1)
			test.push_back(ts[0].parameters[q]);
		    if(exp == 2)
			inc.push_back(ts[0].parameters[q]);
		}

		// evaluate the init statement
		EvalExp(init);

		// while the test is true, evaluate the block then the inc statement
		while(EvalExp(test) == ScriptVariable(1))
		{
		    ScriptVariable r;
		    r = Eval(block);
		    if(r != ScriptVariable())
			ret = r;
		    EvalExp(inc);
		}
		i = j;
	    }

	    prev = i+1;
	    cnt = -1;
	}
    }
    ENDRETURN(ret);
}

// compile a text file then evaluate the commands outside functions in it
bool Script::EvalTextFile(string file)
{
    FUNC("Script::EvalTextFile");

    ifstream ifs;
    ifs.open(file.c_str());

    string data;
    getline(ifs, data, '\0');

    vector<token> toks;
    Compile(toks, data);

    Eval(toks);

    ifs.close();

    ENDRETURN(true);
}

// compile a text file
bool Script::CompileTextFile(string file)
{
    FUNC("Script::CompileTextFile");

    ifstream ifs;
    ifs.open(file.c_str());

    string data;
    getline(ifs, data, '\0');

    vector<token> toks;
    Compile(toks, data);
    ifs.close();

#undef TrimFirst
#undef TrimLast

    ENDRETURN(true);
}

// a constructor to set default variable values
Script::Script()
{
    FUNC("Script");
    vars["false"] = 0;
    vars["true"] = 1;
    vars["pi"] = M_PI;
    compilefuncs["print"] = &Script::print;

    func = NULL;
    END();
}
