#ifndef SCRIPT_H
#define SCRIPT_H

#include "general.h"
#include "debug.h"

/*!
 * \file   script.h
 * \author  <baldurk@baldurk.org>
 * \date   Wed Feb  1 16:58:37 2006
 * 
 * \brief  A scripting system  
 */

//! An enumeration for the different types a variable could be
enum Type {VAR_NULL=-1, //!< NULL variable
	   VAR_INT=0, //!< integer variable
	   VAR_STRING, //!< string variable
	   VAR_FLOAT, //!< floating point variable
	   VAR_ARRAY //!< array variable
};

//! A structure to govern a variable in the script
struct ScriptVariable
{
    //! a constructor - sets the variable to NULL.
    ScriptVariable() : ro(false)
    { FUNC("ScriptVariable::ScriptVariable");i = 0; type = VAR_NULL;END(); }
    //! a constructor - sets the variable to an integer
    ScriptVariable(int x) : ro(false)
    { FUNC("ScriptVariable::ScriptVariable(int)");i = x; type = VAR_INT;END(); }
    //! a constructor - sets the variable to a float
    ScriptVariable(float x) : ro(false)
    { FUNC("ScriptVariable::ScriptVariable(float)");f = x; type = VAR_FLOAT;END(); }
    //! a constructor - sets the variable to a std::string
    ScriptVariable(string x) : ro(false)
    { FUNC("ScriptVariable::ScriptVariable(string)");s = x; type = VAR_STRING;END(); }
    //! a constructor - sets the variable to a C-style string
    ScriptVariable(const char *x) : ro(false)
    { FUNC("ScriptVariable::ScriptVariable(const char*)");s = x; type = VAR_STRING;END(); }
    //! a constructor - sets the variable to an array
    ScriptVariable(vector<ScriptVariable> x) : ro(false)
    { FUNC("ScriptVariable::ScriptVariable(vector<ScriptVariable>)");v = x; type = VAR_ARRAY;END(); }

    //! what type this variable is
    Type type;

    //! a union of the two basic types a variable could be
    union
    {
	//! the integer data in the variable
	int i;
	//! the float data in the variable
	float f;
    };
    //! the string data in the variable
    string s;
    //! the array data in the variable
    vector<ScriptVariable> v;
    //! read-only. If this is true, any assignments will fail an assert
    bool ro;

    //! raise this variable to the power of another.
    //! If either variable is a string or an array, this returns NULL
    ScriptVariable operator ^(ScriptVariable v)
    {
	FUNC("ScriptVariable^");
	ScriptVariable var;
	if(type == VAR_INT && v.type == VAR_INT)
	    var = (int)powf((float)i, (float)v.i);
	else if(type == VAR_INT && v.type == VAR_FLOAT)
	    var = powf((float)i, v.f);
	else if(type == VAR_FLOAT && v.type == VAR_INT)
	    var = powf(f, (float)v.i);
	else if(type == VAR_FLOAT && v.type == VAR_FLOAT)
	    var = powf(f, v.f);
	else
	    var = ScriptVariable();
	ENDRETURN(var);
    }

    //! return this variable multiplied by another.
    //! If either variable is a string or an array, this returns NULL
    ScriptVariable operator *(ScriptVariable v)
    {
	FUNC("ScriptVariable*");
	ScriptVariable var;
	if(type == VAR_INT && v.type == VAR_INT)
	    var = i * v.i;
	else if(type == VAR_INT && v.type == VAR_FLOAT)
	    var = float(i * v.f);
	else if(type == VAR_FLOAT && v.type == VAR_INT)
	    var = float(f * v.i);
	else if(type == VAR_FLOAT && v.type == VAR_FLOAT)
	    var = float(f * v.f);
	else
	    var = ScriptVariable();
	ENDRETURN(var);
    }

    //! return this variable divided by another.
    //! If either variable is a string or an array, this returns NULL
    ScriptVariable operator /(ScriptVariable v)
    {
	FUNC("ScriptVariable/");
	ScriptVariable var;
	if(type == VAR_INT && v.type == VAR_INT)
	    var = float(i) / float(v.i);
	else if(type == VAR_INT && v.type == VAR_FLOAT)
	    var = float(i) / v.f;
	else if(type == VAR_FLOAT && v.type == VAR_INT)
	    var = f / float(v.i);
	else if(type == VAR_FLOAT && v.type == VAR_FLOAT)
	    var = float(f / v.f);
	else
	    var = ScriptVariable();
	ENDRETURN(var);
    }

    //! return this variable modulo another.
    //! If either variable is a string or an array, this returns NULL
    ScriptVariable operator %(ScriptVariable v)
    {
	FUNC("ScriptVariable%");
	ScriptVariable var;
	if(type == VAR_INT && v.type == VAR_INT)
	    var = i % v.i;
	else if(type == VAR_INT && v.type == VAR_FLOAT)
	    var = i % (int)v.f;
	else if(type == VAR_FLOAT && v.type == VAR_INT)
	    var = (int)f % v.i;
	else if(type == VAR_FLOAT && v.type == VAR_FLOAT)
	    var = (int)f % int(v.f);
	else
	    var = ScriptVariable();
	ENDRETURN(var);
    }

    //! return this variable plus another.
    //! If either variable is an array, this returns NULL.
    //! If either variable is a string, the other is promoted
    //! to a string, and the two are concatenated
    ScriptVariable operator +(ScriptVariable v)
    {
	FUNC("ScriptVariable+");
	ScriptVariable var;
	if(type == VAR_INT && v.type == VAR_INT)
	    var = i + v.i;
	else if(type == VAR_INT && v.type == VAR_FLOAT)
	    var = float(i + v.f);
	else if(type == VAR_FLOAT && v.type == VAR_INT)
	    var = float(f + v.i);
	else if(type == VAR_FLOAT && v.type == VAR_FLOAT)
	    var = float(f + v.f);
	else if(type == VAR_STRING && v.type == VAR_STRING)
	    var = s + v.s;
	else
	{
	    ostringstream oss;
	    if(type == VAR_STRING && v.type == VAR_INT)
		oss << s << v.i;
	    else if(type == VAR_STRING && v.type == VAR_FLOAT)
		oss << s << v.f;
	    else if(type == VAR_INT && v.type == VAR_STRING)
		oss << i << v.s;
	    else if(type == VAR_FLOAT && v.type == VAR_STRING)
		oss << f << v.s;
	    else if(type == VAR_NULL && v.type == VAR_STRING)
		oss << "NULL" << v.s;
	    else if(type == VAR_STRING && v.type == VAR_NULL)
		oss << s << "NULL";

	    if(type != VAR_STRING && v.type != VAR_STRING)
		var = ScriptVariable();
	    else
		var = ScriptVariable(oss.str());
	}
	ENDRETURN(var);
    }

    //! return this variable minus another.
    //! If either variable is a string or an array, this returns NULL
    ScriptVariable operator -(ScriptVariable v)
    {
	FUNC("ScriptVariable-");
	ScriptVariable var;
	if(type == VAR_INT && v.type == VAR_INT)
	    var = i - v.i;
	else if(type == VAR_INT && v.type == VAR_FLOAT)
	    var = float(i - v.f);
	else if(type == VAR_FLOAT && v.type == VAR_INT)
	    var = float(f - v.i);
	else if(type == VAR_FLOAT && v.type == VAR_FLOAT)
	    var = float(f - v.f);
	else
	    var = ScriptVariable();
	ENDRETURN(var);
    }

    //! return this variable compared with another - less than.
    //! If either variable is a string or an array, this returns NULL
    ScriptVariable operator <(ScriptVariable v)
    {
	FUNC("ScriptVariable<");
	ScriptVariable var;
	if(type == VAR_INT && v.type == VAR_INT)
	    var = i < v.i;
	else if(type == VAR_INT && v.type == VAR_FLOAT)
	    var = float(i) < v.f;
	else if(type == VAR_FLOAT && v.type == VAR_INT)
	    var = f < float(v.i);
	else if(type == VAR_FLOAT && v.type == VAR_FLOAT)
	    var = f < v.f;
	else
	    var = ScriptVariable();
	ENDRETURN(var);
    }

    //! return this variable compared with another - more than.
    //! If either variable is a string or an array, this returns NULL
    bool operator >(ScriptVariable v)
    {
	FUNC("ScriptVariable>");
	bool var;
	if(type == VAR_INT && v.type == VAR_INT)
	    var = i > v.i;
	else if(type == VAR_INT && v.type == VAR_FLOAT)
	    var = float(i) > v.f;
	else if(type == VAR_FLOAT && v.type == VAR_INT)
	    var = f > float(v.i);
	else if(type == VAR_FLOAT && v.type == VAR_FLOAT)
	    var = f > v.f;
	else
	    var = false;
	ENDRETURN(var);
    }

    //! cast this variable to a bool. The only way for a string to be
    //! true is if it is "true". An array is never true, neither is NULL.
    operator bool()
    {
	FUNC("ScriptVariable::bool");
	if(type == VAR_INT)
	{
	    ENDRETURN(i == 0 ? false : true);
	}
	if(type == VAR_FLOAT)
	{
	    ENDRETURN(f == 0 ? false : true);
	}
	else if(type == VAR_STRING && s == "true")
	{
	    ENDRETURN(true);
	}

	ENDRETURN(false);
    }

    //! access a variable in an array, referenced by an array.
    //! If either variable is not an array, or the reference is not an int, this returns NULL
    ScriptVariable operator [](ScriptVariable vr)
    {
	FUNC("ScriptVariable[]");

	if(vr.type != VAR_INT)   
	{
	    ENDRETURN(ScriptVariable());
	}

	type=VAR_ARRAY;
	for(size_t x=v.size(); x <= (size_t)vr.i; x++)
	    v.push_back(ScriptVariable());

	ScriptVariable var;
	if(type == VAR_ARRAY && vr.type == VAR_INT)
	    var = v[vr.i];
	else
	    var = ScriptVariable();

	ENDRETURN(var);
    }

    //! return this variable compared with another - equals.
    //! If either variable is a string or an array, this returns NULL
    bool operator ==(ScriptVariable vr) const
    {
	if(type != vr.type)
	    return false;
	else if(type == VAR_INT)
	    return i == vr.i;
	else if(type == VAR_FLOAT)
	    return f == vr.f;
	else if(type == VAR_STRING)
	    return s == vr.s;
	else if(type == VAR_ARRAY)
	    return v == vr.v;
	else if(type == VAR_NULL)
	    return true;

	return false;
    }

    //! Assignment to an integer. Fails if the variable is read only
    //! This will set the type of the variable to match
    ScriptVariable &operator =(int o)
    { FUNC("ScriptVariable=int");assert(!ro);type=VAR_INT;i = o; ENDRETURN(*this); }
    //! Assignment to a C-style string. Fails if the variable is read only
    //! This will set the type of the variable to match
    ScriptVariable &operator =(const char *o)
    { FUNC("ScriptVariable=int");assert(!ro);type=VAR_STRING;s = o; ENDRETURN(*this); }
    //! Assignment to a std::string. Fails if the variable is read only
    //! This will set the type of the variable to match
    ScriptVariable &operator =(string o)
    { FUNC("ScriptVariable=int");assert(!ro);type=VAR_STRING;s = o; ENDRETURN(*this); }
    //! Assignment to a float. Fails if the variable is read only
    //! This will set the type of the variable to match
    ScriptVariable &operator =(float o)
    { FUNC("ScriptVariable=int");assert(!ro);type=VAR_FLOAT;f = o; ENDRETURN(*this); }
    //! Assignment to a double. Fails if the variable is read only
    //! This will set the type of the variable to match
    ScriptVariable &operator =(double o)
    { FUNC("ScriptVariable=int");assert(!ro);type=VAR_FLOAT;f = (float)o; ENDRETURN(*this); }

    //! access a variable in an array, referenced by an int.
    //! If either variable is not large enough, this will resize it.
    ScriptVariable &operator [](int i)
    {
	FUNC("ScriptVariable=int");
	type=VAR_ARRAY;
	for(size_t x=v.size(); x <= (size_t)i; x++)
	    v.push_back(ScriptVariable());

	ENDRETURN(v[i]);
    }
};

//! an overloaded operator to output a variable
ostream &operator <<(ostream &o, ScriptVariable var);

//! a macro to define a function.
#define SCRIPTFUNC(name) ScriptVariable name(string funcnum, int varnum, ScriptVariable *vars)

//! a macro to define a pointer to a function
#define ScriptFunction(name) ScriptVariable (*name)(string, int, ScriptVariable *)

//! the type of a bytecode token
enum toktype { EOL=-2, //!< End of line
	       NOP=-1, //!< No operation
	       LBRACKET=0, //!< Opening bracket. Used only in mathematical expressions
	       RBRACKET, //!< Closing bracket. Used only in mathematical expressions
	       INC, //!< Increment
	       DEC, //!< Decrement
	       POW, //!< Power
	       MUL, //!< Multiply
	       DIV, //!< Divide
	       MOD, //!< Modulo
	       ADD, //!< Add
	       SUB, //!< Subtract
	       NOT, //!< Not
	       AND, //!< And
	       OR, //!< Or
	       LE, //!< Less than or equal
	       LT, //!< Less than
	       GE, //!< Greater than or equal
	       GT, //!< Greater than
	       EQ, //!< Equal
	       NE, //!< Not equal
	       ASSIGN, //!< Assign
	       LASTOPNUM, //!< the last oepration number
	       BLOCKOPEN, //!< Open of a curly brace block
	       BLOCKCLOSE, //!< Close of a curly brace block
	       FUNCCALL, //!< Function call
	       VARNAME, //!< ScriptVariable name
	       STRING, //!< A literal string
	       VALUE, //!< A value
	       FUNCDEF, //!< A function definition
	       IF, //!< An if statement
	       FOR, //!< A for loop
	       WHILE, //!< A while loop
	       RET //!< A return statement
};

//! the first operation that performs an action
#define FIRSTOP INC
//! the number of tokens with a direct character translation
#define NUMTOKS LASTOPNUM

//! a lookup table to translate characters to tokens and vice-versa
const char toktrans[NUMTOKS][3] = {"(", ")",
				   "++", "--", "^", "*", "/", "%", "+", "-",
				   "!", "&&", "||", "<=", "<", ">=", ">", "==",
				   "!=", "="};

//! a bytecode token
struct token
{
    //! a constructor - sets the token to a NOP
    token()
	: type(NOP), var(), name("") { }
    //! a constructor to set the type, variable and the name
    token(toktype t, ScriptVariable v = ScriptVariable(), string n="")
	: type(t), var(v), name(n) {}

    //! the type of the token
    toktype type;
    //! the variable value of this token
    ScriptVariable var;
    //! the name of this token - used for variable names and function calls/definitions
    string name;

    //! the parameters of this token - used for function calls/definitions
    vector<token> parameters;
};

//! A scripting system
//! 
//! \todo validate input.
class Script
{
  public:
    //! a constructor to initialise the system
    Script();

    /*!
     *  compile a text file. This will not execute anything,
     * but it will store any function definitions
     * 
     * \param file the filename to compile
     */

    bool CompileTextFile(string file);

    /*!
     * Compile an expression into a vector of tokens
     *
     * \param ts   vector of tokens to insert the compiled bytecode into
     * \param expr the expression t ocompile
     */
    bool Compile(vector<token> &ts, string expr);

    /*!
     * Evaluate bytecode. This compiles it to bytecode then evaluates
     * the bytecode
     * 
     * \param expr the bytecode to evaluate
     */
    bool Eval(string expr)
    {
	FUNC("Script::EvalExpression");
	vector<token> toks;
	Compile(toks, expr);
	toks.push_back(EOL);
	Eval(toks);
	ENDRETURN(true);
    }

    /*!
     * Evaluate a text file. This compiles the file, then evaluates the results.
     * Any function defined will not be called. Only code outside the definitions
     * will be executed
     *
     * \param file the filename of the script to evaluate
     */
    bool EvalTextFile(string file);

    /*!
     * Adds a function to the list of functions that can be called by scripts
     *
     * \param name the name of the function
     * \param func the function pointer to the function
     */
    bool AddFunction(string name, ScriptFunction(func))
    {
	FUNC("Script::AddFunction");
	compilefuncs[name] = func;
	ENDRETURN(true);
    }
  private:
    //! the current level of indentation of curly braces
    int blocklevel;

    /*!
     * Evaluate a series of tokens
     *
     * \param tokens the tokens to evaluate
     */
    ScriptVariable Eval(vector<token> tokens);

    /*!
     * Evaluate a series of tokens that comprises one single expression
     *
     * \param tokens the tokens to evaluate
     */
    ScriptVariable EvalExp(vector<token> tokens);

    /*!
     * Perform an operation between two variables. Given by o
     *
     * \note if a or b should not be written to, they should not be written
     *       to, as they may not be writable.
     *
     * \param a   a pointer to the lvalue variable
     * \param o   an enumeration defining the operation to perform
     * \param b   a pointer to the rvalue variable
     */
    ScriptVariable PerformOp(ScriptVariable *a, toktype o, ScriptVariable *b);

    /*!
     * Get a reference to the storage of a variable of a given name.
     *
     * \param varname the name of the variable to look up
     */
    ScriptVariable *GetVar(string varname);

    /*!
     * returns whether or not the test string begins the str string
     * 
     * \param str   the haystack
     * \param test  the needle
     */
    bool Start(string str, string test);

    /*!
     * returns whether or not the test string is contained in the str string
     * 
     * \param str   the haystack
     * \param test  the needle
     */
    bool Contains(string str, string test);

    /*!
     * returns a number from the string
     * 
     * \param data  the data to retrieve a number from
     */
    string GetNumber(string &data);

    /*!
     * returns whether or not the test string begins with a number
     * 
     * \param data  the string to test
     */
    bool StartNumber(string data);

    /*!
     * returns whitespace from the string
     * 
     * \param data  the data to retrieve whitespace from
     */
    string GetSpace(string &data);

    /*!
     * returns whether or not the test string begins with whitespace
     * 
     * \param data  the string to test
     */
    bool StartSpace(string data);

    /*!
     * returns whether or not the test string is only whitespace
     * 
     * \param data  the string to test
     */
    bool IsSpace(string data);

    /*!
     * returns a variable from the string
     * 
     * \param data  the data to retrieve a variable name from
     */
    string GetVariable(string &data);

    /*!
     * returns a variable name from the string. ie. only blah from blah[0][5]
     * 
     * \param data  the data to retrieve a variable name from
     */
    string GetVariableName(string &data);

    /*!
     * returns whether or not the test string begins with a variable
     * 
     * \param data  the string to test
     */
    bool StartVariable(string data);

    /*!
     * returns a string from the string. This means "blah" - with quotes.
     * 
     * \param data  the data to retrieve a string from
     */
    string GetString(string &data);

    /*!
     * returns whether or not the test string begins with a string
     * 
     * \param data  the string to test
     */
    bool StartString(string data);

    //! a scripted function
    struct scrfunc
    {
	//! the parameters this function expects
	vector<token> parameters;
	//! the tokens in this function
	vector<token> tokens;
    };

    //! the current function being read into
    scrfunc *func;

    //! a map from function names to scripted functions
    map<string, scrfunc> scriptfuncs;

    //! a map from function names to variables
    map<string, ScriptVariable> vars;

    //! a map from function names to compiled functions
    map<string, ScriptFunction()> compilefuncs;

  public:
    //! A function to print out a series of variables
    static SCRIPTFUNC(print)
    {
	for(int i=0; i < varnum; i++)
	    Log << vars[i] << " ";
	Log << endl;
	return ScriptVariable(varnum);
    }
};

#endif
