#include "sdlinput.h"

/*!
 * \file   sdlinput.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:56:17 2006
 * 
 * \brief  Functions for dealing with input via SDL
 */


const int LSHIFT         = KMOD_LSHIFT;
const int RSHIFT         = KMOD_RSHIFT;
const int SHIFT          = (KMOD_LSHIFT|KMOD_RSHIFT);
const int LCONTROL       = KMOD_LCTRL;
const int RCONTROL       = KMOD_RCTRL;
const int CONTROL        = (KMOD_LCTRL|KMOD_RCTRL);
const int LALT           = KMOD_LALT;
const int RALT           = KMOD_RALT;
const int ALT            = (KMOD_LALT|KMOD_RALT);
const int CAPS           = KMOD_CAPS;
const int CAPITAL        = (CAPS|SHIFT);

const int BACKSPACE      = SDLK_BACKSPACE;
const int TAB            = SDLK_TAB;
const int CLEAR          = SDLK_CLEAR;
const int RETURN         = SDLK_RETURN;
const int PAUSE          = SDLK_PAUSE;
const int ESCAPE         = SDLK_ESCAPE;
const int SPACE          = SDLK_ESCAPE;

const int DELETEKEY      = SDLK_DELETE;

const int KP0            = SDLK_KP0;
const int KP1            = SDLK_KP1;
const int KP2            = SDLK_KP2;
const int KP3            = SDLK_KP3;
const int KP4            = SDLK_KP4;
const int KP5            = SDLK_KP5;
const int KP6            = SDLK_KP6;
const int KP7            = SDLK_KP7;
const int KP8            = SDLK_KP8;
const int KP9            = SDLK_KP9;
const int KP_PERIOD      = SDLK_KP_PERIOD;
const int KP_DIVIDE      = SDLK_KP_DIVIDE;
const int KP_MULTIPLY    = SDLK_KP_MULTIPLY;
const int KP_MINUS       = SDLK_KP_MINUS;
const int KP_PLUS        = SDLK_KP_PLUS;
const int KP_ENTER       = SDLK_KP_ENTER;
const int KP_EQUALS      = SDLK_KP_EQUALS;
const int UP             = SDLK_UP;
const int DOWN           = SDLK_DOWN;
const int RIGHT          = SDLK_RIGHT;
const int LEFT           = SDLK_LEFT;
const int HOME           = SDLK_HOME;
const int INSERT         = SDLK_INSERT;
const int ENDKEY         = SDLK_END;
const int PAGEUP         = SDLK_PAGEUP;
const int PAGEDOWN       = SDLK_PAGEDOWN;
const int F1             = SDLK_F1;
const int F2             = SDLK_F2;
const int F3             = SDLK_F3;
const int F4             = SDLK_F4;
const int F5             = SDLK_F5;
const int F6             = SDLK_F6;
const int F7             = SDLK_F7;
const int F8             = SDLK_F8;
const int F9             = SDLK_F9;
const int F10            = SDLK_F10;
const int F11            = SDLK_F11;
const int F12            = SDLK_F12;
const int F13            = SDLK_F13;
const int F14            = SDLK_F14;
const int F15            = SDLK_F15;
const int NUMLOCK        = SDLK_NUMLOCK;
const int SCROLLOCK      = SDLK_SCROLLOCK;
const int CAPSLOCK       = SDLK_CAPSLOCK;
const int LSUPER         = SDLK_LSUPER;
const int RSUPER         = SDLK_RSUPER;
const int MODE           = SDLK_MODE;
const int HELP           = SDLK_HELP;
const int PRINTSCREEN    = SDLK_PRINT;
const int SYSREQ         = SDLK_SYSREQ;
const int BREAK          = SDLK_BREAK;
const int MENU           = SDLK_MENU;
const int POWER          = SDLK_POWER;
const int EURO           = SDLK_EURO;

bool SDLInput::GotEvent(Event e)
{
    FUNC("SDLInput::GotEvent");
    if(e.type == MSGUpdate)
	ENDRETURN(Update(e.update.time));

    if(e.type == MSGLoadResources)
	ENDRETURN(Init());

    if(e.type == MSGReleaseResources)
	ENDRETURN(Close());

    ENDRETURN(true);
}

bool SDLInput::Update(int time)
{
    FUNC("SDLInput::Update");
    assert(Kern);

    SDL_Event sdlEvent;

    bool sendmouse = false;

    SDL_PumpEvents();

    vector<SDL_Event> saved;

    while(SDL_PollEvent(&sdlEvent))
    {
	switch(sdlEvent.type)
	{
	    case SDL_MOUSEMOTION:
	    {
		mouse.x = sdlEvent.motion.x;
		mouse.y = sdlEvent.motion.y;
		sendmouse = true;
		break;
	    }
	    case SDL_MOUSEBUTTONDOWN:
	    case SDL_MOUSEBUTTONUP:
	    {
		mouse.x = sdlEvent.button.x;
		mouse.y = sdlEvent.button.y;
		sendmouse = true;
		
		if(sdlEvent.button.button == SDL_BUTTON_LEFT)
		{
		    if(sdlEvent.button.state == SDL_PRESSED)
			mouse.Left = true;
		    else
			mouse.Left = false;
		}
		else if(sdlEvent.button.button == SDL_BUTTON_MIDDLE)
		{
		    if(sdlEvent.button.state == SDL_PRESSED)
			mouse.Middle = true;
		    else
			mouse.Middle = false;
		}
		else if(sdlEvent.button.button == SDL_BUTTON_RIGHT)
		{
		    if(sdlEvent.button.state == SDL_PRESSED)
			mouse.Right = true;
		    else
			mouse.Right = false;
		}
		break;
	    }
	    case SDL_KEYUP:
	    {
		keydata[sdlEvent.key.keysym.unicode].State = OFF;
		keydata[sdlEvent.key.keysym.unicode].Count = 0;
		keydata[sdlEvent.key.keysym.sym].State = OFF;
		keydata[sdlEvent.key.keysym.sym].Count = 0;
		break;
	    }
	    case SDL_KEYDOWN:
	    {
		keydata[sdlEvent.key.keysym.unicode].State = ON;
		keydata[sdlEvent.key.keysym.sym].State = ON;
		break;
	    }
	    default:
	    {
		saved.push_back(sdlEvent);
		break;
	    }
	}
    }

    SDL_Event *events = new SDL_Event[saved.size()];

    if(saved.size() && sdlEvent.type != SDL_VIDEOEXPOSE)
    {
	for(int i=0; i < (signed)saved.size(); i++)
	    events[i] = saved[i];
	SDL_PeepEvents(events, saved.size(), SDL_ADDEVENT, 0);
    }

    delete[] events;

    Event ev;
    ev.type = MSGKey;

    ev.key.mod = SDL_GetModState();

    for(int i=0; i < 512; i++)
    {
	ev.key.keys[i] = (keydata[i].State != OFF);

	keydata[i].Speed += time;

	if(keydata[i].State == ON)
	{
	    keydata[i].State = PAUSED;

	    ev.key.txtkeys[i] = true;
	    continue;
	}

	if(keydata[i].State == PAUSED)
	{
	    keydata[i].Count += time;
	    
	    if(keydata[i].Count > 700)
	    {
		keydata[i].State = HELD;

		ev.key.txtkeys[i] = true;
		continue;
	    }
	}

	if(keydata[i].State == HELD)
	{
	    if(keydata[i].Speed > 30)
	    {
		keydata[i].Speed = 0;
		
		ev.key.txtkeys[i] = true;
		continue;
	    }
	}

	if(keydata[i].State == OFF)
	    keydata[i].Count = keydata[i].Speed = 0;

	ev.key.txtkeys[i] = false;
    }

    if(!Kern->SendEvent(CATCHER(this), ev))
	ENDRETURN(false);

    if(sendmouse)
    {
	bool first = false;
	if(prevy == -1)
	    first = true;
	ev.type = MSGMouse;
	ev.mouse.dx = mouse.x - prevx;
	ev.mouse.dy = prevy - mouse.y;
	ev.mouse.dz = 0;

	ev.mouse.x = mouse.x;
	ev.mouse.y = mouse.y;
	ev.mouse.z = 0;
	//! \todo get the z value
	
	prevx = mouse.x;
	prevy = mouse.y;

	ev.mouse.left = mouse.Left;
	ev.mouse.middle = mouse.Middle;
	ev.mouse.right = mouse.Right;

	if(!first && !Kern->SendEvent(CATCHER(this), ev))
	    ENDRETURN(false);
    }

    if(Grab)
    {
	prevx = prevy = 200;
	SDL_WarpMouse(200, 200);
	SDL_ShowCursor(SDL_DISABLE);
    }

    ENDRETURN(true);
}

SDLInput::SDLInput(Pointer<Kernel> k, bool grab)
{
    FUNC("SDLInput::SDLInput(k)");
    Kern = k;
    prevx = prevy = -1;
    k->AddChild(CATCHER(this),
		MASKWindow | MASKResources | MASKUpdate |
		MASKMouseMove);
    Grab = grab;
    END();
}

bool SDLInput::Init()
{
    FUNC("SDLInput::Init");
    if(!SDL_WasInit(SDL_INIT_JOYSTICK))
    {
	if(SDL_Init(SDL_INIT_JOYSTICK | SDL_INIT_NOPARACHUTE) < 0)
	{
	    Log << "SDL_Init() failed. SDL Error: " << SDL_GetError() << endl;
	    SDL_Quit();
	    ENDRETURN(false);
	}
    }
    for(int i=0; i < 512; i++)
	keydata[i].State = OFF;
    
    mouse.Left = mouse.Middle = mouse.Right = false;

    SDL_EnableUNICODE(1);

    ENDRETURN(true);
}

bool SDLInput::Close()
{
    FUNC("SDLInput::Close");
    SDL_QuitSubSystem(SDL_INIT_JOYSTICK);
    ENDRETURN(true);
}
