#ifndef SDLINPUT_H
#define SDLINPUT_H

#include "general.h"
#include "memory.h"
#include "kernel.h"
#include "event.h"

/*!
 * \file   sdlinput.h
 * \author  <baldurk@baldurk.org>
 * \date   Tue Jan  3 00:46:57 2006
 * 
 * \brief  A class for Input via SDL  
 */

//! an enum to define the state of a key
enum state { OFF = 0, ON, PAUSED, HELD };

//! a structure to contain details of a key
struct Key
{
    //! the state of the key
    state State;
    //! The number of milliseconds since the key was pressed
    int  Count,
	//! the speed of the repetition
	Speed;
    //! if this key has been processed
    bool proccessed;
};

//! a structure to contain details of the mouse
struct Mouse
{
    //! whether the left mouse button is down or not
    bool Left;
    //! whether the middle mouse button is down or not
    bool Middle;
    //! whether the right mouse button is down or not
    bool Right;
    //! the x co-ord of the mouse
    int  x,
	//! the y co-ord of the mouse
	y;
};

/*! 
 * \brief A class to handle the SDL input subsystem
 *
 * This class will check SDL for new input events (mouse
 * movements, mouse clicks and keyboard events).
 * When any updates are received they are sent back to
 * the kernel.
 *
 * \todo support joysticks.
 */

class SDLInput : public EventCatcher
{
  public:
    //! A default constructor. This shouldn't be used and will error when used.
    SDLInput()
    { FUNC("SDLInput::SDLInput");assert(1==0 && "Don't use the default constructor");END(); }
    /*!
     * \brief A constructor that sets the kernel and adds itself as a child.
     * 
     * This constructor sets the kernel that the class should send
     * updates about the keyboards and mice. It then adds itself as
     * a child of that kernel. Optionally grabs and keeps the mouse pointer
     *
     * \param k    the kernel this class should send updates to
     * \param grab whether to grab the mouse pointer
     */

    SDLInput(Pointer<Kernel> k, bool grab = false);

    /*!
     * \brief This allows the class to receive events
     *
     * When the class receives an event, it comes into this
     * function, which is parsed and possibly passed into internal
     * functions.
     * 
     * \param e the event being received.
     */
    bool GotEvent(Event e);
  private:
    /*!
     * \brief This function pumps the SDL loop and grabs any new events
     *
     * \param time the time passed since the last update.
     */
    bool Update(int time);

    //! This function initialises the input part of SDL and sets up input via SDL
    bool Init();

    //! This function closes down the SDL Input subsystem
    bool Close();

    //! An internal pointer to the kernel controlling the input so that it can send
    //! messages back to the other children
    Pointer<Kernel> Kern;

    //! Grab the mouse
    bool Grab;

    //! previous mouse x
    int prevx,
	//! previous mouse y
	prevy;

    //! details of the mouse
    Mouse mouse;
    //! details of the keyboard
    Key keydata[512];
};

#include "input.h"

#endif
