#ifndef TEXTURE_H
#define TEXTURE_H

#include "general.h"
#include "GLee.h"
#include <GL/glu.h>
#include "compress.h"

#ifdef WIN32
#include <windows.h>
#include <winsock2.h>
#else
#include <netinet/in.h>
#endif

/*!
 * \file   texture.h
 * \author  <baldurk@baldurk.org>
 * \date   Wed Jan  4 00:35:06 2006
 *
 * \brief  Texture loading class
 */

//! Details of a PNG image
struct PNGImage
{
    //! A default constructor
    PNGImage() : cnt(0), byte(0) {}
    //! the image data
    unsigned short *data;
    //! the current short
    int cnt,
	//! the current byte (0/1)
	byte;
    
    //! get a byte
    unsigned char GetByte();
    //! get a long (4 bytes)
    unsigned int GetLong();
};

//! Details of a chunk in a PNG file
struct PNGChunk
{
    long len,			/*!< the length of the chunk (excluding the header) */
	type,			/*!< the type of the chunk */
	crc;			/*!< the CRC of the chunk \todo implement crc checking */
    unsigned char *data;	/*!< the chunk's data */
};

//! Details of a texture - its filename and properties.
//! This is used to determine if a texture has already been
//! loaded and can be reused.
struct TextureDetails
{
    //TextureDetails(string s="", int f=0) : name(s), flags(f) {}
    //! A constructor to set details
    TextureDetails(const char *s="", int f=0) : name(s), flags(f) {}
    string name;		/*!< the filename used to load the texture */
    int flags;			/*!< the flags specifying the texture's properties */
    bool operator < (TextureDetails other) const
    { FUNC("TextureDetails::<");ENDRETURN(name < other.name && flags < other.flags); } /*!< the < operator */
    bool operator == (TextureDetails other) const
    { FUNC("TextureDetails::==");ENDRETURN(name == other.name && flags == other.flags); } /*!< the == operator */
};

//! Use linear filtering instead of mipmapping
#define LINEAR     0x0001
//! This texture should be clamped in the X direction
#define XCLAMP     0x0002
//! This texture should be clamped in the Y direction
#define YCLAMP     0x0004
//! This texture is clamped in X and Y
#define XYCLAMP    (XCLAMP | YCLAMP)
//! This texture should be loaded as a cubemap
#define CUBEMAP    0x0008
//! This texture should be stored in floats instead of bytes
#define FLOAT      0x0010
//! This texture is black and white - allows optimisations
//! \todo autodetect this
#define BLACKWHITE 0x0020
//! This texture has no alpha channel - allows optimisations
//! \todo autodetect this
#define NOALPHA    0x0040

//! a macro to check flags
#define HASFLAGS(a, f) ((a) & (f) == (f))

//! the different texture operations

enum TexOp {ZERO=0, ONE, A_ALPHA, B_ALPHA, A_INVALPHA, B_INVALPHA,
	    A_COL, B_COL, A_INVCOL, B_INVCOL};

/*!
 * \brief A class to contain the data for a texture
 * 
 * this class allows you to load texture data from a
 * file then play around with it. If you only need to
 * load a texture from a file, just pass the details to
 * the Texture class which will invoke TextureData for you.
 *
 * \todo Add functions to modify/generate data.
 * 
 * \note This class is API-agnostic
 * \warning This class stores textures as RGBA floats. That means
 *          it is BIG. Using it on a 4096x4096 image will use 268MB
 *          of memory.
 */
class TextureData
{
  public:
    //! A default constructor
    TextureData() : width(2), height(2), data(NULL) { data = new float[width*height*4];}

    //! A default destructor \todo clean up the data if it hasn't been cleaned
    ~TextureData() { /*delete[] data;*/ }
    /*!
     * \brief Loads the texture data from a file
     * 
     * \param details the details of the texture to load
     */
    bool Load(TextureDetails details);

    /*!
     * \brief Loads the texture data from memory
     * 
     * \param width  the width of the image data
     * \param height the height of the image data
     * \param alpha    true if there is alpha, false if not
     * \param imdata   the data to load
     */
    bool Load(int width, int height, bool alpha, unsigned char *imdata);

    /*!
     * \brief Creates a blank black, full alpha texture
     * 
     * \param width  the width of texture to create
     * \param height the height of texture to create
     */
    bool Blank(int width, int height);

    /*!
     * \brief Inverts the colour (and optionally the alpha) of the texture
     * 
     */
    bool Invert(bool alpha = false);

    /*!
     * \brief Generates Perlin noise
     * 
     * \param numoct  The number of octaves to generate
     * \param Freq    The frequency to start with - the lowest
     * \param Amp     The amplitude to start with
     * \param ampmod  The modifier on the amplitude each octave
     * \param freqmod The modifier on the frequency each octave
     */
    bool Perlin(int numoct, float Freq, float Amp, float freqmod, float ampmod);

    //! Scales the data to fit 0 - 1
    bool ScaleClamp();

    //! Clamps the data to fit 0 - 1
    bool Clamp();
    
    //! Wraps the data to fit 0 - 1. 1.5 becomes 0.5 and 2.7 becomes 0.7. -0.3 becomes 0.7
    inline bool Wrap()
    {
	FUNC("TextureData::Wrap");
	float foo;
	for(int i=0; i < width*height*4; i++)
	{
	    if(data[i] > 0.0f)
		data[i] = modf(data[i], &foo);
	    else
		data[i] = 1.0f + modf(data[i], &foo);
	}
	ENDRETURN(true);
    }

    /*!
     * \brief Blurs the image via a gaussian blur
     * 
     * \param radius the radius of the blur
     */
    bool Blur(int radius);

    /*!
     * \brief Blend another image on this one
     *
     * \param aop   how to blend data a (this data)
     * \param bop   how to blend data b
     * \param other the other image
     */
    bool Blend(TexOp aop, TexOp bop, TextureData other);

    /*!
     * \brief Translates the image
     * 
     * \param trans the vector to translate by
     * \param col   the border colour to use
     * \param wrap  true if the texture should wrap
     */
    bool Translate(Vectori trans, Vectorf col, bool wrap);

    //! Scales the data up or down by a constant
    //! \param constant the constant to scale by
    inline bool Scale(float constant)
    { FUNC("TextureData::Scale");for(int i=0; i < width*height*4; i++) data[i] += constant;ENDRETURN(true); }

    //! Return the bits per pixel
    inline int Bpp()
    {
#ifdef DEBUG_SLOW
	FUNC("TextureData::Bpp");ENDRETURN(32);
#else
	return 32;
#endif
    }
    //! Return the width of the image
    inline int Width() 
    {
#ifdef DEBUG_SLOW
	FUNC("TextureData::Width");ENDRETURN(width);
#else
	return width;
#endif
    }
    //! Return the height of the image
    inline int Height()
    {
#ifdef DEBUG_SLOW
	FUNC("TextureData::Height");ENDRETURN(height);
#else
	return height;
#endif
    }
    //! Return the image data
    inline float *Data()
    {
#ifdef DEBUG_SLOW
	FUNC("TextureData::Data");ENDRETURN(data);
#else
	return data;
#endif
    }
  private:

    /*!
     * \brief Loads a TGA texture
     * 
     * \param details the details of the file/texture to load
     */
    bool LoadTGA(TextureDetails details);

    /*!
     * \brief Loads a PNG texture
     * 
     * \param details the details of the file/texture to load
     */
    bool LoadPNG(TextureDetails details);

    /*!
     * \brief Generate a pseudorandom noise value based on parameters
     * 
     * \param x the x parameter to the function
     * \param y the y parameter to the function
     */
    float Noise2D(int x, int y);

    /*!
     * \brief Generate smoothed noise value based on a gaussian spread of Noise2D values
     * 
     * \param x the x parameter to the function
     * \param y the y parameter to the function
     */
    float SmoothedNoise2D(int x, int y);

    /*!
     * \brief Interpolate between two values
     * 
     * \param a the parameter to interpolate from
     * \param b the parameter to interpolate to
     * \param x the amount to interpolate by in the range 0 - 1
     */
    float Interpolate1D(float a, float b, float x);

    /*!
     * \brief Interpolate between 2D noise values
     *
     * This function interpolates between 2D smoothed noise values
     * so that it is possible to get smoothed 2D noise. The
     * co-ordinates are floats, and the value returned is interpolated
     * between the nearest integer values.
     * 
     * \param x the x co-ordinate
     * \param y the y co-ordinate
     */
    float InterpolatedNoise2D(float x, float y);

    int width,			/*!< the width of the texture */
	height;			/*!< the height of the texture */
    float *data;		/*!< the texture data */
};

//! a textures details + ID structure
struct Tex
{
    //! the details of the texture
    TextureDetails details;
    //! the texture ID
    GLuint ID;
};

/*!
 * \brief A Texture and associated data
 * 
 * This class contains the details of a single texture.
 * The class provides static methods for the loading and
 * managing of textures. This allows textures to be reference
 * counted.
 */
class Texture
{
  public:
    //! A default constructor
    Texture() : ID(0) {}

    /*!
     * \brief Loads the texture
     * 
     * This function loads the specified texture if it's not
     * already been loaded, and gets the old ID.
     *
     * \param details the details of the texture to load
     */
    bool Load(TextureDetails details);

    /*!
     * \brief Loads the texture
     * 
     * This function loads the specified texture This
     * function will take a TextureData class as its paramater.
     * The TextureData class is used for loading texture data,
     * but can also perform operations on it - this allows you
     * to modify your textures before loading, possibly even
     * procedurally generate them.
     *
     * \note This function will always get a new texture, no matter
     *       what, given the infeasibility of tracking duplicates from
     *       just texture data.
     *
     * \param data    the TextureData to load
     * \param details the TextureDetails of the data to load
     */
    bool Load(TextureData data, TextureDetails details=TextureDetails());

    /*!
     * \brief Binds this texture
     * 
     * This function binds the texture as the current texture
     * on the current texture unit.
     */
    bool Bind()
    {FUNC("Texture::Bind"); assert(this);glBindTexture(GL_TEXTURE_2D, ID); ENDRETURN(true);}

    /*!
     * \brief Releases the texture
     * 
     * This function releases the texture, and possibly
     * removes it completely from memory if there are no
     * other textures using it.
     */
    bool Release();

    /*!
     * \brief Releases all textures that are still loaded
     * 
     * Textures which are released, or no longer used, are
     * kept in memory until that memory needs to be replaced.
     * This function releases all the textures - in use or not.
     */
    bool ReleaseAll();
  private:
    GLuint ID;			/*!< the current ID to use */
    //! The list of all the textures.
    static vector<Tex> textures;
    //static map<TextureDetails, GLuint> textures;
    //! the counts of which textures are used
    static vector<short> refcounts;
};

#endif
