#include "texture.h"

/*!
 * \file   texturedata.cpp
 * \author  <baldurk@baldurk.org>
 * \date   Fri Feb  3 15:59:50 2006
 * 
 * \brief  functions for dealing with Texture data -
 *         performing operations on it
 */


bool TextureData::Load(int imwidth, int imheight, bool alpha, unsigned char *imdata)
{
    FUNC("TextureData::Load(imdata)");

    width = imwidth;
    height = imheight;

    delete[] data;
    data = new float[width*height*4];
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	{
	    data[4*(x*height+y)+0] = float(imdata[(3+alpha)*(x*height+y)+0])/255.0f;
	    data[4*(x*height+y)+1] = float(imdata[(3+alpha)*(x*height+y)+1])/255.0f;
	    data[4*(x*height+y)+2] = float(imdata[(3+alpha)*(x*height+y)+2])/255.0f;
	    if(alpha)
		data[4*(x*height+y)+3] = float(imdata[(3+alpha)*(x*height+y)+3])/255.0f;
	    else
		data[4*(x*height+y)+3] = 1.0f;
	}
    ENDRETURN(true);
}

bool TextureData::Blank(int imwidth, int imheight)
{
    FUNC("TextureData::Blank");

    width = imwidth;
    height = imheight;

    delete[] data;
    data = new float[width*height*4];
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	{
	    data[4*(x*height+y)+0] = 0.0f;
	    data[4*(x*height+y)+1] = 0.0f;
	    data[4*(x*height+y)+2] = 0.0f;
	    data[4*(x*height+y)+3] = 1.0f;
	}
    ENDRETURN(true);
}

bool TextureData::Invert(bool alpha)
{
    FUNC("TextureData::Invert");
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	{
	    data[4*(x*height+y)+0] = 1.0f-data[4*(x*height+y)+0];
	    data[4*(x*height+y)+1] = 1.0f-data[4*(x*height+y)+1];
	    data[4*(x*height+y)+2] = 1.0f-data[4*(x*height+y)+2];
	    if(alpha)
		data[4*(x*height+y)+3] = 1.0f-data[4*(x*height+y)+3];
	}
    ENDRETURN(true);
}

float TextureData::Noise2D(int x, int y)
{
#ifdef DEBUG_SLOW
    FUNC("TextureData::Noise2D");
#endif
    int n = x + y * 57;
    n = (n<<13) ^ n;
#ifdef DEBUG_SLOW
    ENDRETURN(1.0f - ((n * (n * n * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0);
#else
    return (1.0f - ((n * (n * n * 15731 + 789221) + 1376312589) & 0x7fffffff) / 1073741824.0);
#endif
}

float TextureData::SmoothedNoise2D(int x, int y)
{
#ifdef DEBUG_SLOW
    FUNC("TextureData::SmoothedNoise2D");
#endif
#ifdef DEBUG_SLOW
    ENDRETURN(Noise2D(x, y)/4.0f +
	(Noise2D(x+1, y) + Noise2D(x-1, y) + Noise2D(x, y+1) + Noise2D(x, y-1))/8.0f +
	(Noise2D(x+1, y+1) + Noise2D(x+1, y-1) + Noise2D(x-1, y+1) + Noise2D(x-1, y-1))/16.0f);
#else
    return (Noise2D(x, y)/4.0f +
	(Noise2D(x+1, y) + Noise2D(x-1, y) + Noise2D(x, y+1) + Noise2D(x, y-1))/8.0f +
	(Noise2D(x+1, y+1) + Noise2D(x+1, y-1) + Noise2D(x-1, y+1) + Noise2D(x-1, y-1))/16.0f);
#endif
}

float TextureData::Interpolate1D(float a, float b, float x)
{
#ifdef DEBUG_SLOW
    FUNC("TextureData::Interpolate1D");
#endif
    float f = (1.0f - cosf(x * PI)) * 0.5f;

#ifdef DEBUG_SLOW
    ENDRETURN(a * (1.0f - f) + b * f);
#else
    return (a * (1.0f - f) + b * f);
#endif
}

float TextureData::InterpolatedNoise2D(float x, float y)
{
#ifdef DEBUG_SLOW
    FUNC("TextureData::InterpolateNoise2D");
#endif
    float v1 = SmoothedNoise2D(int(x),   int(y));
    float v2 = SmoothedNoise2D(int(x)+1, int(y));
    float v3 = SmoothedNoise2D(int(x),   int(y)+1);
    float v4 = SmoothedNoise2D(int(x)+1, int(y)+1);

    float i1 = Interpolate1D(v1, v2, x-int(x));
    float i2 = Interpolate1D(v3, v4, x-int(x));

#ifdef DEBUG_SLOW
    ENDRETURN(Interpolate1D(i1, i2, y-int(y)));
#else
    return (Interpolate1D(i1, i2, y-int(y)));
#endif
}

bool TextureData::Perlin(int numoct, float Freq, float Amp, float freqmod, float ampmod)
{
    FUNC("TextureData::Perlin");
    float freq = Freq, amp = Amp;

    for(int i=0; i < numoct; i++)
    {
	for(int x=0; x < width; x++)
	{
	    for(int y=0; y < height; y++)
	    {
		float X = amp*(InterpolatedNoise2D(float(y)/float(width*freq),
						   float(x)/float(height*freq)));
		data[4*(height*x+y)+0] += X;
	    }
	}
	freq *= freqmod;amp *= ampmod;
    }

    for(int x=0; x < width; x++)
    {
	for(int y=0; y < height; y++)
	{
	    data[4*(height*x+y)+1] = data[4*(height*x+y)+0];
	    data[4*(height*x+y)+2] = data[4*(height*x+y)+0];
	}
    }
    ENDRETURN(true);
}

bool TextureData::Clamp()
{
    FUNC("TextureData::Clamp");
    for(int x=0; x < width; x++)
    {
	for(int y=0; y < height; y++)
	{
	    if(data[4*(height*x+y)+0] > 1.0f)
		data[4*(height*x+y)+0] = 1.0f;
	    if(data[4*(height*x+y)+0] < 0.0f)
		data[4*(height*x+y)+0] = 0.0f;

	    if(data[4*(height*x+y)+1] > 1.0f)
		data[4*(height*x+y)+1] = 1.0f;
	    if(data[4*(height*x+y)+1] < 0.0f)
		data[4*(height*x+y)+1] = 0.0f;

	    if(data[4*(height*x+y)+2] > 1.0f)
		data[4*(height*x+y)+2] = 1.0f;
	    if(data[4*(height*x+y)+2] < 0.0f)
		data[4*(height*x+y)+2] = 0.0f;
	}
    }
    ENDRETURN(true);
}

bool TextureData::ScaleClamp()
{
    FUNC("TextureData::ScaleClamp");
    float minval = data[0], maxval = data[0];
    for(int x=0; x < width; x++) 
    {
	for(int y=0; y < height; y++)
	{ 
	    if(data[4*(height*x+y)+0] < minval)
		minval = data[4*(height*x+y)+0];
	    if(data[4*(height*x+y)+0] > maxval)
		maxval = data[4*(height*x+y)+0];
	}
    }
    for(int x=0; x < width; x++)
    {
	for(int y=0; y < height; y++)
	{
	    data[4*(height*x+y)+0] -= minval;
	    data[4*(height*x+y)+0] /= maxval-minval;
	    data[4*(height*x+y)+1] = data[4*(height*x+y)+0];
	    data[4*(height*x+y)+2] = data[4*(height*x+y)+0];
	}
    }
    ENDRETURN(true);
}

bool TextureData::Blur(int radius)
{
    FUNC("TextureData::Blur");
    float *buf = new float[4*height*width];

    // blur in the X direction, then the Y direction

    float sigma = 1.0f+radius/2.0f;

    float *factors = new float[(radius*2+1)];

    // calculate a convolution matrix - only 1D
    if(radius == 0)
	for(int i=0; i < radius*2+1; i++)
	    factors[i] = 1.0f;
    else
	for(int X=-radius; X <= radius; X++)
	    factors[(X+radius)] = exp(-sqrtf(X*X)/(2*sigma*sigma))/(2.0f*PI*sigma);

    // blur the data into a temporary buffer in the X direction
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	    for(int i=0; i < 4; i++)
	    {
		float total = 0.0f;
		buf[4*(height*x+y)+i] = total;
		// iterate over the matrix, summing the neighbouring
		// pixels and multiplying them by the factor
		for(int X=-radius; X <= radius; X++)
		{
		    if(X+x >= width || X+x < 0 ||
		       (data[4*(height*(x+X)+y)+3] == 0.0f && i != 3))
			continue;
		    float f = factors[X+radius];
		    total += f;
		    buf[4*(height*x+y)+i] += data[4*(height*(x+X)+y)+i]*f;
		}
		if(total != 0.0f)
		    buf[4*(height*x+y)+i] /= total;
	    }

    // copy the temporary buffer data into the data
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	    for(int i=0; i < 4; i++)
		data[4*(height*x+y)+i] =  buf[4*(height*x+y)+i];

    // blur in the Y direction
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	    for(int i=0; i < 4; i++)
	    {
		float total = 0.0f;
		buf[4*(height*x+y)+i] = total;
		// iterate over the matrix, summing the neighbouring
		// pixels and multiplying them by the factor
		for(int Y=-radius; Y <= radius; Y++)
		{
		    if(Y+y >= height || Y+y < 0)
			continue;
		    float f = factors[Y+radius];
		    total += f;
		    buf[4*(height*x+y)+i] += data[4*(height*x+y+Y)+i]*f;
		}
		if(total != 0.0f)
		    buf[4*(height*x+y)+i] /= total;
	    }

    // copy the temporary buffer into the real buffer
    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	    for(int i=0; i < 4; i++)
		data[4*(height*x+y)+i] =  buf[4*(height*x+y)+i];

    delete[] buf;
    delete[] factors;

    ENDRETURN(true);
}

bool TextureData::Translate(Vectori trans, Vectorf col, bool wrap)
{
    FUNC("TextureData::Translate");
    if(trans.x > 0)
    {
	for(int y=0; y < height; y++)
	{
	    for(int x=width-1-trans.x; x >= 0; x--)
	    {
		data[4*(width*y+(x+trans.x))+0] = data[4*(width*y+x)+0];
		data[4*(width*y+(x+trans.x))+1] = data[4*(width*y+x)+1];
		data[4*(width*y+(x+trans.x))+2] = data[4*(width*y+x)+2];
		data[4*(width*y+(x+trans.x))+3] = data[4*(width*y+x)+3];
	    }
	    for(int x=0; x < trans.x; x++)
	    {
		data[4*(width*y+x)+0] = col.x;
		data[4*(width*y+x)+1] = col.y;
		data[4*(width*y+x)+2] = col.z;
		data[4*(width*y+x)+3] = col.w;
	    }
	}
    }
    else
    {
	for(int y=0; y < height; y++)
	{
	    for(int x=0; x < width+trans.x; x++)
	    {
		data[4*(width*y+x)+0] = data[4*(width*y+(x-trans.x))+0];
		data[4*(width*y+x)+1] = data[4*(width*y+(x-trans.x))+1];
		data[4*(width*y+x)+2] = data[4*(width*y+(x-trans.x))+2];
		data[4*(width*y+x)+3] = data[4*(width*y+(x-trans.x))+3];
	    }
	    for(int x=width+trans.x; x < width; x++)
	    {
		data[4*(width*y+x)+0] = col.x;
		data[4*(width*y+x)+1] = col.y;
		data[4*(width*y+x)+2] = col.z;
		data[4*(width*y+x)+3] = col.w;
	    }
	}
    }

    if(trans.y > 0)
    {
	for(int x=0; x < width; x++)
	{
	    for(int y=height-1-trans.y; y >= 0; y--)
	    {
		data[4*(width*(y+trans.y)+x)+0] = data[4*(width*y+x)+0];
		data[4*(width*(y+trans.y)+x)+1] = data[4*(width*y+x)+1];
		data[4*(width*(y+trans.y)+x)+2] = data[4*(width*y+x)+2];
		data[4*(width*(y+trans.y)+x)+3] = data[4*(width*y+x)+3];
	    }
	    for(int y=0; y < trans.y; y++)
	    {
		data[4*(width*y+x)+0] = col.x;
		data[4*(width*y+x)+1] = col.y;
		data[4*(width*y+x)+2] = col.z;
		data[4*(width*y+x)+3] = col.w;
	    }
	}
    }
    else
    {
	for(int x=0; x < width; x++)
	{
	    for(int y=0; y < height+trans.y; y++)
	    {
		data[4*(width*y+x)+0] = data[4*(width*(y-trans.y)+x)+0];
		data[4*(width*y+x)+1] = data[4*(width*(y-trans.y)+x)+1];
		data[4*(width*y+x)+2] = data[4*(width*(y-trans.y)+x)+2];
		data[4*(width*y+x)+3] = data[4*(width*(y-trans.y)+x)+3];
	    }
	    for(int y=height+trans.y; y < height; y++)
	    {
		data[4*(width*y+x)+0] = col.x;
		data[4*(width*y+x)+1] = col.y;
		data[4*(width*y+x)+2] = col.z;
		data[4*(width*y+x)+3] = col.w;
	    }
	}
    }
    ENDRETURN(true);
}

bool TextureData::Blend(TexOp aop, TexOp bop, TextureData other)
{
    FUNC("TextureData::Blend");
    float *buf = new float[4*height*width];

    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	    for(int i=0; i < 4; i++)
		buf[4*(height*x+y)+i] = 0.0f;

    for(int x=0; x < width; x++)
    {
	for(int y=0; y < height; y++)
	{
	    int X = int(float(x)/float(width)*float(other.width));
	    int Y = int(float(y)/float(height)*float(other.height));
	    int idxa = 4*(x*height+y), idxb = 4*(X*other.height+Y);

	    Vectorf col;
	    TexOp op;
	    for(int i=0; i < 2; i++)
	    {
		if(i == 0)
		{
		    op = aop;
		    col.x = data[idxa+0];col.y = data[idxa+1];col.z = data[idxa+2];col.w = data[idxa+3];
		}
		else
		{
		    op = bop;
		    col.x = other.data[idxb+0];col.y = other.data[idxb+1];col.z = other.data[idxb+2];col.w = other.data[idxb+3];
		}

		float factor = 0.0f;
		switch(op)
		{
		    case ZERO:
			break;
		    case ONE:
			buf[idxa+0] += col.x;
			buf[idxa+1] += col.y;
			buf[idxa+2] += col.z;
			buf[idxa+3] += col.w;
			break;
		    case A_ALPHA:
			buf[idxa+0] += col.x*data[idxa+3];
			buf[idxa+1] += col.y*data[idxa+3];
			buf[idxa+2] += col.z*data[idxa+3];
			buf[idxa+3] += col.w*data[idxa+3];
			break;
		    case B_ALPHA:
			buf[idxa+0] += col.x*other.data[idxb+3];
			buf[idxa+1] += col.y*other.data[idxb+3];
			buf[idxa+2] += col.z*other.data[idxb+3];
			buf[idxa+3] += col.w*other.data[idxb+3];
			break;
		    case A_INVALPHA:
			buf[idxa+0] += col.x*(1.0f-data[idxa+3]);
			buf[idxa+1] += col.y*(1.0f-data[idxa+3]);
			buf[idxa+2] += col.z*(1.0f-data[idxa+3]);
			buf[idxa+3] += col.w*(1.0f-data[idxa+3]);
			break;
		    case B_INVALPHA:
			buf[idxa+0] += col.x*(1.0f-other.data[idxb+3]);
			buf[idxa+1] += col.y*(1.0f-other.data[idxb+3]);
			buf[idxa+2] += col.z*(1.0f-other.data[idxb+3]);
			buf[idxa+3] += col.w*(1.0f-other.data[idxb+3]);
			break;
		    case A_COL:
			factor = (data[idxa+0]+data[idxa+1]+data[idxa+2])/3.0f;
			buf[idxa+0] += col.x*factor;
			buf[idxa+1] += col.y*factor;
			buf[idxa+2] += col.z*factor;
			buf[idxa+3] += col.w*factor;
			break;
		    case B_COL:
			factor = (other.data[idxb+0]+other.data[idxb+1]+other.data[idxb+2])/3.0f;
			buf[idxa+0] += col.x*factor;
			buf[idxa+1] += col.y*factor;
			buf[idxa+2] += col.z*factor;
			buf[idxa+3] += col.w*factor;
			break;
		    case A_INVCOL:
			factor = 1.0f-((data[idxa+0]+data[idxa+1]+data[idxa+2])/3.0f);
			buf[idxa+0] += col.x*factor;
			buf[idxa+1] += col.y*factor;
			buf[idxa+2] += col.z*factor;
			buf[idxa+3] += col.w*factor;
			break;
		    case B_INVCOL:
			factor = 1.0f-((other.data[idxb+0]+other.data[idxb+1]+other.data[idxb+2])/3.0f);
			buf[idxa+0] += col.x*factor;
			buf[idxa+1] += col.y*factor;
			buf[idxa+2] += col.z*factor;
			buf[idxa+3] += col.w*factor;
			break;
		    default:
			Log << "Unknown texture operation" << endl;
			assert(0);
		}
	    }
	}
    }

    for(int x=0; x < width; x++)
	for(int y=0; y < height; y++)
	    for(int i=0; i < 4; i++)
	    {
		if(buf[4*(height*x+y)+i] < 0.0f)
		    data[4*(height*x+y)+i] = 0.0f;
		else if(buf[4*(height*x+y)+i] > 1.0f)
		    data[4*(height*x+y)+i] = 1.0f;
		else
		    data[4*(height*x+y)+i] = buf[4*(height*x+y)+i];
	    }

    delete[] buf;

    ENDRETURN(true);
}
