#ifndef TIMER_H
#define TIMER_H

/*!
 * \file   timer.h
 * \author  <baldurk@baldurk.org>
 * \date   Mon Jan  2 23:23:31 2006
 *
 * \brief  A timing class  
 */

#if defined(NO_SDL)
#error "timer not defined without SDL"
#else
#include <SDL.h>
#endif

/*!
 * \brief A class to keep time
 * 
 * This class, when constructed, will intialising a timing
 * system and record the initial time. Then it can be used
 * to obtain the current time in milliseconds at any point.
 */

class Timer : public MMObject
{
  public:
    /*!
     * \brief A default constructor. Intialises the time system and gets the initial time.
     * 
     * Depending on compile time flags, it uses a given system. Then
     * it records the time at the point of initialising the system.
     * This means the current time is relative to the initialisation of
     * the timer, not some arbitrary point depending on the system.
     */

    Timer()
    {
	FUNC("Timer::Timer");
#if defined(NO_SDL)
#else
	if(SDL_WasInit(SDL_INIT_TIMER) == 0)
	{
	    if(SDL_InitSubSystem(SDL_INIT_TIMER | SDL_INIT_NOPARACHUTE) < 0)
	    {
		Log << "SDL_InitSubSystem() failed. SDL Error: " << SDL_GetError() << endl;
		SDL_QuitSubSystem(SDL_INIT_TIMER);
		END();
	    }
	}

	first = SDL_GetTicks();
#endif
	END();
    }

    /*!
     * \brief This function returns the time in milliseconds
     * 
     * This function returns the time passed in milliseconds
     * since the initialisation of this timer.
     */

    int GetTime()
    {
	FUNC("Timer::GetTime");
#if defined(NO_SDL)
#else
	ENDRETURN(SDL_GetTicks() - first);
#endif
    }
  private:
    int first;			/*!< This is the time at the point of initialisation */
};

#endif
